 ! C68 support routine to get a value in a bit field to a 32 bit value
 !-----------------------------------------------------------------------------
 !  #1  original version: David J. Walker                               10/95
 !-----------------------------------------------------------------------------
 !
 !  Name:       .Xbfget
 !
 !  Parameters:
 !              Address of bit field
 !              Long value encoded with size/type information as follows
 !
 !                  Byte    Description
 !                  ~~~~    ~~~~~~~~~~~
 !                  0      Offset of bit field in bits (0-31)
 !                  1      Size of bit field in bits (1-32)
 !                  2      Not used
 !                  3      Set if unsigned, 0 if signed
 !
 !  The following assumptions are made:
 !     1. The combination of offset and size cannot be larger than 32.
 !        (i.e the value always fits into a long).
 !     2. If the address passed is odd, the combination is limited to 24.
 !
 !---------------------------------------------------------------------------

    .sect .text

    .globl .Xbfget
    .globl .Xbfgetreg

VALPTR      equ     4+0
UNSIGNED    equ     4+4
SIZE        equ     4+6
OFFSET      equ     4+7

.Xbfget:
        bclr.b  #0,VALPTR+3(a7)     ! ensure even address ?
        beq     waseven             ! ... it already was so jump
        addq.b  #8,OFFSET(a7)       ! ... and adjust offset
waseven:

!   Load up registers as follows:
!       a0      Address of bit field
!       a1      Address of sign field
!       d1      Offset of bit field
!       d2      Size of bit field

        move.l  VALPTR(a7),a0       ! get value pointer value off stack
        lea     UNSIGNED(a7),a1
        move.b  OFFSET(a7),d1       ! get offset to start
        move.b  SIZE(a7),d2         ! less size (to give amount to shift)
        andi.b  #$1f,d2             ! ensure modulo 32

!   Special internal entry point where relevant
!   registers have already been loaded

.Xbfgetreg:
        move.l  (a0),d0             ! get long that contains value
        lsl.l   d1,d0               ! left align field
        neg.b   d2
        addi.b  #32,d2              ! 32 - size
        andi.b  #$1f,d2             ! ... but modulo 32
        tst.b   (a1)                ! is value unsigned?
        bne     unsigned            ! ... YES - jump
        asr.l   d2,d0               ! move up signed value
        rts
unsigned:
        lsr.l   d2,d0               ! move up unsigned value
        rts

