 ! C68 support routine to put a value into a bit field from a 32 bit value
 !-----------------------------------------------------------------------------
 !  #1  original version: David J. Walker                               10/95
 !-----------------------------------------------------------------------------
 !
 !  Name:       .Xbfget
 !
 !  Parameters:
 !              Long value to be stored
 !              Address of bit field
 !              Short value encoded with size/offset information as follows
 !
 !                  Byte    Description
 !                  ~~~~    ~~~~~~~~~~~
 !                   0      Offset of bit field in bits (0-31)
 !                   1      Size of bit field in bits (1-32)
 !
 !  The following assumptions are made:
 !     1. The combination of offset and size cannot be larger than 32.
 !        (i.e the value always fits into a long).
 !     2. If the address passed is odd, the combination is limited to 24.
 !
 !---------------------------------------------------------------------------

    .sect .text

    .globl  .Xbfput
    .globl  .Xbfputreg

VALUE       equ     4+0
BFPTR       equ     4+4
OFFSET      equ     4+8
SIZE        equ     4+9

.Xbfput:
        bclr.b  #0,BFPTR+3(a7)      ! ensure even address ?
        beq     waseven             ! ... it already was, so jump
        addq.b  #8,OFFSET(a7)       ! ... and adjust offset
waseven:

!   Load up registers as follows:
!       a0      Address of bit field
!       a1      Address of value
!       d1      Offset of bit field
!       d2      Size of bit field

        move.l  BFPTR(a7),a0        ! get bit field pointer value off stack
        lea     VALUE(a7),a1        ! get value address
        move.b  OFFSET(a7),d1       ! get offset to start
        move.b  SIZE(a7),d2         ! get size
        andi.b  #$1f,d2             ! ensure module 32

!   Special internal entry point where relevant
!   registers have already been loaded

.Xbfputreg:

!   Start by making certain that bits in memory
!   where bit field is to be stored are zero

        move.l  (a0),d0             ! get Long containing value
        rol.l   d1,d0               ! left align field
        lsl.l   d2,d0               ! clear bits that belong to value
        lsr.l   d2,d0               ! ... and then back again
        ror.l   d1,d0               ! back into position
        move.l  d0,(a0)             ! and store value with zero in bit field

!   Now get the bits that are to be stored in the
!   bit field into position ensuring all other
!   bits are set to zero

        move.l  (a1),d0             ! get value to be stored
        neg.b   d1                  ! negate size
        addi.b  #32,d1              ! work out 32 - size
        andi.b  #$1f,d1             ! ensure modulo 32
        lsl.l   d1,d0               ! move to top clearing other bits
        lsr.l   d2,d0               ! move into position

!   Finally store the new bit field value

        or.l    d0,(a0)
        rts

