/*
 *  Test program for checking out hardware FPU support
 */

#include <stdio.h>
#include <time.h>

#define LOOP_COUNT 1000000

main()
{
    int     x;
    char    save, *ptrsave = (char *)0x280d0;
    double  d1, d2;
    float   f1, f2;
    time_t  start, looptime;
    time_t  swsfmul, hwsfmul;
    time_t  swsfdiv, hwsfdiv;
    time_t  swsfadd, hwsfadd;
    time_t  swsfsub, hwsfsub;
    time_t  swdfmul, hwdfmul;
    time_t  swdfdiv, hwdfdiv;
    time_t  swdfadd, hwdfadd;
    time_t  swdfsub, hwdfsub;

    save = *ptrsave;

    printf ("\nTest of Hardware FPU support");
    printf ("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");

    printf ("\nCalibrating loop overheads\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
    }
    looptime = time(NULL)-start;

    printf ("\n=== Multiply tests ===\n");
    *ptrsave = 0x00;
    printf ("... software (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 * 3.3;
    }
    swdfmul = time(NULL)-start - looptime;
    printf ("... software (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 * 3.3;
    }
    swsfmul = time(NULL)-start - looptime;

    *ptrsave = 0x42;
    printf ("... hardware (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 * 3.3;
    }
    hwdfmul = time(NULL)-start - looptime;
    printf ("... hardware (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 * 3.3;
    }
    hwsfmul = time(NULL)-start - looptime;

    printf ("\n=== Divide tests ===\n");
    *ptrsave = 0x00;
    printf ("... software (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 / 3.3; 
    }
    swdfdiv = time(NULL)-start - looptime;
    printf ("... software (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 / 3.3; 
    }
    swsfdiv = time(NULL)-start - looptime;


    *ptrsave = 0x42;
    printf ("... hardware (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 / 3.3;
    }
    hwdfdiv = time(NULL)-start - looptime;
    printf ("... hardware (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 / 3.3;
    }
    hwsfdiv = time(NULL)-start - looptime;

    printf ("\n=== Add tests ===\n");
    *ptrsave = 0x00;
    printf ("... software (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 + 3.3; 
    }
    swdfadd = time(NULL)-start - looptime;
    printf ("... software (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 + 3.3; 
    }
    swsfadd = time(NULL)-start - looptime;


    *ptrsave = 0x42;
    printf ("... hardware (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 / 3.3;
    }
    hwdfadd = time(NULL)-start - looptime;
    printf ("... hardware (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 / 3.3;
    }
    hwsfadd = time(NULL)-start - looptime;

    printf ("\n=== Subtract tests ===\n");
    *ptrsave = 0x00;
    printf ("... software (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 - 3.3; 
    }
    swdfsub = time(NULL)-start - looptime;
    printf ("... software (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 - 3.3; 
    }
    swsfsub = time(NULL)-start - looptime;


    *ptrsave = 0x42;
    printf ("... hardware (double)\n");
    for (x=0, d1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        d2 = d1 - 3.3;
    }
    hwdfsub = time(NULL)-start - looptime;
    printf ("... hardware (float)\n");
    for (x=0, f1 = 2.3, start=time(NULL) ; x < LOOP_COUNT ; x++) {
        f2 = f1 - 3.3;
    }
    hwsfsub = time(NULL)-start - looptime;


    *ptrsave = save;

    printf ("\n\n****** TEST RESULTS *******\n");
    printf ("operation       Float     Double\n");
    printf ("~~~~~~~~~       ~~~~~     ~~~~~~\n");
    printf (" multiply         %d         %d\n", swsfmul/hwsfmul, swdfmul/hwdfmul);
    printf (" divide           %d         %d\n", swsfdiv/hwsfdiv, swdfdiv/hwdfdiv);
    printf (" add              %d         %d\n","sdsfadd/hwsfadd, swdfadd/hddfadd);
    printf (" subtract         %d         %d\n","sdsfsub/hwsfsub, swdfsub/hddsubd);
    printf ("\n");
}
