
/*
 * C compiler
 * ==========
 *
 * Copyright 1989, 1990, 1991 Christoph van Wuellen.
 * Credits to Matthew Brandt.
 * All commercial rights reserved.
 *
 * This compiler may be redistributed as long there is no
 * commercial interest. The compiler must not be redistributed
 * without its full sources. This notice must stay intact.
 *
 * History:
 *
 * 1989   starting an 68000 C compiler, starting with material
 *		  originally by M. Brandt
 * 1990   68000 C compiler further bug fixes
 *		  started i386 port (December)
 * 1991   i386 port finished (January)
 *		  further corrections in the front end and in the 68000
 *		  code generator.
 *		  The next port will be a SPARC port
 */

#include "config.h"

#ifdef MC680X0

#include "chdr.h"
#include "expr.h"
#include "cglbdec.h"
#include "proto.h"
#include "gen68k.h"
#include "outproto.h"

#ifdef MULTIPLE_PROCESSORS
#define PRIVATE static
#undef	g_expression
#undef	g_jtrue
#undef	g_jfalse
#undef	g_stack
#undef	g_switch_table
#undef	g_switch_compare
#undef	g_entry
#undef	g_return
#undef	g_epilogue
#undef	g_allocate
#undef	g_flush
#undef	g_auto_align
#undef	g_alignment
#undef	g_label
#undef	g_branch
#undef	g_line
#undef	g_isbigendian
#undef	g_transform
#else
#define PRIVATE
#endif /* MULTIPLE_PROCESSORS */

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

/* gen68k.c */

static	ADDRESS *	as_fcall	P_((const EXPR *, FLAGS, const CHAR *, const CHAR *));
static	ADDRESS *	g_add		P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_aincdec	P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_asadd 	P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_asbitfield	P_((const EXPR *, FLAGS, OPCODE, BOOL));
static	ADDRESS *	g_asdiv 	P_((const EXPR *, FLAGS, BOOL));
static	ADDRESS *	g_aslogic	P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_asmul 	P_((const EXPR *, FLAGS));
static	ADDRESS *	g_asshift	P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_assign	P_((const EXPR *, FLAGS));
static	ADDRESS *	g_cast		P_((ADDRESS *, TYP *, TYP *, FLAGS));
static	ADDRESS *	g_deref 	P_((const EXPR *, TYP *, FLAGS));
static	ADDRESS *	g_div		P_((const EXPR *, FLAGS, BOOL));
static	ADDRESS *	g_expr		P_((const EXPR *, FLAGS));
static	ADDRESS *	g_extend	P_((ADDRESS *, TYP *, TYP *));
static	ADDRESS *	g_fcall 	P_((const EXPR *, FLAGS));
static	ADDRESS *	g_fderef	P_((const EXPR *, FLAGS));
static	ADDRESS *	g_hook		P_((const EXPR *, FLAGS));
static	ADDRESS *	g_index 	P_((const EXPR *));
static	ADDRESS *	g_logic 	P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_mul		P_((const EXPR *, FLAGS));
static	ADDRESS *	g_offset	P_((ADDRESS *, SIZE));
static	ADDRESS *	g_shift 	P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_unary 	P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	g_xmul		P_((const EXPR *, FLAGS, OPCODE));
static	ADDRESS *	mk_amode	P_((enum e_am));
static	ADDRESS *	mk_expr 	P_((enum e_am, EXPR *));
static	ADDRESS *	mk_direct	P_((EXPR *));
static	ADDRESS *	mk_high 	P_((ADDRESS *));
static	ADDRESS *	mk_indirect P_((REG, EXPR *));
static	ADDRESS *	mk_legal	P_((ADDRESS *, FLAGS, TYP *));
static	ADDRESS *	mk_low		P_((ADDRESS *));
static	ADDRESS *	mk_rmask	P_((REGMASK));
static	ADDRESS *	mk_smask	P_((REGMASK));
static	BOOL		g_compare	P_((const EXPR *));
static	BOOL		is_byte 	P_((const EXPR *));
static	BOOL		is_ushort	P_((const EXPR *));
static	BOOL		tst_short	P_((const EXPR *));
static	BOOL		tst_ushort	P_((const EXPR *));
static	SIZE		g_parms 	P_((const EXPR *));
static	SIZE		push_param	P_((const EXPR *));
static	void		call_library	P_((const CHAR *));
static	void		g_cbranch	P_((OPCODE, LABEL));
static	void		g_falsejp	P_((const EXPR *, LABEL));
static	void		g_immed 	P_((OPCODE, TYP *, IVAL, ADDRESS *));
static	void		g_rotate	P_((ADDRESS *, TYP *, int, TYP *, int));
static	void		g_test		P_((const EXPR *));
static	void		g_truejp	P_((const EXPR *, LABEL));
static	void		structassign	P_((ADDRESS *, ADDRESS *, SIZE, TYP *));

#ifdef FLOAT_SUPPORT
static	void		push_rtl_params P_((const EXPR *, const EXPR *));
#ifdef FLOAT_IEEE
static	ADDRESS *	func_result P_((FLAGS, SIZE, TYP*));
static	ADDRESS *	double_bin	P_((const EXPR *, FLAGS, const CHAR *));
static	ADDRESS *	double_as	P_((const EXPR *, FLAGS, const CHAR *));
#else
static	ADDRESS *	func_result P_((FLAGS, SIZE));
#endif /* FLOAT_MFFP */
#endif /* FLOAT_SUPPORT */

PRIVATE BOOL		g_isbigendian	P_((void));
PRIVATE void		g_auto_align	P_((void));
PRIVATE void		g_flush 	P_((SYM *));
PRIVATE void		g_allocate	P_((CSE *));
PRIVATE void		g_entry 	P_((SIZE));
PRIVATE void		g_epilogue	P_((void));
PRIVATE void		g_expression	P_((const EXPR *));
PRIVATE void		g_jfalse	P_((const EXPR *, LABEL));
PRIVATE void		g_jtrue 	P_((const EXPR *, LABEL));
PRIVATE void		g_return	P_((const EXPR *, TYP *));
PRIVATE void		g_stack 	P_((SIZE));
PRIVATE void		g_switch_compare	P_((const EXPR *, STMT *));
PRIVATE void		g_switch_table	P_((const EXPR *,struct swtab *,UVAL, UVAL));

#undef P_

static	int regs_used = 0;	/* number of register variable allocated */
static	REGMASK restore_mask;	/* register restore mask */
static	SIZE	max_stack_adjust = 0L;	/* largest amount stack is altered */
static	ADDRESS push = {am_adec, STACKPTR, 0, 0, NIL_EXPR};
static	ADDRESS pop =  {am_ainc, STACKPTR, 0, 0, NIL_EXPR};
static	REG frameptr = FRAMEPTR;

/*
 * this module contains all of the code generation routines for evaluating
 * expressions and conditions.
 */

PRIVATE SIZE
g_alignment P1(const TYP*, tp)
{
	SYM 	*sp;
	SIZE	 algn, al;

	switch (tp->type) {
	  case bt_void:
	return AL_ILLEGAL;
	  case bt_char:
	  case bt_charu:
	  case bt_uchar:
	  case bt_schar:
	return AL_CHAR;
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
	return AL_SHORT;
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	return AL_LONG;
	  case bt_float:
	return AL_FLOAT;
	  case bt_double:
	  case bt_longdouble:
	return AL_DOUBLE;
	  case bt_pointer32:
	return is_array_type(tp) ? g_alignment (referenced_type(tp)) : AL_POINTER;
	  case bt_struct:
	  case bt_union:
	if (align_option) {
		/* align structures to the alignment of the worst aligned member */
		algn = AL_CHAR;
		for (sp = members(tp)->symbols.head; sp != NIL_SYM; sp = sp->next) {
		al = g_alignment (sp->tp);
		if (al > algn)
			algn = al;
		}
		return algn;
	} else
		return AL_STRUCT;
	  case bt_func:
	return AL_FUNC;
	  case bt_bitfield:
	  case bt_ubitfield:
	return g_alignment (tp_int);
	  default:
	FATAL ((__FILE__,"g_alignment", "illegal type %d", tp->type));
	break;
	}
	return AL_ILLEGAL;
}

static ADDRESS *
mk_amode P1(enum e_am, mode)
{
	ADDRESS   *ap;
	ap = (ADDRESS *) xalloc((size_t) sizeof(ADDRESS));
	ap->mode = mode;
	return ap;
}

static ADDRESS *
mk_expr P2(enum e_am, mode, EXPR *, ep)
{
	ADDRESS *ap;
	ap = mk_amode (mode);
	ap->offset = ep;
	return ap;
}

/*
 * make a node to reference a line number.
 */
static ADDRESS	 *
mk_line P1(LINE, i)
{
	return mk_expr(am_line, mk_const((IVAL)i));
}

/*
 * copy an address mode structure.
 */
ADDRESS   *
copy_addr P2(ADDRESS*, ap, enum e_am, mode)
{
	ADDRESS *newap;
	if (ap == NIL_ADDRESS)
	FATAL ((__FILE__,"copy_addr","ap == 0"));
	newap = (ADDRESS *) xalloc((size_t) sizeof(ADDRESS));
	*newap = *ap;
	newap->mode = mode;
	return newap;
}

/*
 * make a node to reference an immediate value i.
 */
static ADDRESS	 *
mk_immed P1(IVAL, i)
{
	return mk_expr(am_immed, mk_const(i));
}

/*
 * construct a reference node for an internal label number.
 */
ADDRESS   *
mk_label P1(LABEL, lab)
{
	return mk_expr(am_direct, mk_lcon(lab));
}

/*
 * generate a direct reference to a string label.
 */
static ADDRESS	 *
mk_strlab P1(const CHAR*, s)
{
	EXPR	*ep;
	ep = mk_node(en_nacon, NIL_EXPR, NIL_EXPR, tp_void);
	ep->v.str = s;
	return mk_expr(am_direct, ep);
}

/*
 * generate a call to a library routine.
 * it is assumed that lib_name won''t be clobbered
 */
static void
call_library P1(const CHAR*, lib_name)
{
	SYM 	*sp;
	sp = sym_search(lib_name);
	if (sp == NIL_SYM) {
	sp = mk_sym(lib_name, sc_external, NIL_TYP);
	sym_append(&sp);
	}
	symbol_used(sp);
	g_code(op_jsr, IL0, mk_strlab(sp->name), NIL_ADDRESS);
}

/*
**	make an address reference to a register.
*/
ADDRESS   *
mk_reg P1(REG, r)
{
	ADDRESS   *ap;
	if (r < A0) {
	ap = mk_amode(am_dreg);
	} else {
	ap = mk_amode(am_areg);
	}
	ap->preg = r;
	return ap;
}

/*
**	make an address reference to 2 registers.
*/
ADDRESS *
mk_mreg P2(REG, r1, REG, r2)
{
	ADDRESS *ap;
	ap = mk_amode(am_mreg);
	ap->preg = r1;
	ap->sreg = r2;
	return ap;
}

/*
 * returns addressing mode of form offset(frameptr)
 * size is rounded up to AL_DEFAULT
 */
static ADDRESS	 *
mk_scratch P1(SIZE, size)
{
	ADDRESS   *ap;

	/* round up the request */
	if (size % AL_DEFAULT)
		size += AL_DEFAULT - (size % AL_DEFAULT);

	/* allocate the storage */
	act_scratch += size;

	/*
	 * The next statement could be deferred and put into the
	 * routine checkstack(), but this is just safer.
	 */
	if (act_scratch > max_scratch)
		max_scratch = act_scratch;

	ap = mk_indirect(frameptr, mk_const(-(lc_auto_max + act_scratch)));
	return ap;
}

/*
 * generate the mask address structure.
 */
static ADDRESS	 *
mk_rmask P1(REGMASK, mask)
{
	return mk_expr(am_rmask, mk_const((IVAL) mask));
}

/*
 * generate the mask address structure.
 */
static ADDRESS	 *
mk_smask P1(REGMASK, mask)
{
	return mk_expr(am_smask, mk_const((IVAL) mask));
}

/*
 * make a direct reference to a node.
 */
static ADDRESS	 *
mk_direct P1(EXPR*, ep)
{
	return mk_expr(am_direct, ep);
}

/*
 * make an indirect reference to a node
 */
static ADDRESS *
mk_indirect P2(REG, reg, EXPR*, ep)
{
	ADDRESS *ap;
	ap = mk_expr(am_indx, ep);
	ap->preg = reg;
	return ap;
}

/*
 * return true if the node passed can be generated as a short offset.
 */
BOOL
is_short P1(const EXPR*, ep)
{
	return ep->nodetype == en_icon &&
	(ep->v.i >= -32768L && ep->v.i <= 32767L);
}

/*
 * return true if the node passed can be generated as a short offset.
 */
static BOOL
is_ushort P1(const EXPR*, ep)
{
	return ep->nodetype == en_icon &&
	(ep->v.i >= 0L && ep->v.i <= 65535l);
}

static BOOL
is_byte P1(const EXPR*, ep)
{
	return ep->nodetype == en_icon &&
	(ep->v.i >= -128L && ep->v.i <= 127L);
}

/*
 * tests if node is a integer constant falling in the range of uns. short or
 * if node is cast from uns. short, uns. char or char.
 */
static BOOL
tst_ushort P1(const EXPR*, ep)
{
	if (is_ushort(ep))
	return TRUE;

	if (ep->nodetype == en_cast) {
	switch (ep->v.p[0]->etp->type) {
	  case bt_char:
	  case bt_charu:
	  case bt_uchar:
	  case bt_schar:
	  case bt_ushort:
	  case bt_uint16:
		return TRUE;
	  default:
		return FALSE;
	}
	}
	return FALSE;
}

/*
 * tests if node is a integer constant falling in the range of short or if
 * node is cast from short.
 */
static BOOL
tst_short P1(const EXPR*, ep)
{
	if (is_short(ep))
	return TRUE;

	if (ep->nodetype == en_cast) {
	switch (ep->v.p[0]->etp->type) {
	  case bt_char:
	  case bt_charu:
	  case bt_uchar:
	  case bt_schar:
	  case bt_short:
	  case bt_int16:
		return TRUE;
	  default:
		return FALSE;
	}
	}
	return FALSE;
}

static void
g_move8 P2(ADDRESS*, ap1, ADDRESS*, ap2)
{
	if (ap1->mode == am_mreg) {
	if (ap2->mode==am_mreg && ap2->preg==ap1->preg && ap2->sreg==ap1->sreg)
		return;
	if (ap1->preg < ap1->sreg && ap2->mode != am_mreg) {
		g_code(op_movem, IL4,
		 mk_smask((REGMASK)(1<<(int)ap1->preg | 1<<(int)ap1->sreg)), ap2);
		return;
	}
	}
	if (ap2->mode == am_mreg) {
	if (ap2->preg < ap2->sreg && ap1->mode != am_mreg) {
		g_code(op_movem, IL4, ap1,
		 mk_rmask((REGMASK)(1<<(int)ap2->preg | 1<<(int)ap2->sreg)));
		return;
	}
	}
	switch (ap2->mode) {
	  case am_adec:
	g_code (op_move, IL4, mk_high(ap1), mk_high(ap2));
	g_code (op_move, IL4,  mk_low(ap1),  mk_low(ap2));
	break;
	  default:
	g_code (op_move, IL4,  mk_low(ap1),  mk_low(ap2));
	g_code (op_move, IL4, mk_high(ap1), mk_high(ap2));
	break;
	}
}

/*
 * mk_legal will coerce the addressing mode in ap1 into a mode that is
 * satisfactory for the flag word.
 */
static ADDRESS	 *
mk_legal P3(ADDRESS*, ap, FLAGS, flags, TYP*, tp)
{
	ADDRESS *ap2;
	ILEN	 ilen = (ILEN)tp->size;

	if (flags & F_NOVALUE) {
		freeop(ap);
	return NIL_ADDRESS;
	}

	if (ap == NIL_ADDRESS)
	FATAL ((__FILE__, "mk_legal","ap = 0"));

	switch (ap->mode) {
	  case am_immed:
	if (flags & F_IMMED) {
		return ap;	/* mode ok */
	}
	break;
	  case am_areg:
	if (flags & F_AREG && (!(flags & F_VOL) || ap->preg <= max_areg))
		return ap;
	break;
	  case am_dreg:
	if (flags & F_DREG && (!(flags & F_VOL) || ap->preg <= max_dreg))
		return ap;
	break;
	  case am_mreg:
	if (flags & F_DREG) {
		if ((flags & F_VOL) && (ap->preg<=max_dreg || ap->sreg<=max_dreg))
		break;
		return ap;
	}
	break;
	  case am_ind:
	  case am_indx:
	  case am_indx2:
	  case am_indx3:
	  case am_indx4:
	  case am_direct:
	  case am_ainc:
	if (flags & F_MEM)
		return ap;
	break;
	  default:
	break;
	}
	if ((flags & (F_DREG | F_AREG)) == (F_DREG | F_AREG)) {
	/* decide, which mode is better */
	if (ap->mode == am_immed) {
		if (is_byte(ap->offset))
		flags = (FLAGS)(flags & F_DREG);
		else if (is_short(ap->offset) && ilen == IL4)
		flags = (FLAGS)(flags & F_AREG);
	}
	if (free_data() && (flags & F_DREG) && tp->type != bt_pointer32) {
		freeop(ap); 	/* maybe we can use it... */
		ap2 = data_register();	/* allocate to dreg */
		g_code(op_move, ilen, ap, ap2);
		return ap2;
	}
	if (free_addr() && (flags & F_AREG)) {
		freeop(ap);
		ap2 = address_register();
		g_code(op_move, ilen, ap, ap2);
		return ap2;
	}
	
	}
	if (flags & F_DREG) {
	freeop(ap); 	/* maybe we can use it... */
	switch (ilen) {
	  case IL1:
	  case IL2:
	  case IL4:
		ap2 = data_register();	/* allocate to dreg */
		g_code(op_move, ilen, ap, ap2);
		return ap2;
	  case IL8:
		ap2 = mdata_register();
		g_move8(ap, ap2);
		return ap2;
	  default:
		break;
	}
	}
	if (flags & F_AREG) {
	if (ilen < IL2)
		FATAL ((__FILE__,"mk_legal","illegal size %d --> An, mode=%d, flags=%d", ilen, ap->mode, flags));
	else {
		freeop(ap);
		ap2 = address_register();
		g_code(op_move, ilen, ap, ap2);
		return ap2;
	}
	}
	if (flags & F_MEM) {
	freeop(ap);
	ap2 = mk_scratch(tp->size);
	switch (tp->size) {
	  case 1L:
	  case 2L:
	  case 4L:
		g_code(op_move, ilen, ap, ap2);
		break;
	  case 8L:
		g_move8(ap, ap2);
		break;
	  default:
		CANNOT_REACH_HERE();
		break;
	}
	return ap2;
	}
	FATAL ((__FILE__,"mk_legal",""));
	return NIL_ADDRESS;
}

/*
 * If ap is an immediate value between -128 and 127 and the size of the
 * operation is 4 bytes then load the value into a data register if
 * there is one free.
 */
static ADDRESS *
mk_quick P2(ADDRESS*, ap, TYP*, tp)
{
	if (tp->size == (SIZE)2 && free_data() &&
	ap->mode == am_immed && is_byte(ap->offset)) {
	ap = mk_legal(ap, F_DREG, tp);
	}
	return ap;
}

/*
 * If ap is an immediate valye between -128 and 127, excluding the values
 * 1 to 8, and the size of the operation is 4 bytes then load the value into
 * a data register if there is one free.
 * This is used for "quick" operations to the "add" and "sub" commands
 * which aren't covered by the addq and subq instructions.
 * Don't "quick"en the value 0 as this is best left for the peephole
 * optimiser as it can then generally remove the instruction.
 */
static ADDRESS *
mk_quick2 P2(ADDRESS*, ap, TYP*, tp)
{
	if (ap->mode == am_immed && ap->offset->nodetype == en_icon &&
	(ap->offset->v.i < 0L || ap->offset->v.i > 8L)) {
	ap = mk_quick(ap, tp);
	}
	return ap;
}

/*
 * add a compiler generated label to the peep list.
 */
PRIVATE void
g_label P1(LABEL, labno)
{
	sync_stack();
	g_code(op_label, IL0, mk_label(labno), NIL_ADDRESS);
}

#ifdef DEBUGOPT
/*
 * add a source line number to the peep list.
 */
PRIVATE void
g_line P1(LINE, line)
{
	g_code(op_line, IL0, mk_line(line), NIL_ADDRESS);
}
#endif /*DEBUGOPT*/

/*
 * add a conditional branch instruction to the peep list.
 */
static void
g_cbranch P2(OPCODE, op, LABEL, labno)
{
	sync_stack();
	g_code(op, IL0, mk_label(labno), NIL_ADDRESS);
}

/*
 * add a branch instruction to the peep list.
 */
PRIVATE void
g_branch P1(LABEL, labno)
{
	g_cbranch(op_bra, labno);
}

/*
 * adjust the stack by "bytes" bytes.
 */
PRIVATE void
g_stack P1(SIZE, bytes)
{
	if (bytes != 0L) {
	/* adjust stack pointer */
	g_code(op_add, IL4, mk_immed(bytes), mk_reg(STACKPTR));
	stack_offset -= bytes;
	if (max_stack_adjust < bytes)
		max_stack_adjust = bytes;
	}
}

/*
 * Generate an instruction which takes an immediate option with optimal
 * (for space) instruction(s).
 */
static void
g_immed P4(OPCODE, op, TYP *, tp, IVAL, i, ADDRESS*, ap)
{
	ADDRESS *ap1 = mk_immed(i);
	ap1 = mk_quick(ap1, tp);
	g_code(op, (ILEN)tp->size, ap1, ap);
	freeop(ap1);
}

static void
g_immed2 P4(OPCODE, op, TYP*, tp, IVAL, i, ADDRESS*, ap)
{
	ADDRESS *ap1 = mk_immed(i);
	ap1 = mk_quick2(ap1, tp);
	g_code(op, (ILEN)tp->size, ap1, ap);
	freeop(ap1);
}

static ADDRESS	 *
g_extend  P3(ADDRESS*, ap, TYP*, tp1, TYP*, tp2)
{
	switch (tp2->type) {
	  case bt_int16:
	  case bt_uint16:
	  case bt_short:
	  case bt_ushort:
	switch (tp1->type) {
	  case bt_char:
	  case bt_schar:
		g_code(op_ext, IL2, ap, NIL_ADDRESS);
		break;
	  case bt_uchar:
	  case bt_charu:
		g_code(op_and, IL2, mk_immed(255l), ap);
	  default:
		break;
	}
	break;
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	switch (tp1->type) {
	  case bt_char:
	  case bt_schar:
		g_code(op_ext, IL2, ap, NIL_ADDRESS);
		/*FALLTHRU*/
	  case bt_int16:
	  case bt_short:
		g_code(op_ext, IL4, ap, NIL_ADDRESS);
		break;
	  case bt_uchar:
	  case bt_charu:
		g_code(op_and, IL4, mk_immed(255l), ap);
		break;
	  case bt_uint16:
	  case bt_ushort:
		g_code(op_and, IL4, mk_immed(65535l), ap);
		break;
	  default:
		break;
	}
	break;
	  default:
	CANNOT_REACH_HERE();
	break;
	}
	return ap;
}

static ADDRESS *
mk_low P1(ADDRESS*, ap)
{
	switch (ap->mode) {
	  case am_dreg:
	  case am_areg:
	  case am_ainc:
	  case am_adec:
	return ap;
	  case am_mreg:
	return mk_reg(ap->preg);
	  case am_ind:
	  case am_indx:
	  case am_indx2:
	  case am_direct:
	return g_offset(ap, 0L);
	  default:
	CANNOT_REACH_HERE();
	break;
	}
	return NIL_ADDRESS;
}

static ADDRESS *
mk_high P1(ADDRESS*, ap)
{
	switch (ap->mode) {
	  case am_ainc:
	  case am_adec:
	return ap;
	  case am_mreg:
	return mk_reg(ap->sreg);
	  case am_ind:
	  case am_indx:
	  case am_indx2:
	  case am_direct:
	return g_offset(ap, 4L);
	  default:
	CANNOT_REACH_HERE();
	break;
	}
	return NIL_ADDRESS;
}

/*
 * return ap, if ap can be switched to address a location with a short
 * offset, otherwise return 0.	  Typical application:
 * cast long -> short: 8(a6) --> 10(a6) offset is a small number (1,2 or 3)
 */
static ADDRESS *
g_offset P2(ADDRESS*, ap, SIZE, off)
{
	switch (ap->mode) {
	  case am_ind:
	ap = copy_addr(ap, am_indx);
	ap->offset = mk_const(off);
	return ap;
	  case am_indx:
	if (ap->offset->nodetype == en_icon && off + ap->offset->v.i <= (SIZE)32767) {
		ap = copy_addr(ap, ap->mode);
		ap->offset = mk_const (ap->offset->v.i + off);
		return ap;
	}
	break;
	  case am_indx2:
	  case am_indx3:
	  case am_indx4:
	if (ap->offset->nodetype == en_icon && off + ap->offset->v.i <= (SIZE)127) {
		ap = copy_addr(ap, ap->mode);
		ap->offset = mk_const (ap->offset->v.i + off);
		return ap;
	}
	break;
	  case am_direct:
	ap = copy_addr(ap, ap->mode);
	ap->offset = mk_add(ap->offset, mk_const(off));
	return ap;
	  default:
	break;
	}
	/* special value indicating that it must be done by hand */
	return NIL_ADDRESS;
}

/*
 * generate code to evaluate an index node and return the addressing
 * mode of the result.
 */
static ADDRESS	 *
g_index P1(const EXPR*, ep)
{
	ADDRESS *ap1, *ap2;
	EXPR *	ep0 = ep->v.p[0];
	EXPR *	ep1 = ep->v.p[1];
#if 1	/* test not yet validated */
	if (ep1->nodetype == en_register && ep1->v.r >= A0) {
		ep1 = ep->v.p[0];
		ep0 = ep->v.p[1];
	 }
#endif
	if (ep1->nodetype == en_register && ep0->nodetype == en_register) {
	if (ep0->v.r >= A0) {
		/* first node is address register */
		ap1 = g_expr(ep0, F_AREG);
		ap1 = copy_addr(ap1, am_none);
		ap2 = g_expr(ep1, (FLAGS)(F_AREG | F_DREG));
		if (ap2->mode == am_dreg) {
		/* 0(Ax,Dx) */
		ap1->mode = am_indx2;
		} else {
		/* 0(Ax,Ay) */
		ap1->mode = am_indx3;
		}
		ap1->sreg = ap2->preg;
		ap1->offset = mk_const(0l);
		return ap1;
	} else if (ep1->v.r >= A0) {
		/* second node is address register */
		ap1 = g_expr(ep1, F_AREG);
		ap1 = copy_addr(ap1, am_indx2); /* 0(Ax,Dx) */
		ap1->sreg = ep0->v.r;
		ap1->offset = mk_const(0l);
		return ap1;
	}
#if 0
	/*
	**	put address register first place
	**	data register second place
	*/
	if (ep0->v.r < A0 || ep1->v.r >= A0) {
		ep0 = ep->v.p[1];
		ep1 = ep->v.p[0];
	}
#endif
	}
	/*
	**	The general case (no register) 
	*/

	ap1 = g_expr(ep0, (FLAGS)(F_AREG | F_IMMED));
	if (ap1->mode == am_areg) {
	ap2 = g_expr(ep1, (FLAGS)(F_AREG | F_DREG | F_IMMED));
	validate(ap1);
	} else {
	ap2 = ap1;
	ap1 = g_expr(ep1, (FLAGS)(F_AREG | F_IMMED));
	validate(ap2);
	}
	/*
	**	possible combinations:
	** 
	**		F_AREG	+ F_AREG
	**		F_AREG	+ F_DREG
	**		F_AREG	+ F_IMMED
	**		F_IMMED + F_IMMED
	**		F_IMMED + F_AREG
	*/

	if (ap1->mode == am_areg) {
	/*
	**	watch out for:
	**		register(addr) + address_register
	**		register(addr) + data_register
	*/
	if (ap1->preg > max_areg) {
		/* ap1 = register variable address register */
		ap1 = copy_addr(ap1, ap1->mode);
		switch (ap2->mode) {
		  case am_dreg:
		/* 0(Ax,Dy) */
		ap1->mode = am_indx2;
		ap1->sreg = ap2->preg;
		ap1->deep = ap2->deep;
		ap1->offset = mk_const(0l);
		return ap1;
		  case am_areg:
		/* 0(Ax,Ay) */
		ap1->mode = am_indx3;
		ap1->sreg = ap2->preg;
		ap1->deep = ap2->deep;
		ap1->offset = mk_const(0l);
		return ap1;
		  case am_immed:
		if (!is_short(ap2->offset))
			/* we want to add to ap1 later... */
			ap1 = mk_legal(ap1, (FLAGS)(F_AREG | F_VOL), tp_pointer);
		break;
		  default:
		break;
		}
	}

	/*
	**	watch out for:
	**		address_register + register(data)
	*/
	if (ap2->mode == am_dreg && ap2->preg > max_dreg) {
		ap1 = copy_addr(ap1, am_indx2);
		ap1->sreg = ap2->preg;
		ap1->offset = mk_const(0l);
		return ap1;
	}
	}

	if (ap2->mode == am_immed) {
	if (ap1->mode == am_immed) {
		ap1 = copy_addr(ap1, am_direct);
		ap1->offset = mk_add(ap1->offset, ap2->offset);
		return ap1;
	}
	if (is_short(ap2->offset)) {
		ap1 = mk_legal(ap1, F_AREG, tp_pointer);
		ap1 = copy_addr(ap1, am_indx);
		ap1->offset = ap2->offset;
		return ap1;
	}
	}
	
	/* ap1 is volatile ... */
	g_code(op_add, IL4, ap2, ap1);/* add left to address reg */
	ap1 = copy_addr(ap1, am_ind);
	freeop(ap2);		/* release any temps in ap2 */
	return ap1; 		/* return indirect */
}

/*
 * return the addressing mode of a dereferenced node.
 */
static ADDRESS *
g_deref P3(const EXPR*, ep, TYP*, tp, FLAGS, flags)
{
	ADDRESS *ap1;
	SIZE	size = tp->size;
	/*
	 * If a reference to a struct/union is required, return a
	 * pointer to the struct instead
	 */
	if (is_structure(tp) || is_array_assignment (tp)) {
		return g_expr(ep, F_ALL);
	}
	switch (ep->nodetype) {
	  case en_add:
	return g_index(ep);
	  case en_autocon:
	if (ep->v.i >= (IVAL)-32768L && ep->v.i < (IVAL)32767L) {
		ap1 = mk_indirect(frameptr, mk_const(ep->v.i));
	} else {
		ap1 = address_register();
		g_code(op_move, IL4, mk_immed(ep->v.i), ap1);
		g_code(op_add, IL4, mk_reg(frameptr), ap1);
		ap1 = copy_addr(ap1, am_ind);
	}
	return ap1;
	  case en_ainc:
	/* special 68000 instructions */
	if ((size == (SIZE)1 || size == (SIZE)2 || size == (SIZE)4)
		&& ep->v.p[1]->v.i == size
		&& ep->v.p[0]->nodetype == en_register
		&& ep->v.p[0]->v.r >=A0 
		&& !(flags & F_USES)) {
		/* (An)+ */
		ap1 = mk_amode(am_ainc);
		ap1->preg = ep->v.p[0]->v.r;
		return ap1;
	}
	break;
	  default:
	break;
	}
	ap1 = g_expr(ep, (FLAGS)(F_AREG | F_IMMED));	/* generate address */
	if (ap1->mode == am_areg) {
	return copy_addr(ap1, am_ind);
	}
	return copy_addr(ap1, am_direct);
}

/*
 * get a bitfield value
 */
static ADDRESS *
g_fderef P2(const EXPR*, ep, FLAGS, flags)
{
	ADDRESS *ap;
	TYP 	*tp = ep->etp;

	ap = g_deref(ep->v.p[0], tp, F_ALL);
	ap = mk_legal(ap, (FLAGS)(F_DREG | F_VOL), tp);
	g_rotate(ap, tp, (int)ep->v.bit.offset, tp, (int)ep->v.bit.width);
	return mk_legal(ap, flags, tp);
}

/*============================================================================*/
#ifdef FLOAT_SUPPORT
static void
push_rtl_params  P2(const EXPR*, ep1, const EXPR*, ep2)
{
#ifdef FLOAT_IEEE
	ADDRESS *ap;
	is_parameter++;
	temp_inv();
	ap = g_expr(ep1, F_MEM);
	g_code(op_pea, IL0, ap, NIL_ADDRESS);
	freeop(ap);
	if (ep2) {
	ap = g_expr(ep2, F_MEM);
	g_code(op_pea, IL0, ap, NIL_ADDRESS);
	freeop(ap);
	}
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	is_parameter++;
	temp_inv();
	VOIDCAST push_param(ep1);
	if (ep2) {
	VOIDCAST push_param(ep2);
	}
#endif /* FLOAT_MFFP */
	is_parameter--;
}
#endif /* FLOAT_SUPPORT */

#ifdef FLOAT_IEEE
/*
 * generate code for floating-point binary nodes
 */
static ADDRESS *
double_bin P3(const EXPR*, ep, FLAGS, flags, const CHAR*, s)
{
	ADDRESS *ap;
	push_rtl_params (ep->v.p[1], ep->v.p[0]);
	ap = mk_scratch(ep->etp->size);
	g_code(op_pea, IL0, ap, NIL_ADDRESS);
	call_library(s);
	return mk_legal(ap, flags, ep->etp);
}

/*
 * generate code for floating-point assignment nodes
 */
static ADDRESS *
double_as P3(const EXPR*, ep, FLAGS, flags, const CHAR*, s)
{
	ADDRESS *ap;
	push_rtl_params (ep->v.p[1], ep->v.p[0]);
	call_library(s);
	if (flags & F_NOVALUE)
	return NIL_ADDRESS;
	/* Returns in d0 the pointer to the result value */
	ap = address_register();
	g_code(op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap);
	ap=copy_addr(ap, am_ind);
	return mk_legal(ap, flags, ep->etp);
}

#if 0
/*
 * generate code for the difficult assignment nodes when the LHS is an
 * integral type and the RHS is a floating-point type.
 */
static ADDRESS *
double_asnop P3(EXPR*, ep, FLAGS, flags, char*, s)
{
	ADDRESS *ap, *ap1, *ap2;
	EXPR	*lnode = ep->v.p[0];
	UVAL	 mask;
	ILEN	 ilen = (ILEN)ep->etp->size;

	is_parameter++;
	temp_inv();

	/* evaluate the LHS */
	switch (lnode->nodetype) {
	  case en_fieldref:
	mask = bitmask(lnode->v.bit.width);
	ap1 = g_expr(lnode->v.[0], F_AREG);
	g_code(op_move, 4, ap, &push);	/* save the address of LHS */
	ap1 = copy_addr(ap, am_indx);
	freeop(ap1);
	ap = data_register();
	g_code(op_move, ilen, ap1, ap);
	g_rotate (ap, ep->etp, lnode->bit_offset, lnode->etp, lnode->v.bit.width);
	g_immed(op_and, ep->etp, mask, ap);
	break;
	  default:
	ap = g_expr(lnode, F_AREG);
	g_code(op_move, 4, ap, &push);	/* save the address of LHS */
	break;
	}

	/* cast the LHS to the type of the RHS */
	ap = g_cast(ap, lnode->etp, ep->v.[1]->etp, F_MEM);

	/* evaluate the RHS */
	ap2 = g_expr(ep->v.p[1], F_MEM);

	/* and push the addresses onto the stack for the runtime routine */
	validate(ap);
	g_code(op_pea, IL0, ap2, NIL_ADDRESS);
	freeop(ap2);
	g_code(op_pea, IL0, ap, NIL_ADDRESS);
	g_code(op_pea, IL0, ap, NIL_ADDRESS);
	freeop(ap);
	is_parameter--;
	call_library(s);

	/* Get address of LHS */
	ap = address_register();
	g_code(op_move, 4, &pop, ap);
	ap = copy_addr(ap, am_ind);

	/* Returns in d0 the pointer to the result value */
	ap1 = address_register();
	g_code(op_move, 4, mk_reg(reg_usage->result->reg[0]), ap1);
	ap1=copy_addr(ap1, am_ind);

	/* cast the result of the operation back to the type of the LHS */
	ap1 = g_cast(ap1, ep->v.p[0]->etp, lnode->etp, F_DREG);

	/* and store the result */
	switch (lnode->nodetype) {
	  case en_fieldref:
	if (flags & F_NOVALUE) {
		ap2 = ap1;
	} else {
		/* result value needed */
		ap2 = data_register();
		g_code(op_move, 4, ap1, ap2);
	}
	g_rotate(ap1, ep->etp, -lnode->v.bit.offset, tp_void, 0);
	mask <<= lnode->v.bit.offset;
	g_immed(op_and, ep->etp, ~mask, ap);
	g_code(op_or, ilen, ap1, ap);
	if (!(flags & F_NOVALUE)) {
		freeop(ap2);
		validate(ap);
	}
	break;
	  default:
	g_code(op_move, ilen, ap1, ap);
	break;
	}
	freeop(ap1);
	return mk_legal(ap, flags, ep->etp);
}
#endif
#endif /* FLOAT_IEEE */

/*============================================================================*/

/*
 * generate code to evaluate a unary minus or complement. float: unary minus
 * calls a library function
 */
static ADDRESS *
g_unary P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	ADDRESS *ap;
	TYP 	*tp = ep->etp;
	switch (tp->type) {
	  case bt_uchar:
	  case bt_schar:
	  case bt_char:
	  case bt_charu:
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	ap = g_expr(ep->v.p[0], (FLAGS)(F_DREG | F_VOL));
	g_code(op, (ILEN) tp->size, ap, NIL_ADDRESS);
	return mk_legal(ap, flags, tp);
#ifdef FLOAT_SUPPORT
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	if (op == op_neg) {
#ifdef FLOAT_IEEE
		push_rtl_params (ep->v.p[0], NIL_EXPR);
			ap = mk_scratch(tp->size);
			g_code(op_pea, IL0, ap, NIL_ADDRESS);
		switch (tp->type) {
		  case bt_float:
		call_library(SUP_SFNEG);
		break;
		  case bt_double:
		call_library(SUP_DFNEG);
		break;
		  case bt_longdouble:
		call_library(SUP_LFNEG);
		break;
		  default:
		CANNOT_REACH_HERE();
		break;
		}
		return mk_legal(ap, flags, tp);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
		push_rtl_params (ep->v.p[0], NIL_EXPR);
		call_library(SUP_FPNEG);
		return func_result(flags, 0l);
#endif /* FLOAT_MFFP */
	}
		break;
#endif /* FLOAT_SUPPORT */
	  default:
	FATAL ((__FILE__,"g_unary","illegal type %d or operation %d",tp->type,op));
	break;
	}
	return NIL_ADDRESS;
}

/*============================================================================*/

/*
 * generate an auto increment or decrement node. op should be either op_add
 * (for increment) or op_sub (for decrement).
 */
static ADDRESS *
g_aincdec P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	ADDRESS *ap1, *ap2;
	TYP 	*tp = ep->etp;
#ifdef FLOAT_IEEE
	ADDRESS *ap3;
	CHAR	*s;
#endif /* FLOAT_IEEE */
	switch (tp->type) {
	  case bt_uchar:
	  case bt_schar:
	  case bt_char:
	  case bt_charu:
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	if (ep->v.p[0]->nodetype == en_fieldref)
		return g_asbitfield(ep, flags, op, TRUE);
	if (flags & F_NOVALUE) {/* dont need result */
		ap1 = g_expr(ep->v.p[0], F_ALL);
		g_code(op, (ILEN) tp->size, mk_immed(ep->v.p[1]->v.i), ap1);
	} else {
		ap1 = temp_reg(flags);
		ap2 = g_expr(ep->v.p[0], (FLAGS)(F_ALL | F_USES));
		validate(ap1);
		g_code(op_move, (ILEN) tp->size, ap2, ap1);
		g_code(op, (ILEN) tp->size, mk_immed(ep->v.p[1]->v.i), ap2);
		freeop(ap2);
	}
	return mk_legal(ap1, flags, tp);
#ifdef FLOAT_IEEE
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	if (flags & F_NOVALUE)
		return g_asadd (ep, flags, op);
	is_parameter++;
	temp_inv();

	/* 1st parameter */
	ap1 = g_expr(ep->v.p[1], F_MEM);
	g_code(op_pea, IL0, ap1, NIL_ADDRESS);
	freeop(ap1);

	/* 2nd parameter */
	ap3 = g_expr(ep->v.p[0], F_MEM);
	freeop(ap3);
	ap1 = mk_scratch(tp->size);
	ap2 = address_register();
	g_code(op_lea, IL0, ap3, ap2);
	g_code(op_move, IL4, ap2, &push);
	ap3 = address_register();
	validate(ap1);
	g_code(op_lea, IL0, ap1, ap3);
	structassign(ap2, ap3, tp->size, tp);
	freeop(ap3);
	freeop(ap2);

	is_parameter--;
	switch (tp->type) {
	  case bt_float:
		s = (op == op_add) ? SUP_ASSFADD : SUP_ASSFSUB;
		break;
	  case bt_double:
		s = (op == op_add) ? SUP_ASDFADD : SUP_ASDFSUB;
		break;
	  case bt_longdouble:
		s = (op == op_add) ? SUP_ASLFADD : SUP_ASLFSUB;
		break;
	  default:
		CANNOT_REACH_HERE();
		break;
	}
	call_library(s);
	return mk_legal(ap1, F_MEM, tp);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	if (op == op_add)
		return as_fcall(ep, flags, SUP_FPINC, SUP_ASFPINC);
	else
		return as_fcall(ep, flags, SUP_FPDEC, SUP_ASFPDEC);
#endif /* FLOAT_MFFP */
	  default:
	FATAL ((__FILE__,"g_aincdec","illegal type %d or float", tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*============================================================================*/

/*
 * generate code to evaluate a binary node and return the addressing mode of
 * the result.
 */
static ADDRESS *
g_add P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	ADDRESS *ap1, *ap2;
	TYP 	*tp = ep->etp;
	switch (tp->type) {
	  case bt_uchar:
	  case bt_schar:
	  case bt_char:
	  case bt_charu:
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_VOL | (flags & (F_DREG | F_AREG))));
	ap2 = g_expr(ep->v.p[1], F_ALL);
	validate(ap1);		/* in case push occurred */
	g_code(op, (ILEN) tp->size, ap2, ap1);
	freeop(ap2);
	return mk_legal(ap1, flags, tp);
#ifdef FLOAT_IEEE
	  case bt_float:
	return double_bin(ep, flags, op == op_add ? SUP_SFADD : SUP_SFSUB);
	  case bt_double:
	return double_bin(ep, flags, op == op_add ? SUP_DFADD : SUP_DFSUB);
	  case bt_longdouble:
	return double_bin(ep, flags, op == op_add ? SUP_LFADD : SUP_LFSUB);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_longdouble:
	  case bt_double:
	  case bt_float:
	push_rtl_params (ep->v.p[1], ep->v.p[0]);
	switch (op) {
	  case op_add:
		call_library(SUP_FPADD);
		break;
	  case op_sub:
		call_library(SUP_FPSUB);
		break;
	}
	return func_result(flags, 0l);
#endif /* FLOAT_MFFP */
	  default:
	FATAL ((__FILE__,"g_add","illegal type %d", tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*
 * generate a plus equal or a minus equal node.
 */
static ADDRESS *
g_asadd P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	FLAGS	 flagx;
	ADDRESS *ap1, *ap2;
	TYP 	*tp = ep->etp;
	switch (tp->type) {
	  case bt_char:
	  case bt_charu:
	  case bt_schar:
	  case bt_uchar:
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	if (ep->v.p[0]->nodetype == en_fieldref)
		return g_asbitfield(ep, flags, op, FALSE);
	if (flags & F_NOVALUE)
		flagx = F_ALL;
	else
		flagx = (FLAGS)(F_ALL | F_USES);
	ap1 = g_expr(ep->v.p[0], flagx);
	ap2 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));
	ap2 = mk_quick2(ap2, tp);
	validate(ap1);
	g_code(op, (ILEN) tp->size, ap2, ap1);
	freeop(ap2);
	return mk_legal(ap1, flags, tp);
#ifdef FLOAT_IEEE
	  case bt_float:
	if (op == op_add)
		return double_as(ep, flags, SUP_ASSFADD);
	else
		return double_as(ep, flags, SUP_ASSFSUB);
	  case bt_double:
	if (op == op_add)
		return double_as(ep, flags, SUP_ASDFADD);
	else
		return double_as(ep, flags, SUP_ASDFSUB);
	  case bt_longdouble:
	if (op == op_add)
		return double_as(ep, flags, SUP_ASLFADD);
	else
		return double_as(ep, flags, SUP_ASLFSUB);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_longdouble:
	  case bt_double:
	  case bt_float:
	if (op == op_add)
		return as_fcall(ep, flags, SUP_FPADD, SUP_ASFPADD);
	else
		return as_fcall(ep, flags, SUP_FPSUB, SUP_ASFPSUB);
#endif /* FLOAT_MFFP */
	  default:
	FATAL ((__FILE__,"asadd","illegal type %d",tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*============================================================================*/

/*
 * generate code to evaluate a restricted binary node and return the
 * addressing mode of the result.
 */
static ADDRESS *
g_logic P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	ADDRESS *ap1, *ap2;
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_VOL | F_DREG));
	if (op == op_eor)
	ap2 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));
	else
	ap2 = g_expr(ep->v.p[1], (FLAGS)(F_ALL & ~F_AREG));
	ap2 = mk_quick(ap2, ep->etp);
	validate(ap1);		/* in case push occurred */
	g_code(op, (ILEN) ep->etp->size, ap2, ap1);
	freeop(ap2);
	return mk_legal(ap1, flags, ep->etp);
}

/*
 * generate a &= or a |= ep.
 */
static ADDRESS *
g_aslogic P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	FLAGS	 flagx;
	ADDRESS *ap1, *ap2, *ap3;
	TYP 	*tp = ep->etp;
	if (ep->v.p[0]->nodetype == en_fieldref)
	return g_asbitfield(ep, flags, op, FALSE);
	if (flags & F_NOVALUE)
	flagx = F_ALL;
	else
	flagx = (FLAGS)(F_ALL | F_USES);
	ap1 = g_expr(ep->v.p[0], flagx);
	ap2 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));
	ap2 = mk_quick(ap2, ep->etp);
	validate(ap1);
	if (ap1->mode == am_areg) {
	ap3 = data_register();
	g_code(op_move, IL4, ap1, ap3);
	g_code(op, (ILEN) tp->size, ap2, ap3);
	g_code(op_move, (ILEN) tp->size, ap3, ap1);
	freeop(ap3);
	} else
	g_code(op, (ILEN) tp->size, ap2, ap1);
	freeop(ap2);
	return mk_legal(ap1, flags, tp);
}

/*============================================================================*/

/*
 * generate code to evaluate a shift node and return the address mode of the
 * result.
 */
static ADDRESS *
g_shift P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	ADDRESS *ap1, *ap2;
	TYP 	*tp = ep->etp;
  
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_DREG | F_VOL));
	ap2 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));

	/* quick constant only legal if 1<=const<=8 */

	if (ap2->mode == am_immed && ap2->offset->nodetype == en_icon
	&& (ap2->offset->v.i > 8L || ap2->offset->v.i < 1L)) {
	ap2 = mk_legal(ap2, F_DREG, tp_char);
	}
	validate(ap1);
	g_code(op, (ILEN) tp->size, ap2, ap1);
	freeop(ap2);
	return mk_legal(ap1, flags, tp);
}

/*
 * generate shift equals operators.
 */
static ADDRESS *
g_asshift P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	FLAGS	 flagx;
	ADDRESS *ap1, *ap2, *ap3;
	TYP 	*tp = ep->etp;
	switch (tp->type) {
	  case bt_uchar:
	  case bt_schar:
	  case bt_char:
	  case bt_charu:
	  case bt_ushort:
	  case bt_short:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_ulong:
	  case bt_long:
	if (ep->v.p[0]->nodetype == en_fieldref)
		return g_asbitfield(ep, flags, op, FALSE);
	if (flags & F_NOVALUE)
		flagx = F_ALL;
	else
		flagx = (FLAGS)(F_ALL | F_USES);
	ap1 = g_expr(ep->v.p[0], flagx);
	if (ap1->mode != am_dreg) {
#if 0
		/*
		 * assemblers cannot agree on the operands for this special
		 * version of the shift operator
		 */
		if (tp->size == 2L &&
		ep->v.p[1]->nodetype == en_icon &&
		ep->v.p[1]->v.i == 1) {
		g_code(op, 2, ap1, NIL_ADDRESS);
		return mk_legal(ap1, flags, tp);
		}
#endif
		ap3 = data_register();
		g_code(op_move, (ILEN) tp->size, ap1, ap3);
	} else
		ap3 = ap1;
	ap2 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));

	/* quick constant if 1<=const<=8 */

	if (ap2->mode == am_immed && ap2->offset->nodetype == en_icon
		&& (ap2->offset->v.i > 8L || ap2->offset->v.i < 1L)) {
		ap2 = mk_legal(ap2, F_DREG, tp_char);
	}
	validate(ap3);
	g_code(op, (ILEN) tp->size, ap2, ap3);
	freeop(ap2);
	if (ap3 != ap1) {
		g_code(op_move, (ILEN) tp->size, ap3, ap1);
		freeop(ap3);
	}
	return mk_legal(ap1, flags, tp);
	  default:
	FATAL ((__FILE__,"g_asshift","illegal type %d", tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*============================================================================*/


static ADDRESS *
g_shft P3(ADDRESS*, ap, IVAL, i, OPCODE, op)
{
	ADDRESS *ap1;
	if (i != 0L) {
	ap1 = mk_immed(i);
	if (i > 8L)
		ap1 = mk_legal(ap1, F_DREG, tp_char);
	validate(ap);
	g_code(op, IL4, ap1, ap);
	freeop(ap1);
	}
	return ap;
}

/*
 * Attempt to inline signed divide by constant which is a power of 2
 */
static ADDRESS *
g_div_icon P3(const EXPR*, ep, FLAGS, flags, BOOL, mod)
{
	ADDRESS *ap1, *ap2;
	int 	power;
	LABEL	label;
	IVAL	i = ep->v.p[1]->v.i;

	if ((power = pwrof2(i)) == -1)
	return NIL_ADDRESS;

	ap1 = g_expr(ep->v.p[0], flags);
	if (mod) {
	ap2 = data_register();
	g_code(op_move, IL4, ap1, ap2);
	validate(ap1);
	} else {
	ap2 = ap1;
	g_code(op_tst, IL4, ap1, NIL_ADDRESS);
	}
	label = nextlabel++;
	g_cbranch(op_bgt, label);
	g_immed2(op_add, ep->etp, i-(IVAL)1, ap2);
	g_label(label);

	if (mod) {
	g_immed (op_and, ep->etp, -i, ap2);
	g_code (op_sub, IL4, ap2, ap1);
	freeop(ap2);
	} else {
	ap1 = g_shft(ap2, (IVAL)power, op_asr);
	}
	return ap1;
}
/*
 * generate code to evaluate a divide operator
 */
static ADDRESS *
g_div P3(const EXPR*, ep, FLAGS, flags, BOOL, mod)
{
	ADDRESS *ap1, *ap2;
	OPCODE	 op = op_divu;
	TYP 	*tp = ep->etp;
	switch (tp->type) {
	  case bt_short:
	  case bt_int16:
	op = op_divs;
	/*FALLTHRU*/
	  case bt_ushort:
	  case bt_uint16:
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_DREG | F_VOL));
	ap2 = g_expr(ep->v.p[1], (FLAGS)(F_ALL & ~F_AREG));
	validate(ap1);
	ap1 = g_extend(ap1, tp, tp_long);
	g_code(op, IL0, ap2, ap1);
	if (mod)
		g_code (op_swap, IL0, ap1, NIL_ADDRESS);
	freeop(ap2);
	return mk_legal(ap1, flags, tp);
	  case bt_int32:
	  case bt_long:
	if (ep->v.p[1]->nodetype == en_icon) {
		ap1 = g_div_icon(ep, (FLAGS)(F_DREG | F_VOL), mod);
		if (ap1)
		return mk_legal(ap1, flags, tp);
	}
	/*FALLTHRU*/
	  case bt_uint32:
	  case bt_ulong:
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
#endif /* FLOAT_MFFP */
	temp_inv();
	is_parameter++;
	VOIDCAST push_param(ep->v.p[1]);
	VOIDCAST push_param(ep->v.p[0]);
	is_parameter--;
	switch (tp->type) {
	  case bt_int32:
	  case bt_long:
		call_library(mod ? SUP_LREM : SUP_LDIV);
		break;
	  case bt_uint32:
	  case bt_ulong:
		call_library(mod ? SUP_ULREM : SUP_ULDIV);
		break;
#ifdef FLOAT_MFFP
	  case bt_longdouble:
	  case bt_double:
	  case bt_float:
		call_library(mod ? SUP_FPREM : SUP_FPDIV);
		break;
#endif /* FLOAT_MFFP */
	  default:
		CANNOT_REACH_HERE();
		break;
	}
#ifdef FLOAT_IEEE
	return func_result(flags, 0l, tp_long);
	  case bt_float:
	return double_bin(ep, flags, SUP_SFDIV);
	  case bt_double:
	return double_bin(ep, flags, SUP_DFDIV);
	  case bt_longdouble:
	return double_bin(ep, flags, SUP_LFDIV);
#else
	return func_result(flags, 0l);
#endif /* FLOAT_IEEE */
	  default:
	FATAL ((__FILE__,"g_div","%d: illegal type %d", mod, tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*
 * generate /= and %= nodes.
 */
static ADDRESS *
g_asdiv P3(const EXPR*, ep, FLAGS, flags, BOOL, mod)
{
	ADDRESS *ap1, *ap2, *ap3;
	OPCODE	 op = op_divu;
	TYP 	*tp = ep->etp;
	switch (tp->type) {
	  case bt_char:
	  case bt_schar:
	op = op_divs;
	/*FALLTHRU*/
	  case bt_charu:
	  case bt_uchar:
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_ALL | F_USES));
	if (ap1->mode != am_dreg) {
		ap2 = data_register();
		g_code(op_move, IL1, ap1, ap2);
	} else
		ap2 = ap1;
	ap2 = g_extend(ap2, tp, tp_long);
	ap3 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));
	validate(ap2);
	freeop(ap3);
	if (ap3->mode == am_dreg)
		ap3 = g_extend(ap3, tp, tp_short);
	g_code(op_divu, IL0, ap3, ap2);
	if (mod)
		g_code(op_swap, IL0, ap2, NIL_ADDRESS);
	if (ap2 != ap1) {
		validate(ap1);
		freeop(ap2);
		g_code(op_move, IL1, ap2, ap1);
	}
	return mk_legal(ap1, flags, tp);
	  case bt_short:
	  case bt_int16:
	op = op_divs;
	/*FALLTHRU*/
	  case bt_ushort:
	  case bt_uint16:
	if (ep->v.p[0]->nodetype == en_fieldref)
		return g_asbitfield(ep, flags, op, FALSE);
	ap1 = data_register();
	ap2 = g_expr(ep->v.p[0], (FLAGS)(F_ALL | F_USES));
	validate(ap1);
	g_code(op_move, IL2, ap2, ap1);
	ap3 = g_expr(ep->v.p[1], (FLAGS)(F_ALL & ~F_AREG));
	validate(ap2);
	validate(ap1);
	ap1 = g_extend(ap1, tp, tp_long);
	g_code(op, IL0, ap3, ap1);
	if (mod)
		g_code(op_swap, IL0, ap1, NIL_ADDRESS);
	freeop(ap3);
	g_code(op_move, IL2, ap1, ap2);
	freeop(ap2);
	return mk_legal(ap1, flags, tp);
	  case bt_int32:
	  case bt_long:
	if (mod)
		return as_fcall(ep, flags, SUP_LREM, SUP_ASLREM);
	else
		return as_fcall(ep, flags, SUP_LDIV, SUP_ASLDIV);
	  case bt_uint32:
	  case bt_ulong:
	  case bt_pointer32:
	if (mod)
		return as_fcall(ep, flags, SUP_ULREM, SUP_ASULREM);
	else
		return as_fcall(ep, flags, SUP_ULDIV, SUP_ASULDIV);
#ifdef FLOAT_IEEE
	  case bt_float:
	return double_as(ep, flags, SUP_ASSFDIV);
	  case bt_double:
	return double_as(ep, flags, SUP_ASDFDIV);
	  case bt_longdouble:
	return double_as(ep, flags, SUP_ASLFDIV);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	if (mod)
		return as_fcall(ep, flags, SUP_FPREM, SUP_ASFPREM);
	else
		return as_fcall(ep, flags, SUP_FPDIV, SUP_ASFPDIV);
#endif /* FLOAT_MFFP */
	  default:
	FATAL ((__FILE__,"asdiv","%d: illegal type %d", mod, tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*============================================================================*/

/*
 * Attempt to inline multiple by constants which are "near" a power of 2
 */
static ADDRESS *
g_mul_icon P2(const EXPR*, ep, FLAGS, flags)
{
	OPCODE	op;
	IVAL	j, value, power, i = ep->v.p[1]->v.i;
	ADDRESS *ap1, *ap2, *ap3;
	BOOL	negate;
	if (i<0L) {
	i = -i;
	negate = TRUE;
	} else
	negate = FALSE;
	for (j = 0L, value = 1L; value != 0L && value < i ; j++, value <<= 1L) {
	if ((power = (IVAL)pwrof2(i-value)) != -1L)
		op = op_add;
	else if((power = (IVAL)pwrof2(i+value)) != -1L)
		op = op_sub;
	else
		continue;
	ap1 = g_expr(ep->v.p[0], flags);
	ap2 = data_register();
	if (ap1->mode == am_dreg) {
		validate(ap1);
		ap3 = ap1;
	} else {
		ap3 = data_register();
		validate(ap1);
		g_code(op_move, IL4, ap1, ap3);
	}
	g_code(op_move, IL4, ap3, ap2);
	ap3 = g_shft(ap3, power-j, op_asl);
	g_code(op, IL4, ap2, ap3);
	ap3 = g_shft(ap3, j, op_asl);
	if (negate)
		g_code(op_neg, IL4, ap3, NIL_ADDRESS);
	if (ap1 != ap3) {
		g_code(op_move, IL4, ap3, ap1);
		freeop(ap3);
	}
	freeop(ap2);
	return ap1;
	}
	return NIL_ADDRESS;
}

static ADDRESS *
g_mul P2(const EXPR*, ep, FLAGS, flags)
{
	ADDRESS *ap1, *ap2;
	OPCODE	 op = op_mulu;
	TYP 	*tp = ep->etp;
	EXPR *	ep0 = ep->v.p[0];
	EXPR *	ep1 = ep->v.p[1];

	switch (tp->type) {
	  case bt_short:
	  case bt_int16:
	op = op_muls;
	/*FALLTHRU*/
	  case bt_ushort:
	  case bt_uint16:
	if (ep0->nodetype == en_icon) {
		ep0 = ep->v.p[1];
		ep1 = ep->v.p[0];
	}
	ap1 = g_expr(ep0, (FLAGS)(F_DREG | F_VOL));
	ap2 = g_expr(ep1, (FLAGS)(F_ALL & ~F_AREG));
	validate(ap1);
	g_code(op, IL0, ap2, ap1);
	freeop(ap2);
	return mk_legal(ap1, flags, tp);
	  case bt_uint32:
	  case bt_ulong:
	  case bt_pointer32:
	  case bt_long:
	  case bt_int32:
	if (ep0->nodetype == en_icon) {
		ep0 = ep->v.p[1];
		ep1 = ep->v.p[0];
	}
	if (ep1->nodetype == en_icon) {
		ap1 = g_mul_icon(ep, (FLAGS)(F_DREG | F_VOL));
		if (ap1)
		return mk_legal(ap1, flags, tp);
	}
	/*FALLTHRU*/
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
#endif /* FLOAT_MFFP */
	temp_inv();
	is_parameter++;
	VOIDCAST push_param(ep1);
	VOIDCAST push_param(ep0);
	is_parameter--;
	switch (tp->type) {
	  case bt_int32:
	  case bt_long:
		call_library(SUP_LMUL);
		break;
	  case bt_uint32:
	  case bt_ulong:
	  case bt_pointer32:
		call_library(SUP_ULMUL);
		break;
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
		call_library(SUP_FPMULT);
#endif /* FLOAT_MFFP */
	  default:
		CANNOT_REACH_HERE();
		break;
	}
#ifdef FLOAT_IEEE
	return func_result(flags, 0l, tp_long);
	  case bt_float:
	return double_bin(ep, flags, SUP_SFMUL);
	  case bt_double:
	return double_bin(ep, flags, SUP_DFMUL);
	  case bt_longdouble:
	return double_bin(ep, flags, SUP_LFMUL);
#else
	return func_result(flags, 0l);
#endif /* FLOAT_IEEE */
	  default:
	FATAL ((__FILE__,"g_mul","illegal type %d",tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*
 * performs a mixed-mode multiplication
 */
static ADDRESS *
g_xmul P3(const EXPR*, ep, FLAGS, flags, OPCODE, op)
{
	ADDRESS *ap1, *ap2;

	ap1 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_VOL));
	ap2 = g_expr(ep->v.p[0], (FLAGS)(F_ALL & ~F_AREG));
	validate(ap1);

	g_code(op, IL0, ap2, ap1);
	freeop(ap2);
	return mk_legal(ap1, flags, ep->etp);
}
		
/*
 * generate a *= ep.
 */
static ADDRESS *
g_asmul P2(const EXPR*, ep, FLAGS, flags)
{
	ADDRESS *ap1, *ap2, *ap3;
	OPCODE	 op = op_mulu;
	TYP 	*tp = ep->etp;
	switch (tp->type) {
	  case bt_char:
	  case bt_schar:
	op = op_muls;
	/* FALLTHRU*/
	  case bt_charu:
	  case bt_uchar:
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_ALL | F_USES));
	if (ap1->mode != am_dreg) {
		ap2 = data_register();
		g_code(op_move, IL1, ap1, ap2);
	} else
		ap2 = ap1;
	ap2 = g_extend(ap2, tp, tp_short);
	ap3 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));
	validate(ap2);
	freeop(ap3);
	if (ap3->mode == am_dreg) {
		ap3 = g_extend(ap3, tp, tp_short);
	}
	g_code(op, IL0, ap3, ap2);
	if (ap2 != ap1) {
		validate(ap1);
		g_code(op_move, IL1, ap2, ap1);
		freeop(ap2);
	}
	return mk_legal(ap1, flags, tp);
	  case bt_short:
	  case bt_int16:
	op = op_muls;
	/*FALLTHRU*/
	  case bt_ushort:
	  case bt_uint16:
	if (ep->v.p[0]->nodetype == en_fieldref)
		return g_asbitfield(ep, flags, op, FALSE);
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_ALL | F_USES));
	ap2 = g_expr(ep->v.p[1], F_ALL);
	if (ap1->mode == am_dreg) {
		validate(ap1);
		g_code(op, IL0, ap2, ap1);
		freeop(ap2);
	} else {
		ap3 = data_register();
		validate(ap1);
		g_code(op_move, IL2, ap1, ap3);
		g_code(op, IL0, ap2, ap3);
		g_code(op_move, IL2, ap3, ap1);
		freeop(ap3);
		freeop(ap2);
	}
	return mk_legal(ap1, flags, tp);
	  case bt_int32:
	  case bt_long:
	if (ep->v.p[1]->nodetype == en_icon) {
		ap1 = g_mul_icon(ep, (FLAGS)(F_ALL | F_USES));
		if (ap1)
		return mk_legal(ap1, flags, tp);
	}
	return as_fcall(ep, flags, SUP_LMUL, SUP_ASLMUL);
	  case bt_uint32:
	  case bt_ulong:
	  case bt_pointer32:
	if (ep->v.p[1]->nodetype == en_icon) {
		ap1 = g_mul_icon(ep, (FLAGS)(F_ALL | F_USES));
		if (ap1)
		return mk_legal(ap1, flags, tp);
	}
	return as_fcall(ep, flags, SUP_ULMUL, SUP_ASULMUL);
#ifdef FLOAT_IEEE
	  case bt_float:
	return double_as(ep, flags, SUP_ASSFMUL);
	  case bt_double:
	return double_as(ep, flags, SUP_ASDFMUL);
	  case bt_longdouble:
	return double_as(ep, flags, SUP_ASLFMUL);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	return as_fcall(ep, flags, SUP_FPMULT, SUP_ASFPMULT);
#endif /* FLOAT_MFFP */
	  default:
	FATAL ((__FILE__,"asmul","illegal type %d",tp->type));
	break;
	}
	return NIL_ADDRESS;
}

/*============================================================================*/

/*
 * generate code to evaluate a condition operator node (?:)
 */
static ADDRESS *
g_hook P2(const EXPR*, ep, FLAGS, flags)
{
	ADDRESS *ap1, *ap2;
	FLAGS	 flagx;
	BOOL	 result_is_void = FALSE;
	TYP 	*tp = ep->etp;
	SIZE	 offset;
	LABEL	 false_label = nextlabel++;
	LABEL	 end_label = nextlabel++;

	switch (tp->type) {
#ifdef FLOAT_IEEE
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	flagx = F_MEM;
	break;
#endif /* FLOAT_IEEE */
	  case bt_void:
	result_is_void = TRUE;
	flagx = (FLAGS)(F_ALL | F_NOVALUE);
	break;
	  case bt_struct:
	  case bt_union:
	tp = tp_pointer;
	/* FALLTHRU*/
	  default:
		flagx = (FLAGS)(flags & (F_DREG | F_AREG)) == F_AREG ?
			(FLAGS)(F_AREG | F_VOL) : (FLAGS)(F_DREG | F_VOL);
	break;
	}

	temp_inv(); /* I do not think I can avoid that */
	offset = stack_offset;
	stack_offset = 0L;

	/* all scratch registers are void */

	g_falsejp(ep->v.p[0], false_label);
	ep = ep->v.p[1];

	/* all scratch registers are void */

	ap1 = g_expr(ep->v.p[0], flagx);
#ifdef FLOAT_IEEE
	if (flagx == F_MEM) {
	ADDRESS *ap;
	freeop(ap1);
	ap = address_register();
	g_code (op_lea, IL0, ap1, ap);
	ap1 = copy_addr(ap, am_ind);
	}
#endif /* FLOAT_IEEE */
	freeop(ap1);

	/* all scratch registers are void */

	g_branch(end_label);
	g_label(false_label);

	ap2 = g_expr(ep->v.p[1], flagx);
#ifdef FLOAT_IEEE
	if (flagx == F_MEM) {
	ADDRESS *ap;
	freeop(ap2);
	ap = address_register();
	g_code (op_lea, IL0, ap2, ap);
	ap2 = copy_addr(ap, am_ind);
	}
#endif /* FLOAT_IEEE */
	if (!result_is_void && !is_equal_address(ap1,ap2))
	   FATAL ((__FILE__,"g_hook","INCONSISTENCY"));

	g_label(end_label);

	g_stack(stack_offset);
	stack_offset = offset;
	return mk_legal(ap2, flags, tp);
}

/*
 * rotate a bitfield into the required position (assumes ap is in a register)
 */
static void
g_rotate P5(ADDRESS*, ap, TYP*, etp, int, offset, TYP*, tp, int, width)
{
	OPCODE	 op;
	int 	 w;
	ADDRESS *ap1;
	ILEN	 ilen = (ILEN)etp->size;

	switch (tp->type) {
	  case bt_int32:
	  case bt_int16:
		/* signed bitfield */
	w = (int)((ilen*8)-offset-width);
	if (w > 0) {
		if (w > 8) {
		ap1 = data_register();
		g_code(op_move, ilen, mk_immed((IVAL)w), ap1);
		} else
		ap1 = mk_immed((IVAL)w);
		g_code(op_rol, ilen, ap1, ap);
		freeop(ap1);
	}
	w = (int)((ilen*8)-width);
	if (w > 0) {
		if (w > 8) {
		ap1 = data_register();
		g_code(op_move, ilen, mk_immed((IVAL)w), ap1);
		} else
		ap1 = mk_immed((IVAL)w);
		g_code(op_asr, ilen, ap1, ap);
		freeop(ap1);
	}
		break;

	  default:
	/* offset is in range -31..31 */
	if (offset < 0)
		offset = (int)(ilen*8 + offset);

	/* offset in range 0..31 */
	if (ilen == IL4 && offset > 8 && offset < 25)
		g_code(op_swap, IL0, ap, NIL_ADDRESS);
	offset = (offset + 16) % 16;

	/* offset in range 0..15 */
	if (offset != 0) {
		if (offset > 8) {
		op = op_rol;
		offset = 16-offset;
		} else
		op = op_ror;
		g_code(op, ilen, mk_immed((IVAL) offset), ap);
	}

	if (tp->type != bt_void) {
		g_immed(op_and, etp, (IVAL)bitmask(width), ap);
	}
	}
}


/*
 * generate the code for assign operators in bitfield
 */
static ADDRESS *
g_asbitfield P4(const EXPR*, ep, FLAGS, flags, OPCODE, op, BOOL, swap)
{
	ADDRESS *ap1, *ap2, *ap3;
	EXPR	*ep1, *lnode = ep->v.p[0];
	int 	 width = (int)lnode->v.bit.width;
	int 	 offset = (int)lnode->v.bit.offset;
	ILEN	 ilen = (ILEN)ep->etp->size;
	UVAL	 mask;

	/* Evaluate the address of the LHS */
	ep1 = mk_ref(lnode->v.p[0], tp_pointer);
	ap2 = g_expr(ep1, F_MEM);

	/* Now get the value of the LHS, rotate and mask out unwanted bits */
	ap1 = data_register();
	g_code(op_move, ilen, ap2, ap1);
	g_rotate (ap1, ep->etp, (int)offset, lnode->etp, width);

	/* evaluate the RHS */
	ap3 = g_expr(ep->v.p[1], (FLAGS)(F_DREG | F_IMMED));
	validate(ap1);
	validate(ap2);

	/* now do the operation, masking the result back to the required size */
	switch (op) {
	  case op_divs:
	ap1 = g_extend(ap1, tp_short, tp_long);
	g_code(op, IL0, ap3, ap1);
	if (swap)
		g_code(op_swap, IL0, ap1, NIL_ADDRESS);
	swap = FALSE;
	break;
	  case op_divu:
	ap1 = g_extend(ap1, tp_ushort, tp_long);
	g_code(op, IL0, ap3, ap1);
	if (swap)
		g_code(op_swap, IL0, ap1, NIL_ADDRESS);
	swap = FALSE;
	break;
	  case op_muls:
	  case op_mulu:
	g_code(op, IL0, ap3, ap1);
	break;
	  default:
	g_code(op, ilen, ap3, ap1);
	break;
	}
	freeop(ap3);
	mask = (UVAL)bitmask(width);
	g_immed(op_and, ep->etp, (IVAL)mask, ap1);

	/* rotate result back into position, and store */
	g_rotate (ap1, ep->etp, -offset, tp_void, 0);
	g_immed(op_and, ep->etp, (IVAL)~(mask<<offset), ap2);
	g_code(op_or, ilen, ap1, ap2);
	freeop(ap1);
	freeop(ap2);

	/* return a result */
	ap2 = data_register();
	g_code(op_move, IL4, ap1, ap2);
	ap1 = ap2;

	if ((FLAGS)(flags & F_NOVALUE) == F_NONE) {
	/* result value needed */
	g_rotate (ap1, ep->etp, offset, tp_void, 0);
	if (swap) {
		/* post increment/decrement restore original value */
		switch (op) {
		  case op_add:
		op = op_sub;
		g_code (op, ilen, mk_immed(1l), ap1);
		g_immed(op_and, ep->etp, (IVAL)mask, ap1);
		break;
		  case op_sub:
		op = op_add;
		g_code (op, ilen, mk_immed(1l), ap1);
		g_immed(op_and, ep->etp, (IVAL)mask, ap1);
		break;
		  default:
		break;
		}
	}
	}
	return mk_legal(ap1, flags, ep->etp);
}


/*
 * assign structure from ap1 to ap2
 * ap1, ap2 are scratch address registers
 */
static void
structassign P4(ADDRESS*, ap1, ADDRESS*, ap2, SIZE, size, TYP*, tp)
{
	SIZE	 loop;
	int 	 rest;
	ADDRESS *ap3;
	LABEL	 label;
	SIZE	 i;
	SIZE	 algn = g_alignment(tp);
	ILEN	 ilen;

	ap1 = copy_addr(ap1, am_ainc);
	ap2 = copy_addr(ap2, am_ainc);
	if ((size % 2l == 0l) && (algn % 2l == 0l)) {
	ilen = IL4;
	} else {
	ilen = IL1;
	}
	loop = size / (SIZE) ilen;
	rest = (int) (size % (SIZE) ilen);
#ifdef MC68000
	if (loop <= 10l)		/* loop-unrolling only on 68000 */
#else
	if (loop <= 2) /* on 68010 and higher: unroll only the trivial case */
#endif /* MC68000 */
	for (i = 1l; i <= loop; i++)
		g_code(op_move, ilen, ap1, ap2);
	else {
	loop--; 		/* for dbra */
	ap3 = data_register();
	freeop(ap3);
	label = nextlabel++;
	if (loop <= 65535l) {   /* single loop */
		g_code(op_move, IL2, mk_immed(loop), ap3);
		g_label(label);
		g_code(op_move, ilen, ap1, ap2);
		g_code(op_dbra, IL0, ap3, mk_label(label));
	} else {        /* extended loop */
		g_code(op_move, IL4, mk_immed(loop), ap3);
		g_label(label);
		g_code(op_move, ilen, ap1, ap2);
		g_code(op_dbra, IL0, ap3, mk_label(label));
		g_code(op_sub, IL4, mk_immed(65536l), ap3);
		g_cbranch(op_bhs, label);
	}
	}
	if (rest >= 2) {
	rest -= 2;
	g_code(op_move, IL2, ap1, ap2);
	}
	if (rest >= 1)
	g_code(op_move, IL1, ap1, ap2);
}

/*
 * generate code for an assignment node.
 */
static ADDRESS *
g_assign P2(const EXPR*, ep, FLAGS, flags)
{
	ADDRESS *ap1, *ap2, *ap3;
	EXPR	*ep1;
	TYP 	*tp = ep->etp;
	ILEN	 ilen;
	UVAL	 mask;
	FLAGS	 flagx = flags & F_NOVALUE ? F_ALL : (F_ALL | F_USES);

	switch (tp->type) {
	  case bt_pointer32:
	if (!is_array_type(tp) && !is_array_assignment(tp))
		goto common;
	/*FALLTHRU*/
	  case bt_struct:
	  case bt_union:
	/*
	 * Other parts of this module return a pointer to a struct in a register,
	 * not the struct itself
	 */
	ap1 = g_expr(ep->v.p[1], (FLAGS)(F_AREG | F_VOL));
	ap2 = g_expr(ep->v.p[0], (FLAGS)(F_AREG | F_VOL));
	validate(ap1);

	/* hacky: save ap1 if needed later, structassign destroys it */
	if (flags & F_NOVALUE) {
		/* no need to save any registers */
		structassign(ap1, ap2, tp->size, tp);
		freeop(ap2);
		freeop(ap1);
		return NIL_ADDRESS;
	} else {
		ap3 = address_register();
		g_code(op_move, IL4, ap1, ap3);
		structassign(ap3, ap2, tp->size, tp);
		freeop(ap3);
		freeop(ap2);
		validate(ap1);
		return mk_legal(ap1, flags, tp_pointer);
	}
	  default:
common:
	ilen = (ILEN) tp->size;
	switch (ilen) {
	  case IL1:
	  case IL2:
	  case IL4:
	  case IL8:
		switch (ep->v.p[0]->nodetype) {
		  case en_fieldref:
		/*
		 * Field assignment
		 */
		/* get the value */
		mask = (UVAL)bitmask((int)ep->v.p[0]->v.bit.width);
		ap1 = g_expr(ep->v.p[1], (FLAGS)(F_IMMED | F_DREG | F_VOL));
		if (ap1->mode == am_immed) {
			ap1->offset->v.i &= (IVAL)mask;
			ap3 = mk_immed(ap1->offset->v.i << (int)ep->v.p[0]->v.bit.offset);
		} else {
			g_immed(op_and, ep->etp, (IVAL)mask, ap1);
			if (flags & F_NOVALUE) {
				ap3 = ap1;
			} else {
			/* result value needed */
			ap3 = data_register();
			g_code(op_move, IL4, ap1, ap3);
			}
			g_rotate(ap3, ep->etp, -(int)ep->v.p[0]->v.bit.offset,
								 tp_void, 0);
		}
		mask <<= ep->v.p[0]->v.bit.offset;
		ep1 = mk_ref(ep->v.p[0]->v.p[0], tp_pointer);
		ap2 = g_expr(ep1, F_MEM);
		validate(ap3);
		g_immed(op_and, ep->etp, (IVAL)~mask, ap2);
		g_code(op_or, ilen, ap3, ap2);
		freeop(ap2);
		if (!(flags & F_NOVALUE)) {
			freeop(ap3);
			validate(ap1);
		}
		break;

		/*
		 * we want to pass the right hand side as the expression value.
		 * This can''t be done if the left side is a register variable 
		 * on which the right hand side addressing mode depends. But if
		 * the left side IS a register variable, it is desirable to pass
		 * the left side, so no problem.
		 */
		  case en_register:
		/* pass the left side as expr. value */
		ap1 = g_expr(ep->v.p[0], flagx);
		ap2 = g_expr(ep->v.p[1], F_ALL);
		validate(ap1);
		g_code(op_move, ilen, ap2, ap1);
		freeop(ap2);
		break;
		  default:
		/* pass the right side as expr. value */
		/* normally, this is more efficient */
		ap1 = g_expr(ep->v.p[1], flagx);
		ap2 = g_expr(ep->v.p[0], F_ALL);
		validate(ap1);
		switch (ilen) {
		  case IL1:
		  case IL2:
		  case IL4:
			g_code(op_move, ilen, ap1, ap2);
			break;
		  case IL8:
			g_move8(ap1, ap2);
			break;
		  default:
			CANNOT_REACH_HERE();
			break;
		}
		freeop(ap2);
		break;
		}
		return mk_legal(ap1, flags, tp);
	  default:
		FATAL ((__FILE__,"g_assign", "size = %ld", tp->size));
		break;
	}
	return NIL_ADDRESS;
	}
}

/*
 * push the operand expression onto the stack. return the number of bytes
 * pushed
 */
static SIZE
push_param P1(const EXPR*, ep)
{
	ADDRESS *ap, *ap1;
	SIZE	 size = ep->etp->size;

	/* pushing of structures and unions */
	switch (ep->etp->type) {
	  case bt_struct:
	  case bt_union:
	/*
	 * It is possible that struct/union is an odd number of bytes if
	 * it is comprised only of chars - but we must ensure that the stack
	 * is always an even number of bytes.
	 */
	if (size & (SIZE)1)
		size++;

	if (is_lvalue(ep))
		ep = ep->v.p[0];
		/* all other cases return a pointer to the struct anyway */
	/* allocate stack space */
	g_code(op_add, IL4, mk_immed(-size), mk_reg(STACKPTR));
	/*
	 * F_VOL was missing in the following line --
	 * it took a hard-core debugging session to find this error
	 */
	ap = g_expr(ep, (FLAGS)(F_AREG | F_VOL));
	ap1 = address_register();
	validate(ap);
	g_code(op_move, IL4, mk_reg(STACKPTR), ap1);
	/* now, copy it on stack - the same as structassign */
	structassign(ap, ap1, size, ep->etp);
	freeop(ap1);
	freeop(ap);
	break;
	  default:
	ap = g_expr(ep, F_ALL);

	/*
	 * This is a hook for the peephole optimizer, which will convert lea
	 * <ea>,An + pea (An) ==> pea <ea>
	 */

	if (ap->mode == am_areg && size == (SIZE)4 && ap->preg <= max_areg) {
		ap = copy_addr(ap, am_ind);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
	} else {
		switch (size) {
		  case 1L:
		/*
		 * char parameters with prototype ... push an extra
		 * "junk" byte to keep the stack aligned.
		 */
		size++;
		ap = mk_legal(ap, F_DREG, ep->etp);
		/* FALLTHRU*/
		  case 2L:
		  case 4L:
		g_code(op_move, (ILEN) size, ap, &push);
		break;
		  case 8L:
		g_move8(ap, &push);
		break;
		  default:
		CANNOT_REACH_HERE();
		break;
		}
	}
	freeop(ap);
	break;
	}
	return size;
}

/*
 * push a list of parameters onto the stack and return the number of
 * bytes that they occupy on the stack.
 */
static SIZE
g_parms P1(const EXPR*, ep)
{
	SIZE	 size;
	is_parameter++;
	for (size = 0L; ep != NIL_EXPR; ep = ep->v.p[1]) {
	size += push_param(ep->v.p[0]);
	}
	is_parameter--;
	return size;
}


/*
**	Return value is the addressing mode of the result
**	This routine does not use mk_legal and takes care of the
**	stuff itself.
*/

#ifdef FLOAT_IEEE
static ADDRESS *
func_result P3(FLAGS, flags, SIZE, bytes, TYP*, tp)
{
	ADDRESS *ap;
#ifdef MOVEMBUG
	ADDRESS *ap1;
#endif /* MOVEMBUG */
	stack_offset += bytes;
	if (is_parameter)
	g_stack(bytes);
	if (flags & F_NOVALUE)
		return NIL_ADDRESS;
	switch (tp->type) {
	  case bt_longdouble:
	if (flags & F_DREG) {
		ap = mdata_register();
		g_move8(mk_mreg(reg_usage->result->reg[0], reg_usage->result->reg[1]), ap);
		return ap;
	} else if (flags & F_MEM) {
			ap = mk_scratch(tp->size);
#ifdef MOVEMBUG
		g_code(op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap);
		ap1 = copy_addr(ap, ap->mode);
		ap1->offset = mk_const(ap->offset->v.i + 4l);
		g_code(op_move, IL4, mk_reg(reg_usage->result->reg[1]), ap1);
		ap1 = copy_addr(ap, ap->mode);
		ap1->offset = mk_const(ap->offset->v.i + 8l);
		g_code(op_move, IL4, mk_reg(reg_usage->result->reg[2]), ap1);
#else
		g_code(op_movem, IL4,
		 mk_smask((REGMASK)(1<<(int)reg_usage->result->reg[0] |
					1<<(int)reg_usage->result->reg[1] |
					1<<(int)reg_usage->result->reg[2])), ap);
#endif /* MOVEMBUG */
		return ap;
	}
	break;
	  case bt_double:
	if (flags & F_DREG) {
		ap = mdata_register();
		g_move8(mk_mreg(reg_usage->result->reg[0], reg_usage->result->reg[1]), ap);
		return ap;
	} else if (flags & F_MEM) {
			ap = mk_scratch(tp->size);
#ifdef MOVEMBUG
		g_code(op_move, IL4, mk_reg(reg_usage->reg[0]), ap);
		ap1 = copy_addr(ap, ap->mode);
		ap1->offset = mk_const(ap->offset->v.i + 4l);
		g_code(op_move, IL4, mk_reg(reg_usage->reg[1]), ap1);
#else
		g_code(op_movem, IL4,
		 mk_smask((REGMASK)(1<<(int)reg_usage->result->reg[0] |
					1<<(int)reg_usage->result->reg[1])), ap);
#endif /* MOVEMBUG */
		return ap;
	}
	break;
	  case bt_float:
	if (flags & F_DREG) {
		ap = data_register();
		g_code (op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap);
		return ap;
	} else if (flags & F_MEM) {
			ap = mk_scratch(tp->size);
		g_code(op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap);
		return ap;
	}
	break;
	  case bt_char:
	  case bt_uchar:
	  case bt_charu:
	  case bt_schar:
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer16:
	  case bt_pointer32:
	  case bt_struct:		/* always returned as a pointer to result */
	  case bt_union:		/* always returned as a pointer to result */
	ap = temp_reg(flags);
	g_code(op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap);
	return ap;
	  default:
	FATAL ((__FILE__,"func_result","illegal mode (%d, %d)", tp->type, flags));
	break;
	}
	return NIL_ADDRESS;
}
#else
static ADDRESS *
func_result P2(FLAGS, flags, SIZE, bytes)
{
	/*
	 * saves a function call result in D0 it is assumed that flags contain
	 * either F_DREG or F_AREG return value is the addressing mode of the
	 * result bytes is the number of bytes to pop off the stack
	 *
	 * This routine does not use mk_legal and takes care of the stuff itself.
	 */
	ADDRESS *ap;
	stack_offset += bytes;
	if (is_parameter)
	g_stack(bytes);
	if (flags & F_NOVALUE)
		return NIL_ADDRESS;
	if (flags & F_DREG) {
	ap = data_register();
	g_code(op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap);
	} else if (flags & F_AREG) {
	ap = address_register();
	g_code(op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap);
	} else {
	FATAL ((__FILE__,"func_result","illegal addressing mode"));
	}
	return ap; 
}
#endif /* FLOAT_IEEE */

/*
 * assignment operations with library calls
 */
static ADDRESS *
as_fcall P4(const EXPR*, ep, FLAGS, flags, const CHAR*, libnam1, const CHAR*, libnam2)
{
	/*
	 * example: libnam1 = ".ldiv", libnam2 = ".asldiv"
	 */
	ADDRESS *ap1, *ap2, *ap3, *ap4;
	EXPR	*lnode;
	UVAL	 mask;
	int 	 width, offset;
	ILEN	 ilen = (ILEN)ep->etp->size;

	temp_inv();
	switch (ep->v.p[0]->nodetype) {
	  case en_register:
	is_parameter++;
	VOIDCAST push_param(ep->v.p[1]);
	/* ap1 cannot be destroyed, no problem */
	ap1 = g_expr(ep->v.p[0], (FLAGS)(F_DREG | F_AREG));
	g_code(op_move, ilen, ap1, &push);
	is_parameter--;
	call_library(libnam1);
	/* ap1 is always valid and not equal to reg[0] */
	g_code(op_move, ilen, mk_reg(reg_usage->result->reg[0]), ap1);
	break;
	  case en_fieldref:
	is_parameter++;
	lnode = ep->v.p[0];
	width = (int)lnode->v.bit.width;
	offset = (int)lnode->v.bit.offset;

	/* evaluate the address of the LHS */
	ap3 = g_deref(lnode->v.p[0], lnode->etp, F_MEM);
	freeop(ap3);
	ap2 = address_register();
	g_code(op_lea, IL0, ap3, ap2);

	/* and save this address on the stack so that it can be used later */
	g_code(op_move, ilen, ap2, &push);
	ap2 = copy_addr(ap2, am_ind);
	freeop(ap2);

	/* Now get the value of the LHS */
	ap1 = data_register();
	g_code(op_move, ilen, ap2, ap1);

	/* evaluate the RHS and push it onto the stack */
	ap2 = g_expr(ep->v.p[1], F_ALL);
	validate(ap1);
	if (ap2->mode == am_areg && ilen == IL4 && ap2->preg <= max_areg) {
		ap2 = copy_addr(ap2, am_ind);
		g_code(op_pea, IL0, ap2, NIL_ADDRESS);
	} else
		g_code(op_move, ilen, ap2, &push);
	freeop(ap2);

	/* rotate to position, mask out unwanted bits */
	g_rotate (ap1, ep->etp, offset, lnode->etp, width);

	/* push the LHS onto the stack */
	g_code(op_move, ilen, ap1, &push);
	is_parameter--;

	/* now do the operation, masking the result back to the required size */
	call_library(libnam1);

	g_code(op_move, ilen, mk_reg(reg_usage->result->reg[0]), ap1);
	mask = bitmask(width);
	g_immed(op_and, ep->etp, (IVAL)mask, ap1);
	if (flags & F_NOVALUE) {
		ap4 = ap1;
	} else {
		/* result value needed */
		ap4 = data_register();
		g_code(op_move, IL4, ap1, ap4);
	}

	/* restore ap2 */
	ap2 = address_register();
	g_code(op_move, ilen, &pop, ap2);
	ap2 = copy_addr(ap2, am_ind);

	/* rotate result back into position, and store */
	mask <<= offset;
	g_rotate (ap4, ep->etp, -offset, tp_void, 0);
	g_immed(op_and, ep->etp, (IVAL)~mask, ap2);
	g_code(op_or, ilen, ap4, ap2);
	freeop(ap2);
	if (!(flags & F_NOVALUE)) {
		freeop(ap4);
		validate(ap1);
	}
	return mk_legal(ap1, flags, ep->etp);
	  default:
	/* pass a pointer to the left side, call special function */
	is_parameter++;
	VOIDCAST push_param(ep->v.p[1]);
	VOIDCAST push_param(ep->v.p[0]->v.p[0]);
	is_parameter--;
	call_library(libnam2);
	break;
	}
	if (flags & F_NOVALUE)
	return NIL_ADDRESS;

	ap1 = temp_reg(flags);
	g_code(op_move, IL4, mk_reg(reg_usage->result->reg[0]), ap1);
	return mk_legal(ap1, flags, ep->etp);
}

/*
 * generate a function call node and return the address mode of the result.
 */
static ADDRESS *
g_fcall P2(const EXPR*, ep, FLAGS, flags)
{
	EXPR	*ep0 = ep->v.p[0];
	TYP 	*tp;
	ADDRESS *ap;
	SIZE	 size;
	OPCODE	 op;

	if (!is_parameter && ep->nodetype != en_call) {
	switch (stackopt_option) {

	  default:
	  case 0:
		/*
		**	no stack optimisation
		*/
		g_stack(stack_offset);
		break;

	  case 1:
		/*
		**	"Safe" stack optimisation.	Perform a stack optimisation
		**	unless:
		**	1.	The function call is via a variable
		**	2.	The function starts with an underscore character
		**	3.	The alloca() routine is called
		*/
		if ((ep0->nodetype != en_nacon) ||
		(ep0->v.str[0] == (CHAR)'_') ||
		(ep0->v.str == alloca_name)) {
		g_stack(stack_offset);
		}
		break ;

	  case 2:
		/*
		**	"Forced" stack optimisation.   This will not suppress
		**	the optimisation on encountering calls to functions
		**	whose names begin with underscore.
		*/
		if ((ep0->nodetype != en_nacon) ||
		(ep0->v.str == alloca_name)) {
		g_stack(stack_offset);
		}
		break ;

	  case 3:
		/*
		**	"Forced" stack optimisation.   This will not suppress
		**	the optimisation on encountering calls to functions
		**	whose names begin with underscore or via a function
		**	variable.
		*/
		if ((ep0->nodetype == en_nacon) &&
		(ep0->v.str == alloca_name)) {
		g_stack(stack_offset);
		}
		break ;
	}
	}

	/* push any used addr&data temps */
	temp_inv();
	size = g_parms(ep->v.p[1]); /* generate parameters */
	tp = ep->etp;
	if (is_structure(tp)) {
	/*
	**	For functions returning a structure or a union, push
	**	a pointer to the return value as additional argument.
	**	The scratch space will be allocated in the stack frame
	**	of the calling function.
	*/
	ap = mk_scratch(tp->size);
	g_code(op_pea, IL0, ap, NIL_ADDRESS);
	size += tp_pointer->size;
	freeop(ap);
	}
	if (ep->nodetype == en_call)
	size = 0L;

	/* call the function */
	switch (ep0->nodetype) {
	  case en_nacon:
	op = (OPCODE)(large_option ? op_jsr : op_bsr);
	ap = mk_direct(ep0);
	break;
	  case en_labcon:
	op = op_jsr;
	ap = mk_direct(ep0);
	break;
	  default:
	op = op_jsr;
	ap = g_expr(ep0, F_AREG);
	ap = copy_addr(ap, am_ind);
	freeop(ap);
	break;
	}
	g_code(op, IL0, ap, NIL_ADDRESS);

#ifdef FLOAT_IEEE
	return func_result(flags, size, tp);
#else
	return func_result(flags, size);
#endif /* FLOAT_IEEE */
}

/*
 * generates code for a en_cast node
 */
static ADDRESS *
g_cast P4(ADDRESS*, ap, TYP*, tp1, TYP*, tp2, FLAGS, flags)
{
	ADDRESS *ap1;
	FLAGS	 flagx;
#ifdef FLOAT_IEEE
	BOOL	 f;
	CHAR	*s;
#endif /* FLOAT_IEEE */

	if (flags & F_NOVALUE) {
		freeop(ap);
	return NIL_ADDRESS;
	}

	if (tp1->type == tp2->type) {
	/*
	**	This can happen with the g_xmul stuff, where a cast from
	**	(u)short to long now casts from (u)short to (u)short for
	**	a 68000 mulu or muls instruction.
	**	It is safe to cut things short then.
	**	It should not happen with types other than (u)short, but
	**	it does not harm either.
	*/
	switch (tp1->type) {
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
		return mk_legal(ap, flags, tp1);
	  default:
		FATAL ((__FILE__,"g_cast","tp1==tp2 (%d)", tp1->type));
		break;
	}
	}

	switch (tp2->type) {
	/* switch: type to cast to */
	  case bt_char:
	  case bt_charu:
	  case bt_schar:
	  case bt_uchar:
	switch (tp1->type) {
	  case bt_uchar:
	  case bt_schar:
	  case bt_char:
	  case bt_charu:
		return mk_legal(ap, flags, tp2);
	  case bt_ushort:
	  case bt_short:
	  case bt_int16:
	  case bt_uint16:
		if ((ap1 = g_offset(ap, 1l)) == NIL_ADDRESS)
		ap1 = mk_legal(ap, F_DREG, tp1);
		return mk_legal(ap1, flags, tp2);
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
		if ((ap1 = g_offset(ap, 3l)) == NIL_ADDRESS)
		ap1 = mk_legal(ap, F_DREG, tp1);
		return mk_legal(ap1, flags, tp2);
	  case bt_float:
		  case bt_double:
	  case bt_longdouble:
		ap = g_cast(ap, tp1, tp_long, F_DREG);
		return mk_legal(ap, flags, tp2);
	  default:
		break;
	}
		break;
	  case bt_ushort:
	  case bt_short:
	  case bt_int16:
	  case bt_uint16:
	switch (tp1->type) {
	  case bt_charu:
	  case bt_uchar:
	  case bt_char:
	  case bt_schar:
		ap = mk_legal(ap, (FLAGS)(F_DREG | F_VOL), tp1);
		ap = g_extend(ap, tp1, tp2);
		return mk_legal(ap, flags, tp2);
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
		return mk_legal(ap, flags, tp2);
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
		if ((ap1 = g_offset(ap, 2l)) == NIL_ADDRESS)
		ap1 = mk_legal(ap, F_DREG, tp1);
		return mk_legal(ap1, flags, tp2);
	  case bt_float:
		  case bt_double:
	  case bt_longdouble:
		ap = g_cast(ap, tp1, tp_long, F_DREG);
		return mk_legal(ap, flags, tp2);
	  default:
		break;
	}
		break;
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	switch (tp1->type) {
	  case bt_charu:
	  case bt_uchar:
	  case bt_char:
	  case bt_schar:
	  case bt_ushort:
	  case bt_uint16:
		ap = mk_legal(ap, (FLAGS)(F_DREG | F_VOL), tp1);
		ap = g_extend(ap, tp1, tp2);
		return mk_legal(ap, flags, tp2);
	  case bt_short:
	  case bt_int16:
			flagx = (FLAGS)(flags & (F_DREG | F_AREG));
			if (flagx == F_NONE)
		flagx = (FLAGS)(F_DREG | F_AREG);
		ap = mk_legal(ap, flagx, tp1);
		if (ap->mode == am_dreg)
		ap = g_extend(ap, tp1, tp2);
		return mk_legal(ap, flags, tp2);
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	  case bt_func:
		return mk_legal(ap, flags, tp2);
#ifdef FLOAT_IEEE
	  case bt_float:
		  case bt_double:
	  case bt_longdouble:
		ap = mk_legal(ap, F_MEM, tp1);
		freeop(ap);
		temp_inv();
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		f = (tp2->type == bt_long || tp2->type == bt_int32);
		switch(tp1->type) {
		  case bt_float:
		call_library(f ? SUP_SFTOL : SUP_SFTOUL);
		break;
		  case bt_double:
		call_library(f ? SUP_DFTOL : SUP_DFTOUL);
		break;
		  case bt_longdouble:
		call_library(f ? SUP_LFTOL : SUP_LFTOUL);
		break;
		  default:
		CANNOT_REACH_HERE();
		break;
		}
		return func_result(flags, 0l, tp2);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_float:
		  case bt_double:
	  case bt_longdouble:
		/* library call */
		freeop(ap);
		temp_inv();
		g_code(op_move, 4, ap, &push);
		call_library(tp2->type == bt_long || tp2->type == bt_int32 ? SUP_FPFTOL : SUP_FPFTOU);
		return func_result(flags, 0l);
#endif /* FLOAT_MFFP */
	  default:
		break;
	}
		break;
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	switch (tp1->type) {
	  case bt_char:
	  case bt_charu:
	  case bt_schar:
	  case bt_uchar:
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
		ap = g_cast(ap, tp1, tp_long, F_ALL);
		/*FALLTHRU*/
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
		/* library call */
		freeop(ap);
		temp_inv();
		g_code(op_move, IL4, ap, &push);
#ifdef FLOAT_IEEE
		ap = mk_scratch(tp2->size);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		f = (tp1->type == bt_ulong || tp1->type == bt_uint32 || tp1->type == bt_pointer32);
		switch (tp2->type) {
		  case bt_float:
		call_library(f ? SUP_SFUTOSF : SUP_SFLTOSF);
		break;
		  case bt_double:
		call_library(f ? SUP_DFUTODF : SUP_DFLTODF);
		break;
		  case bt_longdouble:
		call_library(f ? SUP_LFUTOLF : SUP_LFLTOLF);
		break;
		  default:
		CANNOT_REACH_HERE();
		break;
		}
		return mk_legal(ap, flags, tp2);
	  case bt_float:
		/* cast from float to double/long double */
		switch(tp2->type) {
		  case bt_double:
		s = SUP_SFTODF;
		break;
		  case bt_longdouble:
		s = SUP_SFTOLF;
		break;
		  default:
		CANNOT_REACH_HERE();
		break;
		}
		ap = mk_legal(ap, F_MEM, tp1);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		freeop(ap);
		ap = mk_scratch(tp2->size);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		call_library(s);
		return mk_legal(ap, flags, tp2);
		  case bt_double:
		/* cast from double to float/long double */
		switch(tp2->type) {
		  case bt_float:
		s = SUP_DFTOSF;
		break;
		  case bt_longdouble:
		s = SUP_DFTOLF;
		break;
		  default:
		CANNOT_REACH_HERE();
		break;
		}
		ap = mk_legal(ap, F_MEM, tp1);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		freeop(ap);
		ap = mk_scratch(tp2->size);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		call_library(s);
		return mk_legal(ap, flags, tp2);
	  case bt_longdouble:
		/* cast from long double to float/double */
		switch(tp2->type) {
		  case bt_float:
		s = SUP_LFTOSF;
		break;
		  case bt_double:
		s = SUP_LFTODF;
		break;
		  default:
		CANNOT_REACH_HERE();
		break;
		}
		ap = mk_legal(ap, F_MEM, tp1);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		freeop(ap);
		ap = mk_scratch(tp2->size);
		g_code(op_pea, IL0, ap, NIL_ADDRESS);
		call_library(s);
		return mk_legal(ap, flags, tp2);
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
		if (tp1->type == bt_uint32 || tp1->type == bt_ulong || tp1->type == bt_pointer32)
		call_library(SUP_FPUTOF);
		else
		call_library(SUP_FPLTOF);
		return func_result(flags, 0l);
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
		return mk_legal(ap, flags, tp2);
#endif /* FLOAT_MFFP */
	  default:
		break;
	}
		break;
	  default:
	break;
	}
	FATAL ((__FILE__,"g_cast","illegal combination type1=%d, type2=%d", tp1->type, tp2->type));
	return NIL_ADDRESS;
}

#ifdef ASM
static ADDRESS *
g_asm P1(const EXPR*, ep)
{
	ADDRESS *ap = mk_expr(am_asm, copynode(ep));
	g_code(op_asm, IL0, ap, NIL_ADDRESS);
	return NIL_ADDRESS;
}
#endif /* ASM */

/*
 * general expression evaluation. returns the addressing mode of the result.
 */
static ADDRESS *
g_expr P2(const EXPR*, ep, FLAGS, flags)
{
	ADDRESS *ap1, *ap2;
	LABEL	 lab0, lab1;
	BTYPE	 type;
	OPCODE	 op;
	TYP 	*tp = ep->etp;
	if (ep == NIL_EXPR) {
	FATAL ((__FILE__,"g_expr","ep == 0"));
	}
	if (tst_const(ep)) {
	ap1 = mk_expr(am_immed, copynode(ep));
	return mk_legal(ap1, flags, tp);
	}
	type = tp->type;
	switch (ep->nodetype) {
	  case en_autocon:
	ap1 = address_register();
	if ((ep->v.i >= (IVAL) -32768L) && (ep->v.i <= (IVAL) 32767L)) {
		ap2 = mk_indirect(frameptr, copynode(ep));
		g_code(op_lea, IL0, ap2, ap1);
	} else {
		g_code(op_move, IL4, mk_immed(ep->v.i), ap1);
		g_code(op_add, IL4, mk_reg(frameptr), ap1);
		ap1 = copy_addr(ap1, am_ind);
	}
	return mk_legal(ap1, flags, tp);
	  case en_register:
	ap1 = mk_reg(ep->v.r);
	return mk_legal(ap1, flags, tp);
	  case en_ref:
	/*
	 * g_deref uses flags and size only to test F_USES
	 *
	 * If the result is not used, autoincrement addressing
	 * modes are wrong!
	 */
	if (flags & F_NOVALUE) {
			ap1 = g_deref(ep->v.p[0], tp, (FLAGS)(flags | F_USES));
	} else {
			ap1 = g_deref(ep->v.p[0], tp, flags);
	}
		if (is_structure(tp) || is_array_type(tp)) {
			return mk_legal(ap1, flags, tp_pointer);
		} else {
		return mk_legal(ap1, flags, tp);
	}
	  case en_fieldref:
	return g_fderef(ep, flags);
	  case en_uminus:
	return g_unary(ep, flags, op_neg);
	  case en_compl:
	return g_unary(ep, flags, op_not);
	  case en_add:
	return g_add(ep, flags, op_add);
	  case en_sub:
	return g_add(ep, flags, op_sub);
	  case en_and:
	return g_logic(ep, flags, op_and);
	  case en_or:
	return g_logic(ep, flags, op_or);
	  case en_xor:
	return g_logic(ep, flags, op_eor);
	  case en_mul:
	/*
	 * special optimization possible if there are patterns matching the
	 * 68000 mulu, muls instructions. ugly, but it gives a big
	 * performance increase
	 */
	switch (type) {
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
		if (tst_ushort(ep->v.p[0]) && tst_ushort(ep->v.p[1])) {
		ep->v.p[0]->etp = tp_ushort;
		ep->v.p[1]->etp = tp_ushort;
		return g_xmul(ep, flags, op_mulu);
		} else if (tst_short(ep->v.p[0]) && tst_short(ep->v.p[1])) {
		ep->v.p[0]->etp = tp_short;
		ep->v.p[1]->etp = tp_short;
		return g_xmul(ep, flags, op_muls);
		}
		break;
	  default:
		break;
	}
	return g_mul(ep, flags);
	  case en_div:
	return g_div(ep, flags, FALSE);
	  case en_mod:
	return g_div(ep, flags, TRUE);
	  case en_lsh:
	return g_shift(ep, flags, op_asl);
	  case en_rsh:
	return g_shift(ep, flags, (is_unsigned_type(tp) ? op_lsr : op_asr));
	  case en_asadd:
	return g_asadd(ep, flags, op_add);
	  case en_assub:
	return g_asadd(ep, flags, op_sub);
	  case en_asand:
	return g_aslogic(ep, flags, op_and);
	  case en_asor:
	return g_aslogic(ep, flags, op_or);
	  case en_asxor:
	return g_aslogic(ep, flags, op_eor);
	  case en_aslsh:
	return g_asshift(ep, flags, op_asl);
	  case en_asrsh:
	return g_asshift(ep, flags, (is_unsigned_type(tp) ? op_lsr : op_asr));
	  case en_asmul:
	return g_asmul(ep, flags);
	  case en_asdiv:
	return g_asdiv(ep, flags, FALSE);
	  case en_asmod:
	return g_asdiv(ep, flags, TRUE);
	  case en_assign:
	return g_assign(ep, flags);
	  case en_ainc:
	return g_aincdec(ep, flags, op_add);
	  case en_adec:
	return g_aincdec(ep, flags, op_sub);
	  case en_eq:
	VOIDCAST g_compare(ep);
	op = op_seq;
	goto cont1;
	  case en_ne:
	VOIDCAST g_compare(ep);
	op = op_sne;
	goto cont1;
	  case en_lt:
	op = (OPCODE)(g_compare(ep) ? op_shi : op_sgt);
	goto cont1;
	  case en_le:
	op = (OPCODE)(g_compare(ep) ? op_shs : op_sge);
	goto cont1;
	  case en_gt:
	op = (OPCODE)(g_compare(ep) ? op_slo : op_slt);
	goto cont1;
	  case en_ge:
	op = (OPCODE)(g_compare(ep) ? op_sls : op_sle);
	goto cont1;
	  case en_not:
	g_test(ep->v.p[0]);
	op = op_seq;
cont1:
	ap1 = data_register();
	g_code(op, IL0, ap1, NIL_ADDRESS);
	g_immed(op_and, tp, 1l, ap1);
	return mk_legal(ap1, flags, tp);
	  case en_land:
	  case en_lor:
	lab0 = nextlabel++;
	lab1 = nextlabel++;
	g_falsejp(ep, lab0);
	ap1 = data_register();
	g_code(op_moveq, IL0, mk_immed(1l), ap1);
	g_branch(lab1);
	g_label(lab0);
	g_code(op_moveq, IL0, mk_immed(0l), ap1);
	g_label(lab1);
	return mk_legal(ap1, flags, tp);
	  case en_cond:
	return g_hook(ep, flags);
	  case en_comma:
	freeop(g_expr(ep->v.p[0], (FLAGS)(F_ALL | F_NOVALUE)));
	return g_expr(ep->v.p[1], flags);
	  case en_fcall:
	  case en_call:
	return g_fcall(ep, flags);
	  case en_cast:
		/*
		 * On the 68000, suppress all casts between any of
		 * long, unsigned long, pointer
		 */
	switch (type) {
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
			switch (ep->v.p[0]->etp->type) {
		  case bt_int32:
		  case bt_uint32:
		  case bt_long:
		  case bt_ulong:
		  case bt_pointer32:
				return g_expr(ep->v.p[0], flags);
		  default:
		break;
		}
		break;
	  default:
		break;
		}
		/*
		 * The cast really results in some work
		 */
	return g_cast(g_expr(ep->v.p[0], (FLAGS)(F_ALL | F_USES)),
			  ep->v.p[0]->etp, tp, flags);
	  case en_deref:
	/*
	 * The cases where this node occurs are handled automatically:
	 * g_assign and g_fcall return a pointer to a structure rather than a
	 * structure.
	 */
	return g_expr(ep->v.p[0], flags);
#ifdef ASM
	  case en_asm:
	return g_asm(ep);
#endif /* ASM */

	  case en_fcon:
	  default:
	FATAL ((__FILE__,"g_expr","uncoded node %d",ep->nodetype));
	return NIL_ADDRESS;
	}
}

PRIVATE void
g_expression P1(const EXPR*, ep)
{
	initstack();
	if (ep != NIL_EXPR)
	VOIDCAST g_expr(ep, (FLAGS)(F_ALL | F_NOVALUE));
	checkstack();
}

/*
 * generate code to do a comparison of the two operands of node. returns 1 if
 * it was an unsigned comparison
 */
static BOOL
g_compare P1(const EXPR*, ep)
{
	EXPR	*ep0 = ep->v.p[0];
	EXPR	*ep1 = ep->v.p[1];
	ADDRESS *ap1, *ap2, *ap3;
	TYP 	*tp = ep0->etp;
	switch (tp->type) {
	  case bt_uchar:
	  case bt_schar:
	  case bt_char:
	  case bt_charu:
	  case bt_ushort:
	  case bt_short:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	ap2 = g_expr(ep0, F_ALL);
	if (ap2->mode == am_immed)
		ap1 = g_expr(ep1, (FLAGS)(F_ALL & ~F_IMMED));
	else
		ap1 = g_expr(ep1, (FLAGS)(F_AREG | F_DREG));
	validate(ap2);
	/*
	 * sorry, no tst.l An on the 68000, but we can move to a data
	 * register if one is free
	 */
	if (ap1->mode == am_areg && ep0->nodetype == en_icon
		&& ep0->v.i == 0L && free_data()) {
		ap3 = data_register();
		g_code(op_move, IL4, ap1, ap3);
		/* tst.l ap3 not needed */
		freeop(ap3);
	} else {
#ifdef CMP_BUG
		if (!IandD_option && ap1->mode == am_direct) {
		ap3 = data_register();
		g_code(op_move, (ILEN)tp->size, ap1, ap3);
		g_code(op_cmp, (ILEN)tp->size, ap2, ap3);
		freeop(ap3);
		} else
#endif
		g_code(op_cmp, (ILEN)tp->size, ap2, ap1);
	}
	freeop(ap1);
	freeop(ap2);
	return is_unsigned_type(tp);
#ifdef FLOAT_IEEE
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	push_rtl_params (ep0, ep1);
	switch (tp->type) {
	  case bt_float:
		call_library(SUP_SFCMP);
		break;
	  case bt_double:
		call_library(SUP_DFCMP);
		break;
	  case bt_longdouble:
		call_library(SUP_LFCMP);
		break;
	  default:
		CANNOT_REACH_HERE();
		break;
	}
	return FALSE;
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	push_rtl_params (ep0, ep1);
	call_library(SUP_FPCMP);
	return FALSE;
#endif /* FLOAT_MFFP */
	  default:
	FATAL ((__FILE__,"g_compare","illegal type %d",tp->type));
	break;
	}
	return FALSE;
}

/*
 * Test the expression and set the condition codes accordingly
 */
static void
g_test P1(const EXPR*, ep)
{
	ADDRESS *ap;
	TYP 	*tp = ep->etp;
	switch(tp->type) {
#ifdef FLOAT_IEEE
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	push_rtl_params (ep, NIL_EXPR);
	switch(tp->type) {
	  case bt_float:
		call_library(SUP_SFTST);
		break;
	  case bt_double:
		call_library(SUP_DFTST);
		break;
	  case bt_longdouble:
		call_library(SUP_LFTST);
		break;
	  default:
		CANNOT_REACH_HERE();
		break;
	}
	/* note that the condition codes are tested,
	   not the return value */
	break;
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
	push_rtl_params (ep, NIL_EXPR);
	call_library(SUP_FPTST);
	break;
#endif /* FLOAT_MFFP */
	  default:
	ap = g_expr(ep, (FLAGS)(F_DREG | F_MEM));
	g_code(op_tst, (ILEN) tp->size, ap, NIL_ADDRESS);
	freeop(ap);
	}
}

/*
 * generate a jump to label if the node passed evaluates to a true condition.
 */
static void
g_truejp P2(const EXPR*, ep, LABEL, label)
{
	LABEL	 lab0;
	OPCODE	 op;
	if (ep == NIL_EXPR)
	FATAL ((__FILE__,"g_truejp","ep == 0"));
	if (ep->nodetype == en_icon) {
	if (ep->v.i)
		g_branch(label);
	return;
	}
	switch (ep->nodetype) {
	  case en_eq:
	VOIDCAST g_compare(ep);
	op = op_beq;
	g_cbranch(op, label);
	break;
	  case en_ne:
	VOIDCAST g_compare(ep);
	op = op_bne;
	g_cbranch(op, label);
	break;
	  case en_lt:
	op = (OPCODE)(g_compare(ep) ? op_bhi : op_bgt);
	g_cbranch(op, label);
	break;
	  case en_le:
	op = (OPCODE)(g_compare(ep) ? op_bhs : op_bge);
	g_cbranch(op, label);
	break;
	  case en_gt:
	op = (OPCODE)(g_compare(ep) ? op_blo : op_blt);
	g_cbranch(op, label);
	break;
	  case en_ge:
	op = (OPCODE)(g_compare(ep) ? op_bls : op_ble);
	g_cbranch(op, label);
	break;
	  case en_land:
	lab0 = nextlabel++;
	g_falsejp(ep->v.p[0], lab0);
	g_truejp(ep->v.p[1], label);
	g_label(lab0);
	break;
	  case en_lor:
	g_truejp(ep->v.p[0], label);
	g_truejp(ep->v.p[1], label);
	break;
	  case en_not:
	g_falsejp(ep->v.p[0], label);
	break;
	  default:
	g_test(ep);
	g_cbranch(op_bne, label);
	break;
	}
}

/*
 * generate code to execute a jump to label if the expression passed is
 * false.
 */
static void
g_falsejp P2(const EXPR*, ep, LABEL, label)
{
	LABEL	 lab0;
	OPCODE	 op;
	if (ep == NIL_EXPR)
	FATAL ((__FILE__,"g_falsejp","ep == 0"));
	if (ep->nodetype == en_icon) {
	if (!ep->v.i)
		g_branch(label);
	return;
	}
	switch (ep->nodetype) {
	  case en_eq:
	VOIDCAST g_compare(ep);
	op = op_bne;
	g_cbranch(op, label);
	break;
	  case en_ne:
	VOIDCAST g_compare(ep);
	op = op_beq;
	g_cbranch(op, label);
	break;
	  case en_lt:
	op = (OPCODE)(g_compare(ep) ? op_bls : op_ble);
	g_cbranch(op, label);
	break;
	  case en_le:
	op = (OPCODE)(g_compare(ep) ? op_blo : op_blt);
	g_cbranch(op, label);
	break;
	  case en_gt:
	op = (OPCODE)(g_compare(ep) ? op_bhs : op_bge);
	g_cbranch(op, label);
	break;
	  case en_ge:
	op = (OPCODE)(g_compare(ep) ? op_bhi : op_bgt);
	g_cbranch(op, label);
	break;
	  case en_land:
	g_falsejp(ep->v.p[0], label);
	g_falsejp(ep->v.p[1], label);
	break;
	  case en_lor:
	lab0 = nextlabel++;
	g_truejp(ep->v.p[0], lab0);
	g_falsejp(ep->v.p[1], label);
	g_label(lab0);
	break;
	  case en_not:
	g_truejp(ep->v.p[0], label);
	break;
	  default:
	g_test(ep);
	g_cbranch(op_beq, label);
	break;
	}
}

PRIVATE void
g_jtrue P2(const EXPR*, ep, LABEL, label)
{
	initstack();
	g_truejp(ep, label);
	checkstack();
}

PRIVATE void
g_jfalse P2(const EXPR*, ep, LABEL, label)
{
	initstack();
	g_falsejp(ep, label);
	checkstack();
}

PRIVATE void
g_switch_table P4(const EXPR*, ep, struct swtab*, sw, UVAL, min_caselabel, UVAL, max_caselabel)
{
	ADDRESS *ap, *ap1, *ap2;
	ILEN	 ilen = (ILEN) ep->etp->size;

	initstack();
	ap = g_expr(ep, (FLAGS)(F_DREG | F_VOL));
	ap1 = address_register();
	g_code(op_lea, IL0, mk_label(sw->tablab), ap1);
	/*
	 * move the interval
	 */
	max_caselabel -= min_caselabel;
	if (min_caselabel != 0) {
	g_immed2(op_sub, ep->etp, (IVAL)min_caselabel, ap);
	}
	g_immed(op_cmp, ep->etp, (IVAL)max_caselabel, ap);
	g_cbranch(op_bhi, sw->deflab);
#ifdef RELOC_BUG
	ilen = IL4;
	ap = g_extend(ap, ep->etp, tp_long);
	g_code(op_asl, ilen, mk_immed(2l), ap);
	freeop(ap1);
	ap1 = copy_addr(ap1, am_indx4);
	ap1->sreg = ap->preg;
	ap1->offset = mk_const(0L);
	ap2 = address_register () ;
	g_code(op_move,IL4,ap1,ap2);
	freeop(ap2);
	ap2 = copy_addr(ap2, am_ind);
	g_code(op_jmp, IL0, ap2, NIL_ADDRESS);
#else
	/* jump table contains 2 byte offset of case branches */
	if (max_caselabel > (UVAL)(65536L/2L)) {
	/* will we ever get this many case labels?	*/
	ilen = IL4;
	ap = g_extend(ap, ep->etp, tp_long);
	} else {
	ilen = IL2;
	ap = g_extend(ap, ep->etp, tp_short);
	}
	g_code(op_add, ilen, ap, ap);
	if (IandD_option) {
	g_code(op_add, ilen, ap, ap1);
	freeop(ap1);
	ap2 = copy_addr(ap1, am_ind);
	ap1 = data_register();
	g_code(op_move, ilen, ap2, ap1);
	freeop(ap1);

	ap2 = address_register();
	g_code(op_lea, IL0, mk_label(sw->beglab), ap2);
	g_code(op_add, IL2, ap1, ap2);
	freeop(ap2);
	ap2 = copy_addr(ap2, am_ind);
	} else {
	freeop(ap1);
	ap2 = copy_addr(ap1, (ilen == 2 ? am_indx4: am_indx2));
	ap2->sreg = ap->preg;
	ap2->offset = mk_const(0l);
	g_code(op_add, IL2, ap2, ap1);
	ap2 = copy_addr(ap2, am_ind);
	}
	g_code(op_jmp, IL0, ap2, NIL_ADDRESS);
	if (IandD_option)
	g_label (sw->beglab);
#endif
	freeop(ap);
	checkstack();
}

/*
 * Generate the body of a switch statement by comparing each case value
 * in turn.   The comparision is infact done by using subtraction as this
 * actually generates more efficient code (and would work best if the
 * labels were sorted!)
 */
PRIVATE void
g_switch_compare P2(const EXPR*, ep, STMT*, stmt)
{
	IVAL	min, max;
	IVAL	min_value = 0L;
	LABEL	label;
	ADDRESS *ap;
	initstack();
	ap = g_expr(ep, (FLAGS)(F_DREG | F_VOL));
	while (stmt != NIL_STMT) {
	if (stmt->stype != st_default) {
		for (min = stmt->v2.i; stmt != NIL_STMT; stmt = stmt->s1) {
		max = stmt->v2.i ;
		stmt->v2.l = label = nextlabel++;
		if ((stmt->s1 == NIL_STMT) ||
			(stmt->s1->stype == st_default) ||
			(stmt->s1 != stmt->v1.s) ||
			(stmt->s1->v2.i != max+1l)) {
			stmt = stmt->s1;
			break ;
		}
		}
		if ((min-min_value) < 0)
		g_immed2 (op_add, ep->etp, min_value-min, ap);
		else
		g_immed2 (op_sub, ep->etp, min-min_value, ap);
		min_value = min;
		switch (max - min) {
		  case 1:
		g_cbranch(op_beq, label);
		g_immed2 (op_sub, ep->etp, 1L, ap);
		min_value++;
		/*FALLTHRU*/
		  case 0:
		g_cbranch(op_beq, label);
		break ;
		  default:
		g_immed2 (op_cmp, ep->etp, max-min_value, ap) ;
		g_cbranch(op_bls, label);
		break ;
		}
	} else
		stmt = stmt->s1;
	}
	freeop(ap);
	checkstack();
}

PRIVATE void
g_entry P1(SIZE, frame_size)
{
	/* WARNING:  This routine generates code in the REVERSE order */

	if (frame_size < 32768l)
	g_coder(op_link, IL0, mk_reg(frameptr), mk_immed(-frame_size));
	else {
	g_coder(op_sub, IL4, mk_immed(frame_size-32768L), mk_reg(STACKPTR));
	g_coder(op_link, IL0, mk_reg(frameptr), mk_immed(-32768L));
	}

#ifdef STACK_CHECK
	if (stackcheck_option) {
	SYM 	*sp = internal_symbol(SUP_STACKCHECK, NIL_TYP);
	symbol_used(sp);
	g_coder(op_jsr, IL0, mk_strlab(sp->name), NIL_ADDRESS);
	g_coder(op_move, IL4, mk_immed(frame_size + max_stack_adjust), &push);
	}
#endif /* STACK_CHECK */
#ifdef PROBES
	if (probe_option) {
	SIZE	 size = frame_size + max_stack_adjust + 32l;
	ADDRESS *ap2, *ap = mk_reg(STACKPTR);
	if (size < 32768l) {
		ap->mode = am_indx;
		ap->offset = mk_const (-size);
		g_coder(op_tst, IL2, ap, NIL_ADDRESS);
	} else {
		ap2 = data_register();
		ap->mode = am_indx2;
		ap->sreg = ap->preg;
		ap->offset = mk_const (0L);
		g_coder(op_tst, IL2, ap, NIL_ADDRESS);
		g_coder(op_move, IL4, mk_immed(-size), ap2);
		freeop(ap2);
	}
	}
#endif /* PROBES */
	max_stack_adjust = 0L;
}

PRIVATE void
g_return P2(const EXPR*, stmtexp, TYP*, tp)
{
	EXPR	*ep, *ep1;
	ADDRESS *ap;
	initstack();
	switch (tp->type) {
	  case bt_struct:
	  case bt_union:
	uses_structassign = 1;
	/* assign structure */
	ep = mk_autocon((SIZE)8);
	ep = mk_ref(ep, tp_pointer);
	ep1 = mk_ref(ep, tp);
	ep1 = mk_node(en_assign, ep1, copynode(stmtexp), tp);
	VOIDCAST g_expr(ep1, (FLAGS)(F_ALL | F_NOVALUE));
	ap = g_expr(ep, F_ALL);
	g_code(op_move, IL4, ap, mk_reg(reg_usage->result->reg[0]));
	freeop(ap);
	break;

#ifdef FLOAT_SUPPORT
	  case bt_longdouble:
#ifdef FLOAT_IEEE
	/* get value in result.lo/result.hi/result.xt */
	ap = g_expr(stmtexp,F_MEM);
#ifdef MOVEMBUG
	{
		ADDRESS *ap1;
		g_code(op_move, IL4, ap, mk_reg(reg_usage->result->reg[0]));
		ap1 = copy_addr(ap, ap->mode);
		ap1->offset = mk_add(ap1->offset, mk_const(4l));
		g_code(op_move, IL4, ap1, mk_reg(reg_usage->result->reg[1]));
		ap1 = copy_addr(ap, ap->mode);
		ap1->offset = mk_add(ap1->offset, mk_const(8l));
		g_code(op_move, IL4, ap1, mk_reg(reg_usage->result->reg[2]));
	}
#else
	g_code(op_movem, IL4, ap, 
		mk_rmask((REGMASK) (1<<(int)reg_usage->result->reg[0] |
					1<<(int)reg_usage->result->reg[1] |
					1<<(int)reg_usage->result->reg[2])));
#endif /* MOVEMBUG */
	freeop(ap);
	break;

	  case bt_double:
	/* get value in result.lo/result.hi */
	ap = g_expr(stmtexp,F_MEM);
#ifdef MOVEMBUG
	{
		ADDRESS *ap1;
		g_code(op_move, IL4, ap, mk_reg(reg_usage->result->reg[0]));
		ap1 = copy_addr(ap, ap->mode);
		ap1->offset = mk_add(ap1->offset, mk_const(4l));
		g_code(op_move, IL4, ap1, mk_reg(reg_usage->result->reg[1]));
	}
#else
	g_code(op_movem, IL4, ap, mk_rmask(
				(REGMASK)(1<<(int)reg_usage->result->reg[0] |
					  1<<(int)reg_usage->result->reg[1])));
#endif /* MOVEMBUG */
	freeop(ap);
	break;
#endif /* FLOAT_IEEE */

#ifdef FLOAT_MFFP
	  case bt_double:
#endif /* FLOAT_MFFP */
	  case bt_float:
	/* fallthrough, return value in result.lo register */
#endif /* FLOAT_SUPPORT */

	  case bt_char:
	  case bt_charu:
	  case bt_uchar:
	  case bt_schar:
	  case bt_short:
	  case bt_ushort:
	  case bt_int16:
	  case bt_uint16:
	  case bt_int32:
	  case bt_uint32:
	  case bt_long:
	  case bt_ulong:
	  case bt_pointer32:
	ap = g_expr(stmtexp, F_ALL);
	g_code(op_move, (ILEN) stmtexp->etp->size, ap, mk_reg(reg_usage->result->reg[0]));
	freeop(ap);
	break;
	  default:
	FATAL((__FILE__,"g_return","illegal type %d",tp->type));
	}
	checkstack();
}

PRIVATE void
g_epilogue P0(void)
{
	ADDRESS *ap;
	SIZE	 stackoffset;

	if (regs_used > 0) {
	if (is_leaf_function)
		g_code(op_movem, IL4, &pop, mk_rmask(restore_mask));
	else {
		if (lc_auto > lc_auto_max)
		lc_auto_max = lc_auto;
		stackoffset = lc_auto_max + max_scratch + (4L*(SIZE)regs_used);
		if (stackoffset <= 32768l) {
		ap=mk_reg(frameptr);
		ap->mode = am_indx;
		ap->offset = mk_const(-stackoffset);
		g_code(op_movem, IL4, ap, mk_rmask(restore_mask));
		} else {
		ap = mk_reg(STACKPTR);
		g_code(op_move, IL4, mk_reg(frameptr), ap);
		g_code(op_sub,	IL4, mk_immed(stackoffset), ap);
		g_code(op_movem, IL4, &pop, mk_rmask(restore_mask));
		}
	}
	}
	g_code(op_unlk, IL0, mk_reg(frameptr), NIL_ADDRESS);
	g_code(op_rts, IL0, NIL_ADDRESS, NIL_ADDRESS);
}

/*
 * allocate will allocate registers for the expressions that have a high
 * enough desirability.
 */
PRIVATE void
g_allocate P1(CSE*, olist)
{
	CSE 	*csp;
	REG 	 datareg = D7;
	REG 	 addreg = (REG)((int)frameptr - 1);
	REGMASK  mask = (REGMASK)0;
	TYP 	*tp;
	EXPR	*ep;
	unsigned	 value;
	USES	 uses;

	regs_used = 0;
	for (value = (unsigned)3; value <= (unsigned)5 ; value++) {
	for (csp = olist; csp != NIL_CSE; csp = csp->next) {
		if (csp->reg != NO_REG)
		continue;		/* register already allocated */

		uses = desire(csp);
		if (uses < (USES)3 || (!reg_option && uses < (USES)5000))
		continue;		/* not wanted in register */

		ep = csp->exp;
		tp = ep->etp;
		if ( (addreg > max_areg)
			/*
			**	integer constants may have different types
			*/
			 && (csp->exp->nodetype != en_icon)
			/*
			**	only those items which have sufficient
			**	dereferences are put into address registers
			*/
			 && (csp->duses > (USES)((unsigned)csp->uses / value))
			/*
			**	only allow pointer types
			*/
			 && (tp->type == bt_pointer32 ||
			 tp->type == bt_func)) {

		csp->reg = addreg--;
		mask |= (REGMASK) (1 << (int)csp->reg);
		regs_used++;
		} else if ( (addreg > max_areg)
			/*
			**	integer constants may have different types
			*/
			 && (csp->exp->nodetype != en_icon)
			/*
			** only those items which have sufficient
			** dereferences are put into address registers
			*/
			 && (csp->duses > (USES)((unsigned)csp->uses / (unsigned)3))
			/*
			**	the types which are fine in address registers.
			**	Allow only 32-bit integral and signed 16-bit
			**	integral types
			*/
			 && (tp->type == bt_short ||
			 tp->type == bt_int16 ||
			 tp->type == bt_int32 ||
			 tp->type == bt_uint32 ||
			 tp->type == bt_long ||
			 tp->type == bt_ulong ||
			 tp->type == bt_pointer32 ||
			 tp->type == bt_func) ) {

		csp->reg = addreg--;
		mask |= (REGMASK) (1 << (int)csp->reg);
		regs_used++;
		} else if ((value == (unsigned)5) && (datareg > max_dreg)
			/*
			 * the types which are fine in data registers.
			 * Allow all types except for 'float', 'double'
			 * and 'long double'
			 */
			 && (tp->type != bt_longdouble)
			 && (tp->type != bt_double)
			 && (tp->type != bt_float)
			 ) {
		csp->reg = datareg--;
		mask |= (REGMASK)(1 << (int)csp->reg);
		regs_used++;
		}
	}
	}

	/*
	**	Now take into account which registers must be saved by the
	**	function.
	*/
	mask &= reglist_to_mask(reg_usage->save);

	if (mask != (REGMASK)0) {
	g_code(op_movem, IL4, mk_smask(mask), &push);
	}
	restore_mask = mask;
}


/*
**	Go through the common sub-expression tree and check to see if
**	any registers must be loaded with a value.
*/

PRIVATE void
g_preload P1(CSE *, olist)
{
	CSE *csp;
	EXPR *ep;
	ADDRESS *ap, *ap2;
	for (csp = olist; csp != NIL_CSE; csp = csp->next) {
	if (csp->reg != NO_REG) {   /* see if preload needed */
		ep = csp->exp;
		if (!is_lvalue(ep) || (ep->v.p[0]->v.i > 0L)) {
		initstack();
		ap = g_expr(ep, F_ALL);
		ap2 = mk_reg(csp->reg);
		g_code(op_move, (ILEN) ep->etp->size, ap, ap2);
		freeop(ap);
		}
	}
	}
}

PRIVATE void
g_flush P1(SYM*, sp)
{
	put_literals();
	put_cseg(AL_FUNC);
	if (sp)
	put_name(sp);
	flush_peep();
}

PRIVATE void
g_auto_align P0(void)
{
	if (frame_option > D0)
	frameptr = (REG)frame_option;
	if (frameptr < A0)
	frameptr += A0;
	if (frameptr > FRAMEPTR)
	frameptr = FRAMEPTR;

	if (lc_auto_max % AL_DEFAULT != 0L)
	lc_auto_max += AL_DEFAULT - (lc_auto_max % AL_DEFAULT);
}

PRIVATE BOOL
g_isbigendian P0(void)
{
	return TRUE;
}

PRIVATE EXPR *
g_transform P1(EXPR *,ep)
{
	SIZE size;
	EXPR *ep1;
	SYM *sp;
	if (ep == NIL_EXPR)
	return ep;
	switch (ep->nodetype) {
#ifdef FLOAT_SUPPORT
#ifndef FLOAT_BOOTSTRAP
	  case en_fcon:
	ep1 = mk_lcon(mk_flabel(&(ep->v.f), ep->etp));
	ep1 = mk_ref(ep1, ep->etp);
	return ep1;
#endif /* FLOAT_BOOTSTRAP */
#endif /* FLOAT_SUPPORT */
	  case en_icon:
	  case en_nacon:
	  case en_labcon:
	  case en_autocon:
	  case en_sym:
	  case en_ref:
	  case en_fieldref:
	  case en_register:
	  case en_size:
#ifdef ASM
	  case en_asm:
#endif /* ASM */
	break;
	  case en_eq:
	  case en_ne:
	  case en_lt:
	  case en_le:
	  case en_gt:
	  case en_ge:
	  case en_add:
	  case en_sub:
	  case en_div:
	  case en_mod:
	  case en_lsh:
	  case en_rsh:
	  case en_and:
	  case en_or:
	  case en_xor:
	  case en_land:
	  case en_lor:
	  case en_cond:
	  case en_comma:
	  case en_list:
	  case en_asadd:
	  case en_assub:
	  case en_asmul:
	  case en_asdiv:
	  case en_asor:
	  case en_asxor:
	  case en_asand:
	  case en_asmod:
	  case en_aslsh:
	  case en_asrsh:
	  case en_fcall:
	  case en_call:
	  case en_assign:
	  case en_mul:
	  case en_ainc:
	  case en_adec:
	ep->v.p[1] = g_transform(ep->v.p[1]);
	/*FALLTHRU*/
	  case en_not:
	  case en_compl:
	  case en_cast:
	  case en_deref:
	ep->v.p[0] = g_transform(ep->v.p[0]);
		break;
	  case en_uminus:
#if 0
	switch (ep->etp->type) {
#ifdef FLOAT_IEEE
	  case bt_float:
		sp = internal_symbol(SUP_SFNEG, ep->etp);
		break;
	  case bt_double:
		sp = internal_symbol(SUP_DFNEG, ep->etp);
		break;
	  case bt_longdouble:
		sp = internal_symbol(SUP_LFNEG, ep->etp);
		break;
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
	  case bt_float:
	  case bt_double:
	  case bt_longdouble:
		sp = internal_symbol(SUP_FPNEG, ep->etp);
		break;
#endif /* FLOAT_MFFP */
	  default:
		return ep;
	}
	ep->nodetype = en_call;
	ep1 = mk_node(en_list, ep->v.p[0]->v.p[0], NIL_EXPR, tp_void);
	ep->v.p[1] = ep1;
	ep->v.p[0] = mk_symnode(sp);
#else
	ep->v.p[0] = g_transform(ep->v.p[0]);
#endif
	break;
	  case en_asmul2:
	sp = internal_symbol(SUP_ASMUL, ep->etp);
	goto cont1;
	  case en_asdiv2:
	sp = internal_symbol(SUP_ASDIV, ep->etp);
cont1:
	ep->v.p[1] = g_transform(ep->v.p[1]);
	ep->v.p[0] = g_transform(ep->v.p[0]);
	size = (ep->etp->size *8) << 8; 	/* size of LHS in bits * 256 */
	switch (ep->v.p[0]->nodetype) {
	  case en_fieldref:
		size  = (SIZE)(ep->v.p[0]->v.bit.width << 8);
		size |= (tp_long->size*8L) -
			(SIZE)(ep->v.p[0]->v.bit.offset) -
			(SIZE)(ep->v.p[0]->v.bit.width);
		/*FALLTHRU*/
	  case en_ref:
		size |= (SIZE)(ep->v.p[1]->etp->size << 16);	/* size of RHS*/
		size |= ((SIZE)is_unsigned_type(ep->etp) <<24); /* sign of LHS*/
		ep->nodetype = en_call;
		ep1 = mk_node(en_list, ep->v.p[0]->v.p[0], NIL_EXPR, tp_void);
		ep1 = mk_node(en_list, ep->v.p[1]->v.p[0], ep1, tp_void);
		ep1 = mk_node(en_list, mk_icon(size, tp_long), ep1, tp_void);
		ep->v.p[1] = ep1;
		ep->v.p[0] = mk_symnode(sp);
		break;
	  default:
		CANNOT_REACH_HERE();
	}
	break;
	  default:
		CANNOT_REACH_HERE();
		break;
	}
	return ep;
}

#ifdef MULTIPLE_PROCESSORS
struct genfuncs mc68k_funcs  = {
	g_expression,
	g_jtrue,
	g_jfalse,
	g_stack,
	g_switch_table,
	g_switch_compare,
	g_entry,
	g_return,
	g_epilogue,
	g_label,
	g_branch,
#ifdef DEBUGOPT
	g_line,
#endif /*DEBUGOPT*/
	g_allocate,
	g_preload,
	g_flush,
	g_auto_align,
	g_alignment,
	g_isbigendian,
	g_transform
	};
#endif /* MULTIPLE_PROCESSORS */
#endif /* MC680X0 */
