/*
 * C compiler
 * ==========
 *
 * Copyright 1989, 1990, 1991 Christoph van Wuellen.
 * Credits to Matthew Brandt.
 * All commercial rights reserved.
 *
 * This compiler may be redistributed as long there is no
 * commercial interest. The compiler must not be redistributed
 * without its full sources. This notice must stay intact.
 *
 * History:
 *
 * 1989   starting an 68000 C compiler, starting with material
 *        originally by M. Brandt
 * 1990   68000 C compiler further bug fixes
 *        started i386 port (December)
 * 1991   i386 port finished (January)
 *        further corrections in the front end and in the 68000
 *        code generator.
 *        The next port will be a SPARC port
 */

#ifndef _GEN68K_H
#define _GEN68K_H

#define	AL_ILLEGAL	((SIZE) 1)

#ifdef MC68000
/*
 * MC68000 is a 16-bit processor. Word alignment is OK in all cases
 */
#define AL_CHAR 	((SIZE) 1)
#define AL_SHORT	((SIZE) 2)
#define AL_LONG 	((SIZE) 2)
#define AL_POINTER	((SIZE) 2)
#define AL_FLOAT	((SIZE) 2)
#define AL_DOUBLE	((SIZE) 2)
#define AL_LONGDOUBLE	((SIZE) 2)
#define AL_STRUCT	((SIZE) 2)
#define AL_FUNC		((SIZE) 2)

#define AL_DEFAULT	((SIZE) 2)	/* alignment suitable for all types */
#endif /* MC68000 */

#ifdef MC68010
/*
 * MC68010 is a 16-bit processor. Word alignment is OK in all cases
 */
#define AL_CHAR 	((SIZE) 1)
#define AL_SHORT	((SIZE) 2)
#define AL_LONG 	((SIZE) 2)
#define AL_POINTER	((SIZE) 2)
#define AL_FLOAT	((SIZE) 2)
#define AL_DOUBLE	((SIZE) 2)
#define	AL_LONGDOUBLE	((SIZE) 2)
#define AL_STRUCT	((SIZE) 2)
#define AL_FUNC		((SIZE) 2)

#define AL_DEFAULT	((SIZE) 2)	/* alignment suitable for all types */
#endif /* MC68010 */

#ifdef MC68020
/*
 * perhaps not necessary, but useful: 32-bit alignment for 32-bit types
 */
#define AL_CHAR 	((SIZE) 1)
#define AL_SHORT	((SIZE) 2)
#define AL_LONG 	((SIZE) 4)
#define AL_POINTER	((SIZE) 4)
#define AL_FLOAT	((SIZE) 4)
#define AL_DOUBLE	((SIZE) 4)
#define	AL_LONGDOUBLE	((SIZE) 4)
#define AL_STRUCT	((SIZE) 4)
#define AL_FUNC		((SIZE) 4)

#define AL_DEFAULT	((SIZE) 4)	/* alignment suitable for all types */
#endif /* MC68020 */

#ifdef MC68030
/*
 * perhaps not necessary, but useful: 32-bit alignment for 32-bit types
 */
#define AL_CHAR 	((SIZE) 1)
#define AL_SHORT	((SIZE) 2)
#define AL_LONG 	((SIZE) 4)
#define AL_POINTER	((SIZE) 4)
#define AL_FLOAT	((SIZE) 4)
#define AL_DOUBLE	((SIZE) 4)
#define	AL_LONGDOUBLE	((SIZE) 4)
#define AL_STRUCT	((SIZE) 4)
#define AL_FUNC		((SIZE) 4)

#define AL_DEFAULT	((SIZE) 4)	/* alignment suitable for all types */
#endif /* MC68030 */

#ifdef MC68040
/*
 * perhaps not necessary, but useful: 32-bit alignment for 32-bit types
 */
#define AL_CHAR 	((SIZE) 1)
#define AL_SHORT	((SIZE) 2)
#define AL_LONG 	((SIZE) 4)
#define AL_POINTER	((SIZE) 4)
#define AL_FLOAT	((SIZE) 4)
#define AL_DOUBLE	((SIZE) 4)
#define	AL_LONGDOUBLE	((SIZE) 4)
#define AL_STRUCT	((SIZE) 4)
#define AL_FUNC		((SIZE) 4)

#define AL_DEFAULT	((SIZE) 4)	/* alignment suitable for all types */
#endif /* MC68040 */

enum e_flags {
	F_NONE    = 0,
	F_DREG    = 1,	/* data register direct mode allowed	*/
	F_AREG    = 2,	/* address register direct mode allowed	*/
	F_MEM     = 4,	/* memory alterable modes allowed	*/
	F_IMMED	  = 8,	/* immediate mode allowed		*/
	F_VOL	  = 16,	/* need volitile operand		*/
	F_NOVALUE = 32,	/* dont need result value		*/
	F_USES	  = 64	/* need result value more than once	*/
			/* this forbids autoincrement modes	*/
};
#define	F_ALL	  ((FLAGS)(F_DREG | F_AREG | F_MEM | F_IMMED))	/* all modes allowed */

/* The instructions */

/*
 * The order of the branch instructions must not be changed
 * since array revcond[] in the peephole optimizer relies on
 * them.
 */
enum e_op {
    op_move,
    op_moveq,
    op_add,
    op_addi,
    op_addq,
    op_sub,
    op_subi,
    op_subq,
    op_muls,
    op_mulu,
    op_divs,
    op_divu,
    op_and,
    op_andi,
    op_or,
    op_ori,
    op_eor,
    op_asl,
    op_lsr,
    op_jmp,
    op_jsr,
    op_bsr,
    op_movem,
    op_rts,
    op_bra,
    op_beq,
    op_bne,
    op_blt,
    op_ble,
    op_bgt,
    op_bge,
    op_bhi,
    op_bhs,
    op_blo,
    op_bls,
    op_btst,
    op_tst,
    op_ext,
    op_lea,
    op_swap,
    op_neg,
    op_not,
    op_cmp,
    op_clr,
    op_link,
    op_unlk,
    op_pea,
    op_cmpi,
    op_dbra,
    op_asr,
    op_rol,
    op_ror,
    op_seq,
    op_sne,
    op_slt,
    op_sle,
    op_sgt,
    op_sge,
    op_shi,
    op_shs,
    op_slo,
    op_sls,
    op_st,
    op_nop,
    op_fabs,
    op_fneg,
    op_fadd,
    op_fsub,
    op_fdiv,
    op_fmul,
    op_fmod,
    op_fcmp,
    op_ftst,
    op_fmove,
    op_fmovem,
#ifdef ASM
    op_asm,
#endif /* ASM */
    op_line,
    op_label
};
#define OP_MIN	op_move
#define OP_MAX	op_label

/* addressing modes */
enum e_am {
    am_dreg,		/* Dn */
    am_areg,		/* An */
    am_ind,		/* (An) */
    am_ainc,		/* (An)+ */
    am_adec,		/* -(An) */
    am_indx,		/* (d16,An) */
    am_indx2,		/* (d8,An,Dn.l) */
    am_indx3,		/* (d8,An,Am.l) */
    am_indx4,		/* (d8,An,Dn.w) */
    am_direct,		/* (xxx) */
    am_immed,		/* #(data) */
    am_smask,		/* Register mask (D0 high, A7 low) */
    am_rmask,		/* Register mask (A7 high, D0 low) */
    am_freg,		/* Floating Pointer Register FPn */
    am_none,
    am_mreg,		/* Dn/Dm */
#ifdef ASM
    am_asm,		/* ASM string */
#endif /* ASM */
    am_line		/* pseudo mode for line numbers */
};

/* register naming, special registers */

#define	D0		((REG) 0)	/* D0 register */
#define	D1		((REG) 1)	/* D1 register */
#define	D2		((REG) 2)	/* D2 register */
#define	D3		((REG) 3)	/* D3 register */
#define	D4		((REG) 4)	/* D4 register */
#define	D5		((REG) 5)	/* D5 register */
#define	D6		((REG) 6)	/* D6 register */
#define	D7		((REG) 7)	/* D7 register */

#define	A0		((REG) 8)	/* A0 register */
#define	A1		((REG) 9)	/* A1 register */
#define	A2		((REG) 10)	/* A2 register */
#define	A3		((REG) 11)	/* A3 register */
#define	A4		((REG) 12)	/* A4 register */
#define	A5		((REG) 13)	/* A5 register */
#define	A6		((REG) 14)	/* A6 register */
#define	A7		((REG) 15)	/* A7 register */

#define	NUM_REGS	16	/* number of distinct registers */

#define FRAMEPTR  	A6	/* frame pointer register */
#define STACKPTR  	A7	/* system stack pointer register */

#define MAX_ADDR	A1	/* max. scratch address register (A1) */
#define MAX_DATA	D2	/* max. scratch data	register (D2) */

#define	MAX_REG_STACK	((DEEP)30)
#define	UNUSED		((DEEP)-1)
#define	EMPTY		((DEEP)0)

/* support routines */
#define	SUP_ASMUL	(CHAR *)".Xasmul"
#define	SUP_ASDIV	(CHAR *)".Xasdiv"
#define	SUP_ASLDIV	(CHAR *)".Xasldiv"
#define	SUP_ASLMUL	(CHAR *)".Xaslmul"
#define	SUP_ASLREM	(CHAR *)".Xaslrem"
#define	SUP_ASULDIV	(CHAR *)".Xasuldiv"
#define	SUP_ASULMUL	(CHAR *)".Xasulmul"
#define	SUP_ASULREM	(CHAR *)".Xasulrem"
#define	SUP_LDIV	(CHAR *)".Xldiv"
#define	SUP_LMUL	(CHAR *)".Xlmul"
#define	SUP_LREM	(CHAR *)".Xlrem"
#define	SUP_ULDIV	(CHAR *)".Xuldiv"
#define	SUP_ULMUL	(CHAR *)".Xulmul"
#define	SUP_ULREM	(CHAR *)".Xulrem"
#ifdef FLOAT_IEEE
#define	SUP_ASDFADD	(CHAR *)".Xasdfadd"
#define	SUP_ASDFDIV	(CHAR *)".Xasdfdiv"
#define	SUP_ASDFMUL	(CHAR *)".Xasdfmul"
#define	SUP_ASDFSUB	(CHAR *)".Xasdfsub"
#define	SUP_ASLFADD	(CHAR *)".Xaslfadd"
#define	SUP_ASLFDIV	(CHAR *)".Xaslfdiv"
#define	SUP_ASLFMUL	(CHAR *)".Xaslfmul"
#define	SUP_ASLFSUB	(CHAR *)".Xaslfsub"
#define	SUP_ASSFSUB	(CHAR *)".Xassfsub"
#define	SUP_ASSFADD	(CHAR *)".Xassfadd"
#define	SUP_ASSFDIV	(CHAR *)".Xassfdiv"
#define	SUP_ASSFMUL	(CHAR *)".Xassfmul"
#define	SUP_DFADD	(CHAR *)".Xdfadd"
#define	SUP_DFCMP	(CHAR *)".Xdfcmp"
#define	SUP_DFDIV	(CHAR *)".Xdfdiv"
#define	SUP_DFLTODF	(CHAR *)".Xdfltodf"
#define	SUP_DFMUL	(CHAR *)".Xdfmul"
#define	SUP_DFNEG	(CHAR *)".Xdfneg"
#define	SUP_DFSUB	(CHAR *)".Xdfsub"
#define	SUP_DFTOL	(CHAR *)".Xdftol"
#define	SUP_DFTOLF	(CHAR *)".Xdftolf"
#define	SUP_DFTOSF	(CHAR *)".Xdftosf"
#define	SUP_DFTOUL	(CHAR *)".Xdftoul"
#define	SUP_DFTST	(CHAR *)".Xdftst"
#define	SUP_DFUTODF	(CHAR *)".Xdfutodf"
#define	SUP_LFADD	(CHAR *)".Xlfadd"
#define	SUP_LFCMP	(CHAR *)".Xlfcmp"
#define	SUP_LFDIV	(CHAR *)".Xlfdiv"
#define	SUP_LFMUL	(CHAR *)".Xlfmul"
#define	SUP_LFNEG	(CHAR *)".Xlfneg"
#define	SUP_LFSUB	(CHAR *)".Xlfsub"
#define	SUP_LFLTOLF	(CHAR *)".Xlfltodf"
#define	SUP_LFTODF	(CHAR *)".Xlftodf"
#define	SUP_LFTOL	(CHAR *)".Xlftol"
#define	SUP_LFTOUL	(CHAR *)".Xlftoul"
#define	SUP_LFTOSF	(CHAR *)".Xlftosf"
#define	SUP_LFTST	(CHAR *)".Xlftst"
#define	SUP_LFUTOLF	(CHAR *)".Xlfutolf"
#define	SUP_SFADD	(CHAR *)".Xsfadd"
#define	SUP_SFCMP	(CHAR *)".Xsfcmp"
#define	SUP_SFDIV	(CHAR *)".Xsfdiv"
#define	SUP_SFLTOSF	(CHAR *)".Xsfltosf"
#define	SUP_SFMUL	(CHAR *)".Xsfmul"
#define	SUP_SFNEG	(CHAR *)".Xsfneg"
#define	SUP_SFSUB	(CHAR *)".Xsfsub"
#define	SUP_SFTODF	(CHAR *)".Xsftodf"
#define	SUP_SFTOL	(CHAR *)".Xsftol"
#define	SUP_SFTOLF	(CHAR *)".Xsftodf"
#define	SUP_SFTOUL	(CHAR *)".Xsftoul"
#define	SUP_SFTST	(CHAR *)".Xsftst"
#define	SUP_SFUTOSF	(CHAR *)".Xsfutosf"
#endif /* FLOAT_IEEE */
#ifdef FLOAT_MFFP
#define	SUP_ASFPADD	(CHAR *)".Xasfpadd"
#define	SUP_ASFPDEC	(CHAR *)".Xasfpdec"
#define	SUP_ASFPDIV	(CHAR *)".Xasfpdiv"
#define	SUP_ASFPINC	(CHAR *)".Xasfpinc"
#define	SUP_ASFPMULT	(CHAR *)".Xasfpmult"
#define	SUP_ASFPREM	(CHAR *)".Xasfprem"
#define	SUP_ASFPSUB	(CHAR *)".Xasfpsub"
#define	SUP_FPADD	(CHAR *)".Xfpadd"
#define	SUP_FPADD	(CHAR *)".Xfpadd"
#define	SUP_FPCMP	(CHAR *)".Xfpcmp"
#define	SUP_FPDEC	(CHAR *)".Xfpdec"
#define	SUP_FPDIV	(CHAR *)".Xfpdiv"
#define	SUP_FPFTOL	(CHAR *)".Xfpftol"
#define	SUP_FPFTOU	(CHAR *)".Xfpftou"
#define	SUP_FPINC	(CHAR *)".Xfpinc"
#define	SUP_FPLTOF	(CHAR *)".Xfpltof"
#define	SUP_FPMULT	(CHAR *)".Xfpmult"
#define	SUP_FPNEG	(CHAR *)".Xfpneg"
#define	SUP_FPREM	(CHAR *)".Xfprem"
#define	SUP_FPSUB	(CHAR *)".Xfpsub"
#define	SUP_FPSUB	(CHAR *)".Xfpsub"
#define	SUP_FPTST	(CHAR *)".Xfptst"
#define	SUP_FPUTOF	(CHAR *)".Xfputof"
#endif /* FLOAT_IEEE */
#ifdef STACK_CHECK
#define	SUP_STACKCHECK	(CHAR *)".stackcheck"
#endif /* STACK_CHECK */

#ifndef _CODE_DEFINED
#define _CODE_DEFINED
typedef struct ocode	CODE;
#endif

/* instruction lengths */

enum ilength {
	IL0 = 0,
	IL1 = 1,
	IL2 = 2,
	IL4 = 4,
	IL8 = 8
	};

typedef struct amode	ADDRESS;
typedef	enum e_flags	FLAGS;
typedef	enum e_op	OPCODE;
typedef	enum ilength	ILEN;
typedef	unsigned int	REGMASK;
typedef	char		DEEP;
typedef	struct reg_use	REGUSAGE;
typedef	struct reglist	REGLIST;

#define NIL_ADDRESS	((ADDRESS *) 0)
#define	NIL_CODE	((CODE *) 0)


/* addressing mode structure */

struct amode {
    enum e_am       mode;	/* addressing mode */
    REG             preg, sreg;	/* register(s) used in addressing mode */
    DEEP   	    deep;
    EXPR	   *offset;	/* expression used in addressing mode */
};

/* output code structure */

struct ocode {
    OPCODE          opcode;	/* opcode for this instruction */
    ILEN	    length;	/* operand length of the instruction */
    ADDRESS	   *oper1;	/* first operand */
    ADDRESS	   *oper2;	/* second operand */
    CODE	   *fwd;	/* next instruction */
    CODE	   *back;	/* previous instruction */
};

struct reglist {
    int		number;		/* number of registers in the list */
    REG *	reg;		/* register list */
};

/*
**	The usage of registers is controlled by the information held
**	within the following structure.
*/
struct	reg_use {
    REGLIST *	parameter;	/* Registers used to pass parameters */
    REGLIST *	save;		/* Registers saved by the function */
    REGLIST *	result;		/* Registers used to return results */
};

#define BRANCH_COUNT	2	/* abandon branch optimistaion if exceeded */

#ifdef MULTIPLE_PROCESSORS
/*
 * remap function names - it is necessary to change the names of functions
 * if there are multiple code generators build into the compiler in order
 * to prevent name clashes.
 *
 * The following defines do the necessary renaming
 */
#define	data_register		data_register68K
#define	address_register	address_register68K
#define	mdata_register		mdata_register68K
#define	alignment		alignment68k
#define	checkstack		checkstack68k
#define	copy_addr		copy_addr68k
#define	is_equal_address	is_equal_address68k
#define	flush_peep		flush_peep68k
#define	freeop			freeop68k
#define	g_code			g_code68k
#define	g_coder			g_coder68k
#define	get_register		get_register68k
#define	initstack		initstack68k
#define	mk_label		mk_label68k
#define	mk_reg			mk_reg68k
#define	mk_mreg			mk_mreg68k
#define	temp_inv		temp_inv68k
#define	validate		validate68k
#define	reg_usage		reg_usage68k
#define	reglist_to_mask		reglist_to_mask68k
#define	max_dreg		max_dreg68K
#define	max_areg		max_areg68K
#endif /* MULTIPLE_PROCESSORS */

extern	REG	max_dreg;	/* maximum temporary data register */
extern	REG	max_areg;	/* maximum temporary address register */
extern	REGUSAGE *reg_usage;	/* register usage */

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

/* gen68k.c */
ADDRESS *mk_reg			P_((REG));
ADDRESS *mk_mreg		P_((REG, REG));
ADDRESS *mk_label		P_((LABEL));
ADDRESS *copy_addr		P_((ADDRESS*, enum e_am));
BOOL	is_short		P_((const EXPR *));

/* peep68k */
void	g_code			P_((OPCODE, ILEN, ADDRESS *, ADDRESS *));
void	g_coder			P_((OPCODE, ILEN, ADDRESS *, ADDRESS *));
void	flush_peep		P_((void));
BOOL	is_equal_address	P_((ADDRESS *, ADDRESS *));

/* reg68k */
ADDRESS *data_register		P_((void));
ADDRESS *mdata_register		P_((void));
ADDRESS *address_register	P_((void));
ADDRESS *temp_reg		P_((FLAGS));
BOOL	free_data		P_((void));
BOOL	free_addr		P_((void));
REGMASK	reglist_to_mask		P_((const REGLIST *));
void	initstack		P_((void));
void	checkstack		P_((void));
void	validate		P_((const ADDRESS *));
void	freeop			P_((const ADDRESS *));
void	temp_inv		P_((void));

#undef	P_

#endif /* _GEN68K_H */
