/*
 *	ieeetest.c
 *
 *	This program is intended to do a first level confidence check
 *	on the IEEE support routines for C68.  More detailed checks
 *	should be performed as part of a serious library test suite.
 *	(such as the Plum Hall validation suite).
 */

#include <stdio.h>

char    _progname[]="IEEE Test";
char    _version[] ="v1.11";
char    _copyright[] =  "(c) D.J.Walker, 1992-1994";

#ifdef QDOS
#include <qdos.h>
struct  WINDOWDEF _condetails = {2,1,0,7,452,235,30,10};
void    (*_consetup)() = consetup_title;
#endif /* QDOS */

double frexp (double value, int *eptr);
double ldexp (double value, int exp);
double modf  (double value, double *iptr);
float  modff (float  value, float *iptr);

union  F {
	float f;
	unsigned long  v;
	} xf, yf, zf;

union  D {
	double d;
	unsigned long  v[2];
	} xd, yd, zd;

void
printf_float (msg, f)
char *msg;
union F *f;
{
	float xf;

	xf = 1.0;

	printf (" float (%s):  sign=%d,  exp= %02.2x  mant=%06.6lx\n",
		msg,
		(int)((f->v >> 31) & 0x01),
		(int)((f->v >> 23) & 0xFF),
		(long)((f->v << 1) & 0xFFFFFF));
}

void
printf_double (msg, d)
char *msg;
union D *d;
{
	printf ("double (%s):  sign=%d,  exp=%03.3x mant=%05.5lx%08.8lx\n",
		msg,
		(int)((d->v[0] >> 31) & 0x01),
		(int)((d->v[0] >> 20) & 0x7FF),
		(long)((d->v[0] ) & 0xFFFFF),
		(unsigned long)d->v[1]);
}

main()
{
	unsigned long ul;
	unsigned int ui;
	long	l;
	int	i;

#ifndef QDOS
printf("\n%s %s\t\t%s\n",_progname, _version, _copyright);
#endif /* QDOS */

	printf("\n---------------- simple assignments -------------\n\n");
	xf.f = 10.0; printf_float ("xf=10.0", &xf);
	xd.d = 10.0; printf_double ("xd=10.0", &xd);

	xf.f = -xf.f; printf_float ("xf = -xf",&xf);
	xd.d = -xd.d; printf_double ("xd = -xd",&xd);

	printf("\n---------------- arithmetic operators -------------\n\n");
	yd.d = 1.0;
	xd.d = 1.0 + yd.d; printf_double ("double add (1.0+1.0)",&xd);
	xd.d = 2.0 - yd.d; printf_double ("double sub (2.0-1.0)",&xd);

	printf("\n");
	yf.f = 1.0;
	xf.f = 1.0 + yf.f; printf_float ("float add (1.0+1.0)",&xf);
	xf.f = 2.0 - yf.f; printf_float ("float sub (2.0-1.0)",&xf);

	printf("\n");
	xd.d = 2.0 * 2.0; printf_double ("multiply 2.0 * 2.0",&xd);
	xd.d = 4.0 / 2.0; printf_double ("divide   4.0 / 2.0",&xd);

	printf("\n");
	xf.f = 2.0 * 2.0; printf_float ("multiply 2.0 * 2.0",&xf);
	xf.f = 4.0 / 2.0; printf_float ("divide   4.0 / 2.0",&xf);

	getchar();
	printf("\n---------------- operator assignment -------------\n\n");
	xd.d = 1.0;
	xd.d += 1.0; printf_double ("plus assignment     1.0 += 1.0", &xd);
	xd.d -= 1.0; printf_double ("minus assignment    2.0 -= 1.0", &xd);
	xd.d *= 1.0; printf_double ("multiply assignment 1.0 *= 1.0", &xd);
	xd.d /= 1.0; printf_double ("divide assignment   1.0 /= 1.0", &xd);

	printf("\n");
	xf.f = 1.0; 
	xf.f += 1.0; printf_float ("plus assignment)    1.0 += 1.0", &xf);
	xf.f -= 1.0; printf_float ("minus assignment    2.0 -= 1.0", &xf);
	xf.f *= 2.0; printf_float ("multiply assignment 1.0 *= 2.0", &xf);
	xf.f /= 4.0; printf_float ("divide assignment   2.0 /= 4.0", &xf);

	getchar();
	printf("\n---------------- arithemtic compares -------------\n\n");
	printf ("double compare (equal       ) is\t%d\t%d\n",1.0==1.0, 1.0==2.0);
	printf ("double compare (not equal   ) is\t%d\t%d\n",1.0!=2.0, 1.0!=1.0);
	printf ("double compare (less than   ) is\t%d\t%d\n",1.0< 2.0, 1.0<-1.0);
	printf ("double compare (greater than) is\t%d\t%d\n",1.0>-1.0, 1.0>2.0);

	printf ("\n");
	printf (" float compare (equal       ) is\t%d\t%d\n",1.0==1.0, 1.0==2.0);
	printf (" float compare (not equal   ) is\t%d\t%d\n",1.0!=2.0, 1.0!=1.0);
	printf (" float compare (less than   ) is\t%d\t%d\n",1.0< 2.0, 1.0<-1.0);
	printf (" float compare (greater than) is\t%d\t%d\n",1.0>-1.0, 1.0>2.0);

	printf("\n---------------- logical compares -------------\n\n");
	printf ("double test true=%d, false=%d\n",(1.0 ? 1 : 0),(!1.0 ? 1 : 0));
	printf ("float  test true=%d, false=%d\n",(1.0 ? 1  :0),(!1.0 ? 1 : 0));

	printf("\n---------------- conversions (fp to fp )-------------\n\n");
	xd.d = 1.0; xf.f = xd.d; printf_float ("from double:  1.0",&xf);
	xd.d =-1.0; xf.f = xd.d; printf_float ("from double: -1.0",&xf);
	xf.f = 1.0; xd.d = xf.f; printf_double ("from float:  1.0",&xd);
	xf.f =-1.0; xd.d = xf.f; printf_double ("from float: -1.0",&xd);

	getchar();
	printf("\n---------------- conversions (integers to fp) -------------\n\n");
	l = 1;  xd.d = l;  printf_double(" long(1) to double",&xd);
	i = 1;  xd.d = i;  printf_double("  int(1) to double",&xd);
	ul = 1; xd.d = ul; printf_double("ulong(1) to double",&xd);
	ui = 1; xd.d = ui; printf_double(" uint(1) to double",&xd);
	l = 1;  xf.f = l;  printf_float (" long(1) to float",&xf);
	i = 1;  xf.f = i;  printf_float ("  int(1) to float",&xf);
	ul = 1; xf.f = ul; printf_float ("ulong(1) to float",&xf);
	ui = 1; xf.f = ui; printf_float (" uint(1) to float",&xf);
	printf("\n");
	l = -1;  xd.d = l;  printf_double(" long(-1) to double",&xd);
	i = -1;  xd.d = i;  printf_double("  int(-1) to double",&xd);
	ul = -1; xd.d = ul; printf_double("ulong(-1) to double",&xd);
	ui = -1; xd.d = ui; printf_double(" uint(-1) to double",&xd);
	l = -1;  xf.f = l;  printf_float (" long(-1) to float",&xf);
	i = -1;  xf.f = i;  printf_float ("  int(-1) to float",&xf);
	ul = -1; xf.f = ul; printf_float ("ulong(-1) to float",&xf);
	ui = -1; xf.f = ui; printf_float (" uint(-1) to float",&xf);

	getchar();
	printf("\n---------------- conversions (fp to integers )-------------\n\n");
	i = 1.1;  printf ("double(1.1) to int   = %d\n", i);
	l = 1.1;  printf ("double(1.1) to long  = %ld\n", l);
	ui = 1.1; printf ("double(1.1) to uint  = %d\n", ui);
	ul = 1.1; printf ("double(1.1) to ulong = %ld\n", ul);
	i = -1.1;  printf ("double(-1.1) to int   = %d\n", i);
	l = -1.1;  printf ("double(-1.1) to long  = %ld\n", l);
	ui = -1.1; printf ("double(-1.1) to uint  = %d\n", ui);
	ul = -1.1; printf ("double(-1.1) to ulong = %ld\n", ul);
	i = 1.1;  printf ("float(1.1) to int   = %d\n", i);
	l = 1.1;  printf ("float(1.1) to long  = %ld\n", l);
	ui = 1.1; printf ("float(1.1) to uint  = %d\n", ui);
	ul = 1.1; printf ("float(1.1) to ulong = %ld\n", ul);
	i = -1.1;  printf ("float(-1.1) to int   = %d\n", i);
	l = -1.1;  printf ("float(-1.1) to long  = %ld\n", l);
	ui = -1.1; printf ("float(-1.1) to uint  = %d\n", ui);
	ul = -1.1; printf ("float(-1.1) to ulong = %ld\n", ul);

/*
 *	The next section checks the standard libc routines that are direectly
 *	aware of the bit representation of the floating point numbers and are
 *	resposnsible for splitting them up/assembling them.
 *	NOTE.	If your routines for printing floating point numbers are not
 *		working properly, then you may get some strange results.  Check
 *		the bit patterns for the floating point numbers as well.
 */
	getchar();
	printf("\n---------------- frexp -------------\n\n");
	printf ("yd.d=frexp (xd.d(=+3.4), &i)\n");
	xd.d = 3.4;	yd.d = frexp (xd.d, &i);
	printf ("    %g 	",xd.d); printf_double ("xd.d", &xd);
	printf ("    %g 	",yd.d); printf_double ("yd.d", &yd);
	printf ("    %d		integer (i)\n",i);

	printf("\n");
	printf ("yd.d=frexp (xd.d(=-3.4), &i)\n");
	xd.d = -3.4;	yd.d = frexp (xd.d, &i);
	printf ("   %g 	",xd.d); printf_double ("xd.d", &xd);
	printf ("   %g 	",yd.d); printf_double ("yd.d", &yd);
	printf ("   %d		integer (i)\n",i);

	printf("\n");
	printf ("yd.d=frexp (xd.d(=+0.34), &i)\n");
	xd.d = 0.34;	yd.d = frexp (xd.d, &i);
	printf ("    %g	",xd.d); printf_double ("xd.d", &xd);
	printf ("    %g	",yd.d); printf_double ("yd.d", &yd);
	printf ("    %d		integer (i)\n",i);

	printf("\n");
	printf ("yd.d=frexp (xd.d(=0.0), &i)\n");
	xd.d = 0.0;	yd.d = frexp (xd.d, &i);
	printf ("    %g		",xd.d); printf_double ("xd.d", &xd);
	printf ("    %g		",yd.d); printf_double ("yd.d", &yd);
	printf ("    %d		integer (i)\n",i);

	getchar();
	printf("\n---------------- ldexp -------------\n\n");
	printf ("yd.d=ldexp (xd.d(=+0.85), 2)\n");
	xd.d = 0.85; 	yd.d = ldexp (xd.d, 2);
	printf ("    %g 	",xd.d); printf_double ("xd.d", &xd);
	printf ("    %g 	",yd.d); printf_double ("yd.d", &yd);

	printf("\n");
	printf ("yd.d=ldexp (xd.d(=-0.85), 2)\n");
	xd.d = -0.85;	yd.d = ldexp (xd.d, 2);
	printf ("   %g 	",xd.d); printf_double ("xd.d", &xd);
	printf ("   %g 	",yd.d); printf_double ("yd.d", &yd);
	printf("\n");

	printf ("yd.d=ldexp (xd.d(=0.0), 0)\n");
	xd.d = 0.0;	yd.d = ldexp (xd.d, 0);
	printf ("    %g		",xd.d); printf_double ("xd.d", &xd);
	printf ("    %g		",yd.d); printf_double ("yd.d", &yd);

	getchar();
	printf("\n---------------- modf -------------\n\n");
	printf ("yd.d=modf (xd.d(=+3.4), &zd.d)\n");
	xd.d = 3.4;	yd.d = modf (xd.d, &zd.d);
	printf ("    %g 	",xd.d); printf_double ("xd.d", &xd);
	printf ("    %g 	",yd.d); printf_double ("yd.d", &yd);
	printf ("    %g		",zd.d); printf_double ("zd.d", &zd);

	printf("\n"); 
	printf ("yd.d=modf (xd.d(=-3.4), &zd.d)\n");
	xd.d = -3.4;	yd.d = modf (xd.d, &zd.d);
	printf ("   %g 	",xd.d); printf_double ("xd.d", &xd);
	printf ("   %g 	",yd.d); printf_double ("yd.d", &yd);
	printf ("   %g		",zd.d); printf_double ("zd.d", &zd);

	printf("\n"); 
	printf ("yd.d=modf (xd.d(=0.0), &zd.d)\n");
	xd.d = 0.0;	yd.d = modf (xd.d, &zd.d);
	printf ("   %g		",xd.d); printf_double ("xd.d", &xd);
	printf ("   %g		",yd.d); printf_double ("yd.d", &yd);
	printf ("   %g		",zd.d); printf_double ("zd.d", &zd);

	getchar();
	printf("---------------- modff -------------\n\n");
	printf ("yf.f=modff (xf.f(=+3.4), &zf.f)\n");
	xf.f = 3.4;	yf.f = modff (xf.f, &zf.f);
	printf ("    %g 	",xf.f); printf_float ("xf.f", &xf);
	printf ("    %g 	",yf.f); printf_float ("yf.f", &yf);
	printf ("    %g  	",zf.f); printf_float ("zf.f", &zf);

	printf("\n");
	printf ("yf.f=modff (xf.f(=-3.4), &zf.f)\n");
	xf.f = -3.4;	yf.f = modff (xf.f, &zf.f);
	printf ("   %g 	",xf.f); printf_float ("xf.f", &xf);
	printf ("   %g 	",yf.f); printf_float ("yf.f", &yf);
	printf ("   %g  	",zf.f); printf_float ("zf.f", &zf);

	printf("\n"); 
	printf ("yf.f=modff (xf.f(=0.0), &zf.f)\n");
	xf.f = 0.0;	yf.f = modff (xf.f, &zf.f);
	printf ("   %g		",xf.f); printf_float ("xf.f", &xf);
	printf ("   %g		",yf.f); printf_float ("yf.f", &yf);
	printf ("   %g		",zf.f); printf_float ("zf.f", &zf);

	printf ("\n\n");
}
