/*
 * C compiler
 * ==========
 *
 * Copyright 1989, 1990, 1991 Christoph van Wuellen.
 * Credits to Matthew Brandt.
 * All commercial rights reserved.
 *
 * This compiler may be redistributed as long there is no
 * commercial interest. The compiler must not be redistributed
 * without its full sources. This notice must stay intact.
 *
 * History:
 *
 * 1989   starting an 68000 C compiler, starting with material
 *        originally by M. Brandt
 * 1990   68000 C compiler further bug fixes
 *        started i386 port (December)
 * 1991   i386 port finished (January)
 *        further corrections in the front end and in the 68000
 *        code generator.
 *        The next port will be a SPARC port
 */

#include "chdr.h"
#include "expr.h"
#include "cglbdec.h"
#include "proto.h"

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

/* optimize.c */
#ifndef MC680X0
static	int	pwrof2		P_(( IVAL));
#endif /* MC680X0 */
static	IVAL	xfold		P_((EXPR *));
static	EXPR *	dooper		P_((EXPR *));
static	EXPR 	*fold_const	P_((EXPR *));

#undef P_

/*
**	dooper() will execute a constant operation in a node and return
**	the node to be the result of the operation.
*/
static EXPR *
dooper P1(EXPR*, ep)
{
    EXPRTYPE     type = ep->nodetype;
#ifdef FLOAT_SUPPORT
    RVAL f;
#endif /* FLOAT_SUPPORT */

#define ep0 ep->v.p[0]
#define ep1 ep->v.p[1]
#define epu ep->v.u
#define epi ep->v.i
#define epf ep->v.f

    ep->nodetype = ep0->nodetype;
    switch (ep0->nodetype) {
      case en_fcon:
#ifdef FLOAT_SUPPORT
#ifndef FLOAT_BOOTSTRAP
	FASSIGN(f, ep0->v.f);
        switch (type) {
          case en_uminus:
	    FASSIGN(epf, f);
	    FNEG(epf);
            break;
          case en_not:
            epi = FTST(f) ? (IVAL) 0 : (IVAL) 1;
            ep->nodetype = en_icon;
            break;
	  case en_cast:
	    FTOL(epi, f);
	    ep->nodetype = en_icon;
	    break;
          case en_add:
	    FADD3(epf, f, ep1->v.f);
            break;
          case en_sub:
	    FSUB3(epf, f, ep1->v.f);
            break;
          case en_mul:
	    FMUL3(epf, f, ep1->v.f);
            break;
          case en_div:
	    if (FTST(ep1->v.f)) {
		FDIV3(epf, f, ep1->v.f);
	    } else {
	        ep->nodetype = en_div;
	    }
            break;
          case en_eq:
            epi = (IVAL) FEQ(f, ep1->v.f);
            ep->nodetype = en_icon;
            break;
          case en_ne:
            epi = (IVAL) FNE(f, ep1->v.f);
            ep->nodetype = en_icon;
            break;
          case en_land:
            epi = (IVAL) (FTST(f) && FTST(ep1->v.f));
            ep->nodetype = en_icon;
            break;
          case en_lor:
            epi = (IVAL) (FTST(f) || FTST(ep1->v.f));
            ep->nodetype = en_icon;
            break;
          case en_lt:
            epi = (IVAL) FLT(f, ep1->v.f);
            ep->nodetype = en_icon;
            break;
          case en_le:
            epi = (IVAL) FLE(f, ep1->v.f);
            ep->nodetype = en_icon;
            break;
          case en_gt:
            epi = (IVAL) FGT(f, ep1->v.f);
            ep->nodetype = en_icon;
            break;
          case en_ge:
            epi = (IVAL) FGE(f, ep1->v.f);
            ep->nodetype = en_icon;
            break;
          default:
	    CANNOT_REACH_HERE();
            break;
        }
#endif /* FLOAT_BOOTSTRAP */
#endif /* FLOAT_SUPPORT */
	break;
      case en_icon:
	if (is_unsigned_type(ep0->etp)) {
	    UVAL u = ep0->v.u;
	    switch (type) {
	      case en_uminus:
		/*
		**	unary minus on an unsigned is normally a mistake so we must
		**	fool the compiler into not giving a warning.
		*/
		epu = (UVAL) (- (IVAL)u);
		break;
	      case en_not:
		epu = (u ? (UVAL) 0 : (UVAL) 1);
		break;
	      case en_compl:
		epu = (UVAL) strip_icon((IVAL) ~u, ep->etp);
		break;
#ifdef FLOAT_SUPPORT
	      case en_cast:
		ep->nodetype = en_fcon;
		UTOF(epf, u);
		break;
#endif /* FLOAT_SUPPORT */
	      case en_add:
		epu = u + ep1->v.u;
		break;
	      case en_sub:
		epu = u - ep1->v.u;
		break;
	      case en_mul:
		epu = u * ep1->v.u;
		break;
	      case en_div:
		if (ep1->v.u == (UVAL) 0) {
		    ep->nodetype = en_div;
		} else {
		    epu = u / ep1->v.u;
		}
		break;
	      case en_mod:
		if (ep1->v.u == (UVAL) 0) {
		    ep->nodetype = en_mod;
		} else {
		    epu = u % ep1->v.u;
		}
		break;
	      case en_and:
		epu = u & ep1->v.u;
		break;
	      case en_or:
		epu = u | ep1->v.u;
		break;
	      case en_xor:
		epu = u ^ ep1->v.u;
		break;
	      case en_eq:
		epu = (UVAL) (u == ep1->v.u);
		break;
	      case en_ne:
		epu = (UVAL) (u != ep1->v.u);
		break;
	      case en_land:
		epu = (UVAL) (u && ep1->v.u);
		break;
	      case en_lor:
		epu = (UVAL) (u || ep1->v.u);
		break;
	      case en_lt:
		epu = (UVAL) (u < ep1->v.u);
		break;
	      case en_le:
		epu = (UVAL) (u <= ep1->v.u);
		break;
	      case en_gt:
		epu = (UVAL) (u > ep1->v.u);
		break;
	      case en_ge:
		epu = (UVAL) (u >= ep1->v.u);
		break;
	      case en_lsh:
		epu = u << ep1->v.u;
		break;
	      case en_rsh:
		epu = u >> ep1->v.u;
		break;
	      default:
		CANNOT_REACH_HERE();
		break;
	    }
	} else {
	    IVAL i = ep0->v.i;
	    switch (type) {
	      case en_uminus:
		epi = - i;
		break;
	      case en_not:
		epi = (i) ? (IVAL) 0 : (IVAL) 1;
		break;
	      case en_compl:
		epi = strip_icon(~i, ep->etp);
		break;
#ifdef FLOAT_SUPPORT
	      case en_cast:
		ep->nodetype = en_fcon;
		LTOF(epf, i);
		break;
#endif /* FLOAT_SUPPORT */
	      case en_add:
		epi = i + ep1->v.i;
		break;
	      case en_sub:
		epi = i - ep1->v.i;
		break;
	      case en_mul:
		epi = i * ep1->v.i;
		break;
	      case en_div:
		if (ep1->v.i == (IVAL) 0) {
		    ep->nodetype = en_div;
		} else {
		    epi = i / ep1->v.i;
		}
		break;
	      case en_mod:
		if (ep1->v.i == (IVAL) 0) {
		    ep->nodetype = en_mod;
		} else {
		    epi = i % ep1->v.i;
		}
		break;
	      case en_and:
		epi = i & ep1->v.i;
		break;
	      case en_or:
		epi = i | ep1->v.i;
		break;
	      case en_xor:
		epi = i ^ ep1->v.i;
		break;
	      case en_eq:
		epi = (IVAL) (i == ep1->v.i);
		break;
	      case en_ne:
		epi = (IVAL) (i != ep1->v.i);
		break;
	      case en_land:
		epi = (IVAL) (i && ep1->v.i);
		break;
	      case en_lor:
		epi = (IVAL) (i || ep1->v.i);
		break;
	      case en_lt:
		epi = (IVAL) (i < ep1->v.i);
		break;
	      case en_le:
		epi = (IVAL) (i <= ep1->v.i);
		break;
	      case en_gt:
		epi = (IVAL) (i > ep1->v.i);
		break;
	      case en_ge:
		epi = (IVAL) (i >= ep1->v.i);
		break;
	      case en_lsh:
		epi = i << ep1->v.i;
		break;
	      case en_rsh:
		epi = i >> ep1->v.i;
		break;
	      default:
		CANNOT_REACH_HERE();
		break;
	    }
	}
	break;
      default:
	CANNOT_REACH_HERE();
	break;
    }
#undef ep0
#undef ep1
#undef epi
#undef epf
    return ep;
}


/*
**	return which power of two i is or -1.
*/
#ifdef MC680X0
int
#else
static int
#endif /* MC680X0 */
pwrof2 P1( IVAL, i)
{
    int             p;
    IVAL            q;
    q = (IVAL) 2;
    p = 1;
    while (q > (IVAL) 0) {
	if (q == i)
	    return p;
	q <<= 1l;
	++p;
    }
    return -1;
}


/*
**	opt0() - delete useless expressions and combine constants.
**
**	opt0 will delete expressions such as
**	 x + 0,
**	 x - 0,
**	 x * 0,
**	 x * 1,
**	 0 / x,
**	 x / 1,
**	 x mod 0,
**	 etc from the tree pointed to by node and combine obvious
**	constant operations. It cannot combine name and label constants
**	but will combine icon type nodes.
*/
EXPR *
opt0 P1(EXPR*, ep)
{
#define ep0 ep->v.p[0]
#define ep1 ep->v.p[1]

    IVAL         val, sc;
    EXPRTYPE     nt;
    TYP	    	*tp;
    if (ep == NIL_EXPR)
	return ep;
    tp = ep->etp;
    nt = ep->nodetype;
    switch (nt) {
      case en_uminus:
      case en_compl:
      case en_not:
	ep0 = opt0(ep0);
        if (nt != en_not && ep0->nodetype == nt) {
	    /*
	    **	This operation applied twice is a no-op
	    */
            ep = ep0->v.p[0];
            break;
        }
	if (ep0->nodetype == en_icon || ep0->nodetype == en_fcon)
	    ep = dooper(ep);
	break;
      case en_cast:
	ep0 = opt0(ep0);
	if (ep0->nodetype == en_icon || ep0->nodetype == en_fcon) {
	    ep = dooper(ep);
#if 1       /* This test does not work! */
	} else if (ep0->nodetype == en_register 
        && ep0->etp->size == 4L 
        && ep->etp->size == 4L
        && is_integral_type(ep0->etp)
        && is_integral_type(ep->etp)) {
	    return (ep0);
#endif
	}
	break;
      case en_add:
      case en_sub:
	ep0 = opt0(ep0);
	ep1 = opt0(ep1);
        /*
         *  a + (-b)  =  a - b
         *  a - (-b)  =  a + b
         *  (-a) + b  =  b - a
         */
        if (ep1->nodetype == en_uminus) {
            ep1 = ep1->v.p[0];
            ep->nodetype = nt = (EXPRTYPE)((nt == en_add) ? en_sub : en_add);
        }
        if (ep0->nodetype == en_uminus && nt == en_add) {
            ep0 = ep0->v.p[0];
            swap_nodes(ep);
            ep->nodetype = nt = en_sub;
        }
        /*
        **	constant expressions
        */
        if ((ep0->nodetype == en_icon && ep1->nodetype == en_icon) ||
            (ep0->nodetype == en_fcon && ep1->nodetype == en_fcon))   {
            ep = dooper(ep);
            break;
        }
	if (nt == en_add) {
	    if (ep1->nodetype == en_autocon)
		swap_nodes(ep);
	}
	if (ep0->nodetype == en_icon) {
	    if (ep0->v.i == 0L) {
		if (nt == en_sub) {
		    ep0 = ep1;
		    ep->nodetype = nt = en_uminus;
		} else
		    ep = ep1;
		    ep->etp = tp;
		break;
	    }
	} else if (ep1->nodetype == en_icon) {
	    if (ep0->nodetype == en_autocon && nt == en_add) {
		ep0->v.i += ep1->v.i;
		ep = ep0;
		ep->etp = tp;
		break;
	    }
	    if (ep1->v.i == 0L) {
		ep = ep0;
		ep->etp = tp;
		break;
	    }
	}
	break;
      case en_mul:
	ep0 = opt0(ep0);
	ep1 = opt0(ep1);
        /*
        **	constant expressions
        */
        if ((ep0->nodetype == en_icon && ep1->nodetype == en_icon) ||
            (ep0->nodetype == en_fcon && ep1->nodetype == en_fcon))  {
            ep = dooper(ep);
            break;
        }
	if (ep0->nodetype == en_icon) {
	    val = ep0->v.i;
	    if (val == (IVAL) 0) {
		ep = mk_node(en_comma, ep1, ep0, tp);
		break;
	    }
	    if (val == (IVAL) 1) {
		ep = ep1;
		break;
	    }
	    sc = (IVAL) pwrof2(val);
	    if (sc != (IVAL) -1) {
		swap_nodes(ep);
		ep1->v.i = sc;
		ep->nodetype = en_lsh;
	    }
	} else if (ep1->nodetype == en_icon) {
	    val = ep1->v.i;
	    if (val == (IVAL) 0) {
		ep = mk_node(en_comma, ep0, ep1, tp);
		break;
	    }
	    if (val == (IVAL) 1) {
		ep = ep0;
		break;
	    }
	    sc = (IVAL) pwrof2(val);
	    if (sc != (IVAL) -1) {
		ep1->v.i = sc;
		ep->nodetype = en_lsh;
	    }
	}
	break;
      case en_div:
	ep0 = opt0(ep0);
	ep1 = opt0(ep1);
        /*
        **	constant expressions
        */
        if ((ep0->nodetype == en_icon && ep1->nodetype == en_icon) ||
            (ep0->nodetype == en_fcon && ep1->nodetype == en_fcon))  {
            ep = dooper(ep);
            break;
        }
	if (ep0->nodetype == en_icon) {
	    if (ep0->v.i == 0L) {	/* 0/x */
		ep = mk_node(en_comma, ep1, ep0, tp);
		break;
	    }
	} else if (ep1->nodetype == en_icon) {
	    val = ep1->v.i;
	    if (val == (IVAL) 1) {	/* x/1 */
		ep = ep0;
		break;
	    }
	    if (is_unsigned_type(tp)) {
	        sc = (IVAL) pwrof2(val);
	        if (sc != (IVAL) -1) {
		    ep1->v.i = sc;
		    ep->nodetype = en_rsh;
	        }
	    }
	}
	break;
      case en_mod:
	ep0 = opt0(ep0);
	ep1 = opt0(ep1);
        /*
        **	constant expressions
        */
        if ((ep0->nodetype == en_icon && ep1->nodetype == en_icon) ||
            (ep0->nodetype == en_fcon && ep1->nodetype == en_fcon))  {
            ep = dooper(ep);
            break;
        }
	if (ep1->nodetype == en_icon) {
	    if (is_unsigned_type(tp)) {
		sc = (IVAL) pwrof2(ep1->v.i);
		if (sc != (IVAL) -1) {
		    ep1->v.i = (IVAL)bitmask((int)sc);
		    ep->nodetype = en_and;
		}
	    }
	}
	break;
      case en_lor:
      case en_and:
      case en_or:
      case en_xor:
      case en_eq:
      case en_ne:
      case en_land:
      case en_lt:
      case en_le:
      case en_gt:
      case en_ge:
	ep0 = opt0(ep0);
	if (nt == en_lor && ep0->nodetype == en_icon && ep0->v.i) {
	    /* short-circuit! */
	    ep0->v.i = (IVAL)1;
	    ep = ep0;
	    break;
	}
	ep1 = opt0(ep1);
        /*
        **	constant expressions
        */
        if ((ep0->nodetype == en_icon && ep1->nodetype == en_icon) ||
            (ep0->nodetype == en_fcon && ep1->nodetype == en_fcon))
            ep = dooper(ep);
	break;
      case en_lsh:
      case en_rsh:
	ep0 = opt0(ep0);
	ep1 = opt0(ep1);
	if (ep0->nodetype == en_icon && ep1->nodetype == en_icon) {
	    ep = dooper(ep);
            break;
        }
        /*
        **	optimize a << 0   and a >> 0
        */
	if (ep1->nodetype == en_icon && ep1->v.i == 0L)
	    ep = ep0;
	break;
      case en_cond:
	ep0 = opt0(ep0);
	ep1 = opt0(ep1);
        if (ep0->nodetype == en_icon) {
	    if (ep0->v.i)
		ep=ep1->v.p[0];
	    else
		ep=ep1->v.p[1];
	}
	break;
      case en_asand:
      case en_asor:
      case en_asxor:
      case en_asadd:
      case en_assub:
      case en_asmul:
      case en_asmul2:
      case en_asdiv:
      case en_asdiv2:
      case en_asmod:
      case en_asrsh:
      case en_aslsh:
      case en_fcall:
      case en_call:
      case en_comma:
      case en_list:
      case en_assign:
	ep1 = opt0(ep1);
	/*FALLTHRU*/
      case en_ref:
      case en_fieldref:
      case en_ainc:
      case en_adec:
      case en_deref:
	ep0 = opt0(ep0);
	break;
      default:
	break;
    }
    return ep;

#undef ep0
#undef ep1
}


/*
**	xfold() will remove constant nodes and return the values to the
**	calling routines.
*/
static IVAL
xfold P1(EXPR*, ep)
{
    IVAL            i;
    if (ep == NIL_EXPR)
	return 0L;
    switch (ep->nodetype) {
      case en_icon:
	i = ep->v.i;
	ep->v.i = 0L;
	return i;
      case en_add:
	return xfold(ep->v.p[0]) + xfold(ep->v.p[1]);
      case en_sub:
	return xfold(ep->v.p[0]) - xfold(ep->v.p[1]);
      case en_mul:
	if (ep->v.p[0]->nodetype == en_icon)
	    return xfold(ep->v.p[1]) * ep->v.p[0]->v.i;
	else if (ep->v.p[1]->nodetype == en_icon)
	    return xfold(ep->v.p[0]) * ep->v.p[1]->v.i;
	else {
	    ep->v.p[1] = fold_const(ep->v.p[1]);
	    ep->v.p[0] = fold_const(ep->v.p[0]);
	    return 0L;
	}
      case en_uminus:
	return -xfold(ep->v.p[0]);
      case en_lsh:
      case en_rsh:
      case en_div:
      case en_mod:
      case en_asadd:
      case en_assub:
      case en_asmul:
      case en_asmul2:
      case en_asdiv:
      case en_asdiv2:
      case en_asmod:
      case en_and:
      case en_land:
      case en_or:
      case en_lor:
      case en_xor:
      case en_asand:
      case en_asor:
      case en_asxor:
      case en_comma:
      case en_list:
      case en_fcall:
      case en_call:
      case en_assign:
      case en_eq:
      case en_ne:
      case en_lt:
      case en_le:
      case en_gt:
      case en_ge:
	ep->v.p[1] = fold_const(ep->v.p[1]);
	/*FALLTHRU*/
      case en_ref:
      case en_fieldref:
      case en_compl:
      case en_not:
      case en_deref:
      case en_cast:
	ep->v.p[0] = fold_const(ep->v.p[0]);
	return 0L;
      default:
	break;
    }
    return 0L;
}

/*
**	reorganize an expression for optimal constant grouping.
*/
static EXPR *
fold_const P1(EXPR*, ep)
{
    EXPR	*ep1;
    IVAL         i;
    if (ep == NIL_EXPR)
	return ep;
    switch (ep->nodetype) {
      case en_add:
	if (ep->v.p[0]->nodetype == en_icon) {
	    swap_nodes(ep);
	}
	if (ep->v.p[1]->nodetype == en_icon) {
	    ep->v.p[1]->v.i += xfold(ep->v.p[0]);
	    return ep;
	}
	break;
      case en_sub:
	if (ep->v.p[0]->nodetype == en_icon) {
	    ep->v.p[0]->v.i -= xfold(ep->v.p[1]);
	    return ep;
	}
	if (ep->v.p[1]->nodetype == en_icon) {
	    ep->v.p[1]->v.i -= xfold(ep->v.p[0]);
	    return ep;
	}
	break;
#if 0
      case en_icon:
	return ep;
#endif
      default:
	break;
    }
    i = xfold(ep);
    if (i != 0L) {
	/*
	**	strip_icon() is in fact harmless here since this value is
	**	just added to ep
	**
	**	consider in 16-bit mode:
	**
	**		int day, year;
	**		day = 365 * (year - 1970);
	**
	**	and look at the code, which is transformed to
	**
	**		day = 365*year + 1846;
	**
	**	which works if the multiplication returns the lower 16 bits
	**	of the result correctly.
	*/
	i = strip_icon(i, ep->etp);
	ep1 = mk_icon(i, ep->etp);
	ep = mk_node(en_add, ep, ep1, ep->etp);
    }
    return ep;
}


/*
**	apply all constant optimizations.
*/
EXPR *
constantopt P1(EXPR*, ep)
{
    ep = fold_const(ep);
    return opt0(ep);
}
