/*
 * C compiler
 * ==========
 *
 * Copyright 1989, 1990, 1991 Christoph van Wuellen.
 * Credits to Matthew Brandt.
 * All commercial rights reserved.
 *
 * This compiler may be redistributed as long there is no
 * commercial interest. The compiler must not be redistributed
 * without its full sources. This notice must stay intact.
 *
 * History:
 *
 * 1989   starting an 68000 C compiler, starting with material
 *        originally by M. Brandt
 * 1990   68000 C compiler further bug fixes
 *        started i386 port (December)
 * 1991   i386 port finished (January)
 *        further corrections in the front end and in the 68000
 *        code generator.
 *        The next port will be a SPARC port
 */

#include "config.h"

#ifdef MC680X0

#include "chdr.h"
#include "expr.h"
#include "cglbdec.h"
#include "proto.h"
#include "gen68k.h"
#include "outproto.h"

static CODE *peep_head = NIL_CODE;
static CODE *next_ip;
static int changes;

static OPCODE revcond[] = { op_bne, op_beq, op_bge, op_bgt, op_ble, op_blt,
			       op_bls, op_blo, op_bhs, op_bhi };

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

/* peep68k.c */
static	CODE *	block_end	P_((CODE*));
static	CODE *	find_label	P_(( LABEL));
static	int	label_references P_((CODE *));
static	BOOL	uses_label	P_((ADDRESS *, LABEL));
static	BOOL	uses_address	P_((ADDRESS *, ADDRESS *));
static	BOOL	dest_overwritten P_((ADDRESS *, CODE *));
static	CODE *	code		P_((OPCODE, ILEN, ADDRESS *, ADDRESS *));
static	void	add_peep	P_((CODE *));
static	void	peep_delete	P_((CODE *));
static	void	peep_pea	P_((CODE *));
static	void	peep_lea	P_((CODE *));
static	void	peep_move	P_((CODE *));
static	void	peep_movem	P_((CODE *));
static	void	peep_add	P_((CODE *));
static	void	peep_and	P_((CODE *));
static	void	peep_clr	P_((CODE *));
static	void	peep_sub	P_((CODE *));
static	void	peep_cmp	P_((CODE *));
static	void	peep_tst	P_((CODE *));
static	void	peep_uctran	P_((CODE *));
static	void	peep_bxx	P_((CODE *));
static	void	peep_ext	P_((CODE *));
static	void	peep_label	P_((CODE *));
static	void	check_label	P_((CODE *, CODE *));
static	void	peep_bra	P_((CODE *));
static	void	peep_jmp	P_((CODE *));
static	void	peep_line	P_((CODE *));
static	void	opt3		P_((void));
static	BOOL	is_equal_oper	P_((ADDRESS *, ADDRESS *));

#define	DEST_MODIFY	((unsigned char)1)
#define	DEST_OVERWRITE	((unsigned char)2)
#define	DEST_ALTERED	(DEST_MODIFY | DEST_OVERWRITE)

static unsigned char op_flags[] = {
	/* op_move */	DEST_OVERWRITE,
	/* op_moveq */	DEST_OVERWRITE,
	/* op_add */	DEST_MODIFY,
	/* op_addi */	DEST_MODIFY,
	/* op_addq */	DEST_MODIFY,
	/* op_sub */	DEST_MODIFY,
	/* op_subi */	DEST_MODIFY,
	/* op_subq */	DEST_MODIFY,
	/* op_muls */	DEST_MODIFY,
	/* op_mulu */	DEST_MODIFY,
	/* op_divs */	DEST_MODIFY,
	/* op_divu */	DEST_MODIFY,
	/* op_and */	DEST_MODIFY,
	/* op_andi */	DEST_MODIFY,
	/* op_or */	DEST_MODIFY,
	/* op_ori */	DEST_MODIFY,
	/* op_eor */	DEST_MODIFY,
	/* op_asl */	DEST_MODIFY,
	/* op_lsr */	DEST_MODIFY,
	/* op_jmp */	0,
	/* op_jsr */	0,
	/* op-bsr */	0,
	/* op_movem */	DEST_MODIFY,
	/* op_rts */	0,
	/* op_bra */	0,
	/* op_beq */	0,
	/* op_bne */	0,
	/* op_blt */	0,
	/* op_ble */	0,
	/* op_bgt */	0,
	/* op_bge */	0,
	/* op_bhi */	0,
	/* op_bhs */	0,
	/* op_blo */	0,
	/* op_bls */	0,
	/* op_btst */	0,
	/* op_tst */	0,
	/* op_ext */	DEST_MODIFY,
	/* op_lea */	DEST_OVERWRITE,
	/* op_swap */	DEST_OVERWRITE,
	/* op_neg */	DEST_MODIFY,
	/* op_not */	DEST_MODIFY,
	/* op_cmp */	0,
	/* op_clr */	DEST_OVERWRITE,
	/* op_link */	DEST_OVERWRITE,
	/* op_unlk */	DEST_OVERWRITE,
	/* op_pea */	DEST_OVERWRITE,
	/* op_cmpi */	0,
	/* op_dbra */	DEST_OVERWRITE,
	/* op_asr */	DEST_MODIFY,
	/* op_rol */	DEST_MODIFY,
	/* op_ror */	DEST_MODIFY,
	/* op_seq */	DEST_OVERWRITE,
	/* op_sne */	DEST_OVERWRITE,
	/* op_slt */	DEST_OVERWRITE,
	/* op_sle */	DEST_OVERWRITE,
	/* op_sgt */	DEST_OVERWRITE,
	/* op_sge */	DEST_OVERWRITE,
	/* op_shi */	DEST_OVERWRITE,
	/* op_shs */	DEST_OVERWRITE,
	/* op_slo */	DEST_OVERWRITE,
	/* op_sls */	DEST_OVERWRITE,
	/* op_st */	DEST_OVERWRITE,
	/* op_fabs */	DEST_OVERWRITE,
	/* op_fneg */	DEST_OVERWRITE,
	/* op_fadd */	DEST_OVERWRITE,
	/* op_fsub */	DEST_OVERWRITE,
	/* op_fdiv */	DEST_OVERWRITE,
	/* op_fmul */	DEST_OVERWRITE,
	/* op_fmod */	DEST_OVERWRITE,
	/* op_fcmp */	DEST_OVERWRITE,
	/* op_ftst */	DEST_OVERWRITE,
	/* op_fmove */	DEST_OVERWRITE,
	/* op_fmovem */	DEST_OVERWRITE,
#ifdef ASM
	/* op_asm */	DEST_OVERWRITE,
#endif /* ASM */
	/* op_line */	0,
	/* op_label */	0,
};

#undef P_

/*
 *	The next two #define statements are to make the
 *	code in the branch optimisation clearer.  Tests
 *	have shown that the in-line size cost is about
 *	the same as making them functions so we keep
 *	them in-line for speed.
 */

#define same_instruction(ip1,ip2) \
    ((ip1 != NIL_CODE) && (ip2 != NIL_CODE) && \
    (ip1->opcode == ip2->opcode) && \
    (ip1->length == ip2->length) && \
    (is_equal_oper (ip1->oper1, ip2->oper1)) && \
    (is_equal_oper (ip1->oper2, ip2->oper2)))


/* backup over any sequence of labels to previous instruction */
#define previous_instruction(ip) \
    do { \
	ip = ip->back; \
    } while (ip != NIL_CODE && ip->opcode == op_label)

#define branch(ip)	(ip->opcode >= op_bra && ip->opcode <= op_bls)

/*
 * find the end of a block of code.
 */
static CODE *
block_end P1(CODE*, ip)
{
    int count = 0;
    while (ip != NIL_CODE && ip->opcode != op_bra
		   && ip->opcode != op_jmp
		   && ip->opcode != op_rts) {
	if (count == BRANCH_COUNT)
	    return NIL_CODE;
	if (branch(ip))
	    count++;
 	ip = ip->fwd;
    }
    return ip;
}

/*
 * find the node which contains the label 'lab'
 */
static CODE *
find_label P1(LABEL, lab)
{
    register CODE *ip;
    for (ip = peep_head; ip != NIL_CODE; ip = ip->fwd) {
	if (ip->opcode == op_label && ip->oper1->offset->v.l == lab)
	    return ip;
    }
    /* we should have found it */
    return NIL_CODE;
}

/*
 * counts the number of times that a label node is referenced
 */
static int
label_references P1(CODE *, ip)
{
    CODE   	*target;
    struct swtab*sw;
    LABEL       i;
    LABEL	lab = ip->oper1->offset->v.l;
    int		count = 0;
    for (target = peep_head; target != NIL_CODE; target = target->fwd) {
	if ((target != ip) &&
	    (uses_label(target->oper1, lab) ||
	     uses_label(target->oper2, lab)))
	    count++;
    }
    for (sw=swtables; sw != 0; sw = sw->next) {
	for (i=0; i < sw->numlabs; i++) {
	    if (sw->labels[i] == lab)
		count++;
	}
    }
    if (ip->back != NIL_CODE && ip->back->opcode == op_label)
	count++;
    if (ip->fwd != NIL_CODE && ip->fwd->opcode == op_label)
	count++;
    return count;
}

/*
 * Returns false if the <ea> does is not a label or else isn't equal to label
 */
static BOOL
uses_label P2(ADDRESS *, ap, LABEL, label)
{
    return (ap != NIL_ADDRESS &&
	    ap->mode == am_direct &&
	    ap->offset->nodetype == en_labcon &&
	    ap->offset->v.l == label);
}

/*
 * Returns false if the <ea> of ap1 is not used in the <ea> of ap2, otherwise
 * it returns true.  If we aren't sure then return true anyway.
 */
static BOOL
uses_address P2(ADDRESS *, ap1, ADDRESS *, ap2)
{
    if (ap1 == NULL || ap2 == NULL)
	return FALSE;
    switch (ap1->mode) {
      case am_areg:
	switch (ap2->mode) {
	  case am_areg:
	  case am_ainc:
	  case am_adec:
	  case am_ind:
	  case am_indx:
	  case am_indx2:
	  case am_indx4:
	    return ap2->preg == ap1->preg;
	  case am_smask:
	  case am_rmask:
	    return ((ap2->offset->v.i >> (15-(int)ap1->preg)) & 1L) != 0L;
	  case am_dreg:
	  case am_immed:
	  case am_direct:
	  case am_none:
	  case am_line:
	    return FALSE;
	  default:
	    break;
	}
	break;
      default:
	break;
    }
    return TRUE;
}

/*
 * Checks to see if the addressing mode ap is overwritten with a new value
 * before the value is used.
 */
static BOOL
dest_overwritten P2(ADDRESS *, ap, CODE *, ip)
{
    CODE   *ip2;
    BOOL   result = FALSE;
    OPCODE op;
    if (ip == NIL_CODE)
	return FALSE;
    if (ap->mode != am_areg)
	return FALSE;
    for (ip2 = ip->fwd; ip2 != NIL_CODE; ip2 = ip2->fwd) {
	switch (ip2->opcode) {
	  case op_unlk:
	  case op_rts:
	    return TRUE;
	  case op_jmp:
	  case op_nop:
	    return FALSE;
	  case op_beq: /* use of A7 unknown */
	  case op_bne:
	  case op_bgt:
	  case op_bge:
	  case op_blt:
	  case op_ble:
	  case op_bls:
	  case op_blo:
	  case op_bhi:
	  case op_bhs:
	    op = ip2->opcode;
	    ip2->opcode = op_nop;	/* to prevent looping */
	    result = dest_overwritten (ap, ip2) &&
		     dest_overwritten (ap, find_label(ip2->oper1->offset->v.l));
	    ip2->opcode = op;
	    return result;
	  case op_bra:
	    op = ip2->opcode;
	    ip2->opcode = op_nop;	/* to prevent looping */
	    result = dest_overwritten (ap, find_label(ip2->oper1->offset->v.l));
	    ip2->opcode = op;
	    return result;
	  case op_pea: /* implicit use of A7 */
	  case op_link:
	    if (ap->preg == STACKPTR)
		return FALSE;
	    break;
	  default:
	    if (uses_address(ap, ip2->oper1) || uses_address(ap, ip2->oper2))
		return FALSE;
	    break;
	  case op_label:
#ifdef ASM
	  case op_asm:
#endif /* ASM */
	    break;
	}
    }
    return FALSE;
}

static CODE *
code P4(OPCODE, op, ILEN, len, ADDRESS *, ap1, ADDRESS *, ap2)
{
    CODE   *ip;
    ip = (CODE *) xalloc((size_t) sizeof(CODE));
    ip->opcode = op;
    ip->length = len;
    ip->oper1 = ap1;
    ip->oper2 = ap2;
    return ip;
}

/*
 * generate a code sequence into the peep list.
 */
void
g_code P4(OPCODE, op, ILEN, len, ADDRESS *, ap1, ADDRESS *, ap2)
{
    add_peep(code(op, len, ap1, ap2));
}

/*
 * generate a code sequence and put it in front of the list
 */
void
g_coder P4(OPCODE, op, ILEN, len, ADDRESS *, ap1, ADDRESS *, ap2)
{
    CODE   *ip;
    ip = code(op, len, ap1, ap2);
    if (peep_head == NIL_CODE) {
	/* must use add_peep to take care of peep_tail */
	add_peep(ip);
    } else {
        ip->back  = NIL_CODE;
        ip->fwd   = peep_head;
	peep_head->back = ip;
	peep_head  = ip;
	/* peep_tail does not change */
    }
}

/*
 * add the instruction pointed to by new to the peep list.
 */
static void
add_peep P1(CODE *, ip)
{
    static CODE *peep_tail;
    if (peep_head == NIL_CODE) {
	peep_head = peep_tail = ip;
	ip->fwd = NIL_CODE;
	ip->back = NIL_CODE;
    } else {
	ip->fwd = NIL_CODE;
	ip->back = peep_tail;
	peep_tail->fwd = ip;
	peep_tail = ip;
    }
}

/*
 * output all code and labels in the peep list.
 */
void
flush_peep P0(void)
{
    register CODE *ip;
    struct swtab  *sw;
    EXPR	  *ep2;
    LABEL	   i;
    opt3();			/* do the peephole optimizations */
    for (ip = peep_head; ip != NIL_CODE; ip = ip->fwd) {
	if (ip->opcode == op_label)
	    put_label(ip->oper1->offset->v.l);
	else
	    put_code(ip);
    }
    peep_head = NIL_CODE;
    for(sw=swtables; sw; sw=sw->next) {
	put_kseg(g_alignment(tp_pointer));
	put_dlabel(sw->tablab);
	ep2 = mk_lcon(UNDEF_LABEL);
#ifdef RELOC_BUG
	/* generate the switch jump table as a series of 4-byte addresses */
	for(i=0; i<sw->numlabs; i++) {
	    ep2->v.l = sw->labels[i];
	    put_pointer(ep2);
	}
#else
	/* generate the switch jump table as a series of 2-byte offsets
	 * This limits the amount of code that can be generated in a
	 * function to less then 32K.  I believe that this is a reasonable
	 * restriction.
	 */
	{
	    EXPR	  *ep, *ep1;
	    ep1 = mk_lcon(sw->beglab);
	    ep = mk_node(en_sub, ep2, ep1, tp_void);
	    for(i=0; i<sw->numlabs; i++) {
		ep2->v.l = sw->labels[i];
		put_short(ep);
	    }
	}
#endif /* RELOC_BUG */
    }
    swtables = NULL;
}

/*
 * delete an instruction referenced by ip
 */
static void
peep_delete P1(CODE *, ip)
{
    if (ip == NIL_CODE || ip->back == NIL_CODE)
	FATAL ((__FILE__,"peep_delete",""));

    if ((ip->back->fwd = ip->fwd) != NIL_CODE)
	ip->fwd->back = ip->back;

    next_ip = ip->back;
    changes++;
}

/*
 * changes lea <ea>,An + pea (An) to pea <ea>
 *  The value of An is not needed (An is scratch register)
 * CAVEAT code generator modifier!
 */
static void
peep_pea P1(CODE *, ip)
{
    CODE   *prev;
    if (ip->oper1->mode != am_ind)
	return;
    if ((prev = ip->back) == NIL_CODE)
	return;
    if (prev->opcode == op_lea && prev->oper2->preg == ip->oper1->preg
	&& ip->oper1->preg <= MAX_ADDR) {
	prev->opcode = op_pea;
	prev->oper2 = NIL_ADDRESS;
	peep_delete(ip);
    }
}

/*
 * deletes lea (An),An
 * changes lea <ea>,An + move.l An,Am to lea <ea>,Am
 * The value of An is not needed (An scratch, Am typically tempref)
 * CAVEAT code generator modifier!
 */
static void
peep_lea P1(CODE *, ip)
{
    CODE   *next;
    if (ip->oper1->mode == am_ind && ip->oper1->preg == ip->oper2->preg) {
	peep_delete(ip);
	return;
    }
    if ((next = ip->fwd) == NIL_CODE)
	return;
    if (next->opcode == op_move && ip->oper2->preg <= MAX_ADDR &&
	next->oper1->mode == am_areg && next->oper1->preg == ip->oper2->preg
	&& next->oper2->mode == am_areg && next->oper2->preg > MAX_ADDR) {
	ip->oper2 = next->oper2;
	peep_delete(ip->fwd);
    }
    if (ip->oper2->preg == STACKPTR && dest_overwritten(ip->oper2, ip)) {
	peep_delete(ip);
	return;
    }
}

/*
 * Determine whether a move is redundant ... this is done by looking back
 * along the code list (following all branches to labels) to determine
 * whether the destination already contains the necessary result.
 */
static BOOL
redundent_move P3(CODE*, ip, CODE*, ip2, BOOL, memory)
{
    BOOL	 altered;
    LABEL	 label;
    struct swtab*sw;

    for (ip2 = ip2->back; ip2 != NIL_CODE; ip2 = ip2->back) {
	switch (ip2->opcode) {
	  case op_label:
	    label = ip2->oper1->offset->v.l;

	    /* first check code before the label */
	    if (!redundent_move (ip, ip2, memory)) {
		return FALSE;
	    }

	    /* ... and then check all branches to this label */
	    for (ip2 = peep_head; ip2 != NIL_CODE; ip2 = ip2->fwd) {
		switch (ip2->opcode) {
		  case op_beq:
		  case op_bne:
		  case op_bgt:
		  case op_bge:
		  case op_blt:
		  case op_ble:
		  case op_bls:
		  case op_blo:
		  case op_bhi:
		  case op_bhs:
		  case op_bra:
		    if (uses_label(ip2->oper1, label)) {
			 OPCODE	op = ip2->opcode;
			ip2->opcode = op_nop;
			if (!redundent_move (ip, ip2, memory)) {
			    ip2->opcode = op;
			    return FALSE;
			}
			ip2->opcode = op;
		    }
		    break;
		  default:
		    break;
		}
	    }

	    /* but if it is via a jump table we cannot determine it */
	    for (sw=swtables; sw != 0; sw = sw->next) {
		LABEL	i;
		for (i=0; i < sw->numlabs; i++) {
		    if (sw->labels[i] == label)
			return FALSE;
		}
	    }
	    return TRUE;
	  case op_nop:
	    return TRUE;
#ifdef ASM
	  case op_asm:
#endif /* ASM */
	  case op_jsr:
	  case op_bsr:
	  case op_jmp:
	  case op_rts:
	  case op_bra:
	    return FALSE;
	  default:
	    break;
	}
	if (same_instruction(ip, ip2)) {
	    return TRUE;
	}
	altered = (BOOL)(op_flags[ip2->opcode] & DEST_ALTERED);
	if (ip2->oper2) {
	    /* two operand instruction */
	    if (altered && is_equal_oper (ip->oper1, ip2->oper2))
		return FALSE;
	    switch (ip2->oper2->mode) {
	      case am_areg:
		if (altered && ((ip->oper2->preg == ip2->oper2->preg) ||
				(ip->oper1->preg == ip2->oper2->preg)))
		    return FALSE;
		break;
	      case am_ainc:
	      case am_adec:
		if (ip->oper2->preg == ip2->oper2->preg)
		    return FALSE;
		/*FALLTHRU*/
	      case am_ind:
	      case am_indx:
		if (altered && memory)
		    return FALSE;
		break;
	      default:
		break;
	    }
	}
	if (ip2->oper1) {
	    switch (ip2->oper1->mode) {
	      case am_areg:
		if (ip2->oper2 != NIL_ADDRESS)
		    break;
		/* one operand instruction */
		if (altered && ((ip->oper2->preg == ip2->oper1->preg) ||
				(ip->oper1->preg == ip2->oper1->preg)))
		    return FALSE;
		break;
	      case am_ainc:
	      case am_adec:
		if (ip->oper2->preg == ip2->oper1->preg)
		    return FALSE;
		/*FALLTHRU*/
	      case am_ind:
	      case am_indx:
		if (ip2->oper2 != NIL_ADDRESS)
		    break;
		/* one operand instruction */
		if (altered && memory)
		    return FALSE;
		break;
	      default:
		break;
	    }
	}
    }
    return FALSE;
}

/*
 * peephole optimization for MOVE instructions.
 */
static void
peep_move P1(CODE*, ip)
{
    CODE   *ip2;
    EXPR   *ep;

    /*
     * move.w An,An changes the contents of An through sign extension
     */
    if (is_equal_address(ip->oper1, ip->oper2) &&
       (ip->oper1->mode != am_areg || ip->length != IL2)) {
	peep_delete(ip);
	return;
    }
    /* eliminate redundant move with registers on sequence:
     *		MOVE	Xn, Xm
     *		MOVE	Xm, Xn
     * except when Xm is an address register and Xn is a data register - this
     * set the flags so cannot be deleted.
     */
    if ((ip->back != NIL_CODE) && (ip->back->opcode == op_move) &&
    	(ip->oper1->mode == am_dreg || ip->oper1->mode == am_areg) &&
	(ip->oper2->mode == am_areg || ip->oper1->mode == ip->oper2->mode) &&
	(ip->length == ip->back->length) &&
	is_equal_address (ip->oper1, ip->back->oper2) &&
	is_equal_address (ip->oper2, ip->back->oper1)) {
	peep_delete(ip);
	return;
    }

    switch (ip->oper1->mode) {
      case am_ind:
      case am_indx:
	if (ip->oper2->mode != am_areg)
	    break;
	if (ip->oper1->preg == ip->oper2->preg)
	    break;
	if (redundent_move (ip, ip, TRUE))
	    peep_delete(ip);
	break;
      case am_dreg:
      case am_areg:
	if (ip->oper2->mode != am_areg)
	    break;
	if (redundent_move (ip, ip, FALSE))
	    peep_delete(ip);
	break;
      case am_immed:
	ep = ip->oper1->offset;

	if (ip->length == IL4) {
	    /*
	     *   Replace move immediate to stack by PEA
	     */
	    if (ip->oper2->mode == am_adec && ip->oper2->preg == STACKPTR) {
		ip->opcode = op_pea;
		ip->oper2 = NIL_ADDRESS;
		ip->oper1 = copy_addr(ip->oper1, am_direct);
		ip->length = IL0;
		changes++;
		return;
	    }
	    /*
	     * changes move immediate to An	to LEA.
	     */
	    if (ip->oper2->mode == am_areg) {
		ip->opcode = op_lea;
		ip->oper1 = copy_addr(ip->oper1, am_direct);
		ip->length = IL0;
		changes++;
		return;
	    }
	}
	if (ep->nodetype == en_icon) {
	    /*
	     *  Replace move immediate An for the following cases
	     *		MOVE #0, An
	     *		MOVE #shortval, An
	     *  by the sequences
	     *		SUB.L An,An
	     *	and	MOVE.W #shortval, An
	     */
	    if (ip->oper2->mode == am_areg) {
		if (ep->v.i == 0L) {
		    ip->length = IL4;
		    ip->opcode = op_sub;
		    ip->oper1 = ip->oper2;
		    changes++;
		} else if ((IVAL)-32768L <= ep->v.i && ep->v.i <= (IVAL)32767L)
		    ip->length = IL2;
		return;
	    }
	    /*
	     *	Replace moving small constants into a data register by
	     *		MOVEQ	#value,Dn
	     */
	    if (ip->oper2->mode == am_dreg) {
		if ((IVAL) -128 <= ep->v.i && ep->v.i <= (IVAL) 127) {
		    ip->opcode = op_moveq;
		    ip->length = IL0;
		    changes++;
		    return;
		}
	    }
	    /*
	     *  Replace the sequence
	     *		MOVE.W #X, -(An)
	     *		MOVE.W #Y, -(An)
	     *  by the sequence
	     *		MOVE.L #(X<<16|Y), -(An)
	     */
	    ip2 = ip->fwd;
	    if (ip->length == IL2 &&
		ip->oper2->mode == am_adec &&
		ip2->length == IL2 &&
		ip2->opcode == op_move &&
		ip2->oper1->mode == am_immed &&
		ip2->oper1->offset->nodetype == en_icon &&
		ip2->oper2->mode == am_adec &&
		ip2->oper2->preg == ip->oper2->preg) {
		ip->length = IL4;
		ep->v.u = (ep->v.u & (UVAL) 0xffffL) |
					 (ip2->oper1->offset->v.u << 16);
		peep_delete(ip2);
		return ;
	    }
	    /*
	     * The M68000 reads the operand to be cleared.
	     * This is unsafe e.g. when manipulating interface
	     * registers, and clr is not too fast for this reason
	     * (although it is a considerablt smaller instruction).
	     * We will therefore only do this optimisation on the
	     * M68000 if volatile is not set.
	     *
	     * M68010, M68020 etc. do not have this bug so always
	     * replace move immediate of zero by a CLR instruction.
	     *
	     * N. B. The special case of moving zero to a register
	     * will already have been optimised to MOVEQ or SUBA.
	     */
#ifdef MC68000
	    if (!volatile_found)
#endif /* ! MC68000 */
		if (ep->v.i == 0L) {
		    ip->opcode = op_clr;
		    ip->oper1 = ip->oper2;
		    ip->oper2 = NIL_ADDRESS;
		    changes++;
		    return;
		}
	}
	break;
      default:
	break;
    }
}

/*
 * compare two address nodes and return true if they are equivalent.
 */
BOOL
is_equal_address P2(ADDRESS *, ap1, ADDRESS *, ap2)
{
    if (ap1 == NIL_ADDRESS || ap2 == NIL_ADDRESS)
	return FALSE;
    if (ap1->mode != ap2->mode)
	return FALSE;
    switch (ap1->mode) {
      case am_areg:
      case am_dreg:
      case am_ind:
	return ap1->preg == ap2->preg;
      case am_indx:
	return ap1->preg == ap2->preg &&
               ap1->offset->nodetype == en_icon &&
               ap2->offset->nodetype == en_icon &&
	       ap1->offset->v.i == ap2->offset->v.i;
      case am_indx2:
      case am_indx3:
      case am_indx4:
	return ap1->sreg == ap2->sreg &&
	       ap1->preg == ap2->preg &&
               ap1->offset->nodetype == en_icon &&
               ap2->offset->nodetype == en_icon &&
	       ap1->offset->v.i == ap2->offset->v.i;
      case am_rmask:
      case am_smask:
	return ap1->offset->v.i == ap2->offset->v.i;
      default:
	break;
    }
    return FALSE;
}

/*
 * peephole optimization for movem instructions. movem instructions are used
 * to save registers on the stack. if only one register is being saved we
 * can convert the movem to a move instruction.
 *
 */
static void
peep_movem P1(CODE*, ip)
{
    REG		 reg;
    ADDRESS	*ap = (ip->oper1->mode == am_smask) ? ip->oper1 : ip->oper2;
    REGMASK mask = (REGMASK)ap->offset->v.i;

    for (reg = A7; reg >= D0; reg--) {
	if (mask == (REGMASK)(1 << (int)reg)) {
	    if (reg >= A0) {
		ap->mode = am_areg;
	    } else
		ap->mode = am_dreg;
	    ap->preg = reg;
	    ip->opcode = op_move;
	    break;
	}
    }

    /* optimises sequence:   move ea1, ea2
			     move ea2, ea1
    */
    if (ip->back && ip->back->opcode == ip->opcode) {
	if (is_equal_address(ip->oper1, ip->back->oper2) &&
	    is_equal_address(ip->oper2, ip->back->oper1))
		peep_delete(ip);
    }
}

/*
 * peephole optimization for add instructions.
 * makes quick immediates out of small constants (redundant for most as's).
 * change add immediate to address registers to lea where possible
 */
static void
peep_add P1(CODE*, ip)
{
    EXPR   *ep;
    if (ip->oper1->mode != am_immed)
	return;
    ep = ip->oper1->offset;
    if (ip->oper2->mode == am_areg) {
	if (ip->oper2->preg == STACKPTR && dest_overwritten(ip->oper2, ip)) {
	    peep_delete(ip);
	    return;
	}
	/* adda.w extents the first argument automatically */
	if (is_short(ep))
	    ip->length = IL2;
    } else {
	ip->opcode = op_addi;
    }
    if (ep->nodetype != en_icon)
	return;
    if (ep->v.i == 0L) {
	if (ip->oper2->mode == am_areg) {
	    peep_delete(ip);
	} else {
	    ip->oper1 = ip->oper2;
	    ip->oper2 = NIL_ADDRESS;
	    ip->opcode = op_tst;
	    changes++;
	}
	return;
    }
    if ((IVAL) 1 <= ep->v.i && ep->v.i <= (IVAL) 8) {
	ip->opcode = op_addq;
	return;
    }
    if ((IVAL) -8 <= ep->v.i && ep->v.i <= (IVAL) -1) {
	ep->v.i = -ep->v.i;
	ip->opcode = op_subq;
	changes++;
	return;
    }
    if (ip->oper2->mode == am_areg && is_short(ep)) {
	/*
	 * 	adda.w #X, An   and 	lea X(An), An 
	 * are the same size, but the lea version is quicker
	 */
	ip->oper1 = copy_addr(ip->oper1, am_indx);
	ip->oper1->preg = ip->oper2->preg;
	ip->length = IL0;
	ip->opcode = op_lea;
	changes++;
	return;
    }
}

/*
 * conversion of unsigned data types often yields statements like
 * move.b source,d0 +  andi.l #255,d0
 * which should be converted to
 * clr.l d0 + move.b source,d0
 * deletes and #-1
 */
static void
peep_and P1(CODE*, ip)
{
    CODE   *prev;
    ILEN    size;
    IVAL    val;
    if (ip->oper1->mode != am_immed ||
	ip->oper1->offset->nodetype != en_icon)
	return;
    val = ip->oper1->offset->v.i;
    /*
     * and #-1 does only set flags, which the code generator does not know
     */
    if (val == -1L) {
	peep_delete(ip);
	return;
    }
    if (ip->oper1->mode != am_immed ||
	(val != 255L && val != 65535L))
	return;

    size = ((val == 255L) ? IL1 : IL2);

    if ((prev = ip->back) == NIL_CODE || prev->length != size
	|| prev->opcode != op_move
	|| ip->oper2->mode != am_dreg || prev->oper2->mode != am_dreg
	|| ip->oper2->preg != prev->oper2->preg
        ||      (prev->oper1->mode == am_indx2 &&
                 prev->oper1->sreg == prev->oper2->preg)
        )
	return;

    prev->length = ip->length;
    ip->length = size;
    prev->opcode = op_clr;
    ip->opcode = op_move;
    ip->oper1 = prev->oper1;
    prev->oper1 = prev->oper2;
    prev->oper2 = NIL_ADDRESS;

    next_ip = prev;
}

/*
 * removes consecutive clr-statements
 *
 */
static void
peep_clr P1(CODE *, ip)
{
    CODE   *prev;

    if ((prev = ip->back) == NIL_CODE || prev->opcode != op_clr ||
	!is_equal_address(ip->oper1, prev->oper1))
	return;

    if (prev->length < ip->length)
	prev->length = ip->length;

    peep_delete(ip);
}

/*
 * peephole optimization for subtract instructions.
 * makes quick immediates out of small constants (redundant for most as's).
 */
static void
peep_sub P1(CODE*, ip)
{
    EXPR   *ep;
    if (ip->oper1->mode != am_immed)
	return;
    ep = ip->oper1->offset;
    if (ip->oper2->mode == am_areg) {
	/* suba.w extents the first argument automatically */
	if (is_short(ep))
	    ip->length = IL2;
    } else
	ip->opcode = op_subi;
    if (ep->nodetype != en_icon)
	return;
    if (ep->v.i == 0L && ip->oper2->mode != am_areg) {
	ip->oper1 = ip->oper2;
	ip->oper2 = NIL_ADDRESS;
	ip->opcode = op_tst;
	changes++;
	return;
    }
    if ((IVAL) 1 <= ep->v.i && ep->v.i <= (IVAL) 8) {
	ip->opcode = op_subq;
	return;
    } else if ((IVAL) -8 <= ep->v.i && ep->v.i <= (IVAL) -1) {
	ep->v.i = -ep->v.i;
	ip->opcode = op_addq;
	changes++;
	return;
    }
}

/*
 * peephole optimization for compare instructions.
 * changes compare #0 to tst
 */
static void
peep_cmp P1(CODE*, ip)
{
    EXPR   *ep;
    if (ip->oper1->mode != am_immed)
	return;
    ep = ip->oper1->offset;
    if (ip->oper2->mode == am_areg) {
	/* cmpa.w extents the first argument automatically */
	if (is_short(ep))
	    ip->length = IL2;
	return;
    }
    ip->opcode = op_cmpi;
    if (ep->nodetype != en_icon || ep->v.i != 0L)
	return;
    ip->oper1 = ip->oper2;
    ip->oper2 = NIL_ADDRESS;
    ip->opcode = op_tst;
    next_ip = ip;
}

/*
 * deletes a tst instruction if the flags are already set.
 */
static void
peep_tst P1(CODE *, ip)
{
    CODE   *prev;
    prev = ip->back;
    if (prev == NIL_CODE)
	return;
    /*
     * All the following cases used to be checked with an obscure
     * if-statement. There was an error in it that caused prev->oper2->mode
     * to be referenced even if prev->opcode is op_clr. (This yields a NIL-
     * pointer reference.
     * I threw all this stuff away and replaced it by this case-statement.
     * This is much more readable.
     */
    switch (prev->opcode) {
      case op_label:
	/*
	 * List all pseudo-instructions here. Of course, they do not do
	 * anything.
	 */
#ifdef ASM
      case op_asm:
#endif /* ASM */
	return;

      case op_add:
      case op_addi:
      case op_addq:
      case op_sub:
      case op_subi:
      case op_subq:
      case op_asl:
      case op_asr:
      case op_neg:
	/*
	 * These instructions can set the carry and overflow flags which a
	 * tst instruction always clears.
	 */
	return;

      case op_mulu:
      case op_muls:
	/*
	 * These instructions set flags on full 32 bit register.
	 */
	if (prev->length < IL4)
	    return;
	break;

      case op_move:
	/*
	 * A move TO an address register does not set the flags
	 */
	if (prev->oper2->mode == am_areg)
	    return;
	/*FALLTHRU*/
      case op_moveq:
      case op_clr:
	/*
	 * All other move an clr instructions set the flags according to
	 * the moved operand, which is prev->oper1
	 */
	if (is_equal_address(prev->oper1, ip->oper1))
	    break;
	/*FALLTHRU*/
      default:
	/*
	 * If instructions have different length then the falgs will not be
	 * set correctly.
	 */
	if (prev->length != ip->length)
	    return;
	/*
	 * All instructions that have a target set the flags according to the
	 * target (prev->oper2)
	 * Note that is_equal_address may be called with a NIL pointer
	 */
	if (is_equal_address(prev->oper2, ip->oper1))
	    break;
	return;
    }
    /*
     * We come here if the flags are already set, thus the tst
     * instruction can be deleted.
     */
    peep_delete(ip);
}

/*
 * peephole optimization for unconditional transfers. deletes instructions
 * which have no path. applies to bra, jmp, and rts instructions.
 */
static void
peep_uctran P1(CODE *, ip)
{
    while (ip->fwd != NIL_CODE && ip->fwd->opcode != op_label)
	peep_delete(ip->fwd);
}

/*
 * optimizes conditional branch over a bra.
 */
static void
peep_bxx P1(CODE *, ip)
{
    CODE   *next = ip->fwd;
    if (next == NIL_CODE)
	return;
    if (next->opcode == op_bra) {
       /* peep_uctran increases the 'hit' probability */
        peep_uctran(next);
	next = next->fwd;
	if (next == NIL_CODE)
	    return;
	if (next->opcode == op_label &&
	    ip->oper1->offset->v.l == next->oper1->offset->v.l) {
	    ip->fwd->opcode = revcond[(int)ip->opcode - (int)op_beq];
	    peep_delete(ip);
	}
    }
}

static void
peep_ext P1(CODE *, ip)
{
    CODE * next = ip->fwd;
    if (next == NIL_CODE)
	return;
    if (ip->length == IL4) {
	switch (next->opcode) {
	  case op_move:
	  case op_add:
	  case op_sub:
	    if ((next->oper2->mode == am_areg) &&
		is_equal_oper (next->oper1, ip->oper1)) {
		/*
		 * convert	ext.l Dn	=>	move.w Dn,An
		 *		move.l Dn, An
		 */
		peep_delete(ip);
		next->length = IL2;
	    }
	    break;
	  default:
	    break;
	}
    }
}

/*
 * For small shifts left, it can be more efficient to generate ADD
 * instructions.  Although there is a small speed gain when the shift
 * is by 2 there is a space penalty ... we go for space not speed.
 */
static void
peep_shift P1(CODE*, ip)
{
    if (ip->oper1->mode != am_immed)
	return;

    if (ip->oper1->offset->v.i == (IVAL) 1) {
	ip->opcode = op_add;
	ip->oper1 = copy_addr(ip->oper2, ip->oper2->mode);
	changes++;
    }
}

/*
 * if a label is followed by a branch to another label, the
 * branch statement can be deleted when the label is moved
 */
static void
peep_label P1(CODE *, ip)
{
    CODE   *prev, *next, *target;
    struct swtab   *sw;
    LABEL	    i, lab, label;

    if ((next = ip->fwd) == NIL_CODE)
	return;

    if (!optimize_option)
	return;

    lab = ip->oper1->offset->v.l;
    switch (next->opcode) {
      case op_label:
	/* if a label is followed by a label then common them up */
	label = next->oper1->offset->v.l;
	for (target = peep_head; target != NIL_CODE; target = target->fwd) {
	    if (uses_label ( target->oper1, label))
		target->oper1->offset->v.l = lab;
	    if (uses_label ( target->oper2, label))
		target->oper2->offset->v.l = lab;
	}
	for (sw=swtables; sw != 0; sw = sw->next) {
	    if (sw->beglab == label)
		sw->beglab = lab;
	    for (i=0; i < sw->numlabs; i++) {
		if (sw->labels[i] == label)
		    sw->labels[i] = lab;
	    }
	}
	peep_delete(next);
	break;

      case op_bra:
	prev = ip->back;
	/*
	 * To make this fast, assume that the label number is really
	 * next->oper1->offset->v.l
	 */
	label = next->oper1->offset->v.l;
	if (label == lab)
	    return;
	target = find_label(label);
	if (target == NIL_CODE) {
	   message(MSG_PEEPLABEL);
	   return;
	}
	/* move label */
	if (target->fwd == ip)
	    return;
	peep_delete(ip);
	ip->fwd = target->fwd;
	ip->back = target;
	target->fwd = ip;
	if (ip->fwd != NIL_CODE)
	    ip->fwd->back = ip;
	/* possibly remove branches */
	/* in fact, prev is always != 0 if peep_delete has succeeded */
	if (prev != NIL_CODE) {
	    if (prev->opcode == op_bra || prev->opcode == op_jmp
		|| prev->opcode == op_rts)
		peep_uctran(prev);
	}
	break;
      default:
	/* check that there are still references to this label */
	if (label_references(ip) == 0)
	    peep_delete(ip);
	break;
    }
}

static BOOL
is_equal_oper  P2(ADDRESS *, ap1, ADDRESS *, ap2)
{
    if (ap1 == NIL_ADDRESS && ap1 == NIL_ADDRESS)
	return TRUE;
    if (ap1 == NIL_ADDRESS || ap2 == NIL_ADDRESS)
	return FALSE;
    if (ap1->mode != ap2->mode)
	return FALSE;
    switch (ap1->mode) {
      case am_areg:
      case am_dreg:
      case am_ind:
      case am_ainc:
      case am_adec:
	return ap1->preg == ap2->preg;
      case am_indx:
	return ap1->preg == ap2->preg &&
	    ap1->offset->v.i == ap2->offset->v.i;
      case am_indx2:
      case am_indx3:
      case am_indx4:
	return
	    ap1->preg == ap2->preg &&
	    ap1->sreg == ap2->sreg &&
	    ap1->offset->v.i == ap2->offset->v.i;
      case am_direct:
	    return is_equalnode (ap1->offset, ap2->offset);
      case am_immed:
      case am_smask:
      case am_rmask:
	return ap1->offset->v.i == ap2->offset->v.i;
      case am_none:
      case am_line:
	return TRUE;
      default:
	break;
    }
    return FALSE;
}


/* ensure we have a label to branch to (create if needed) */
static void
check_label P2(CODE *, ip, CODE *, target)
{
    if (target->fwd->opcode == op_label) {
	ip->oper1->offset->v.l = target->fwd->oper1->offset->v.l;
    } else {
	CODE   *p;

	p = code (op_label, IL0, mk_label(nextlabel), NIL_ADDRESS);
	p->fwd = target->fwd;
	p->back= target;
	target->fwd = p->fwd->back = p;
	ip->oper1->offset->v.l = nextlabel++;
    }
}


/* delete branches to the following statement */
static void
peep_bra P1(CODE *, ip)
{
    CODE   *p = ip->fwd;
    CODE   *target;
    LABEL           label = ip->oper1->offset->v.l;
    int		    count;

    /* delete branches to the following statement */
    while (p != NIL_CODE && p->opcode == op_label) {
	if (p->oper1->offset->v.l == label) {
	    peep_delete(ip);
	    return;
	}
	p = p->fwd;
    }

    if (!optimize_option)
	return;

    target = find_label(label);
    /* we should have found it */
    if (target == NIL_CODE)
	FATAL ((__FILE__,"peep_bra","target == 0"));

    /* Space optimisation:
     * if the code before the target of the branch is itself a branch
     * then we can move the destination block of code to eliminate the branch
     */
    p = target->back;
    if (p != NULL && (  (p->opcode == op_bra)
		     || (p->opcode == op_jmp)
		     || (p->opcode == op_rts))) {
	p = block_end(target);
	if (p != NULL && p != ip) {
	    if (ip->fwd)
		ip->fwd->back = p;
	    if (p->fwd)
		p->fwd->back = target->back;
	    target->back->fwd = p->fwd;
	    p->fwd = ip->fwd;
	    target->back = ip;
	    ip->fwd = target;
	    peep_delete(ip);
	    return;
	}
    }

    /* Space optimisation:
     * if the code before the branch instruction is the same as the
     * instruction before the label the branch can be moved back an
     * instruction.
     */

    p = ip->back;
    previous_instruction(target);
    if (p==target)
	return;
    /* now skip back over identical instruction sequences */
    while (same_instruction(p,target)) {
	p = p->back;
	previous_instruction(target);
	peep_delete(p->fwd);
    }
    check_label(ip, target);
    label = ip->oper1->offset->v.l;

    /*  Space optimisation:
     *  Optimise for the situation where two branches to the same
     *  target label have identical instruction sequences 
     *  leading up to the branch.   We can instead eliminate one
     *  of these instruction sequences by branching to the other one.
     */
    for (target = peep_head; target != NIL_CODE; target = target->fwd) {
	if ((target != ip) &&
	    (target->opcode == op_bra) &&
	    (target->oper1->offset->v.l == label)) {

    	    CODE   *t = target;
	    p = ip->back;
	    previous_instruction(t);
	    count=0;
	    while (same_instruction(p,t)) {
		p = p->back;
		previous_instruction(t);
		peep_delete(p->fwd);
		count++;
	    }
	    if (count != 0) {
		check_label (ip, t);
		break;
	    }
 
	}
    }

    peep_uctran(ip);
}

static void
peep_jmp P1(CODE *, ip)
{
    CODE   *target;
    if (!optimize_option)
	return;
    /*
     * see if we can find a jmp just like this one - if so we shall branch
     * to it.  This allows the tail end of the code for switch jump tables
     * to be commoned up.  This is space optimisation!
     */
    for (target = peep_head; target != NIL_CODE; target = target->fwd) {
	if ((target != ip) && (same_instruction(ip, target))) {
	    ip->opcode = op_bra;
	    ip->oper1 = mk_label(0);	/* label filled in by next line */
	    check_label(ip, target->back);
	}
    }
    peep_uctran(ip);
}

/* delete multiple debugging line statements */
static void
peep_line P1(CODE *, ip)
{
    CODE   *ip2;
    if (ip->fwd == NULL)
	return;
    switch (ip->fwd->opcode) {
      case op_line:
	peep_delete(ip);
	break;
      case op_label:
	/* move the line number to after the label */
	ip2 = ip->fwd;
	if (ip->back)
	    ip->back->fwd = ip2;
	if (ip2->fwd)
	    ip->fwd->back = ip;
	ip2->back = ip->back;
	ip->fwd = ip2->fwd;
	ip2->fwd = ip;
	ip->back = ip2;
	break;
      default:
	break;
    }
}

/*
 * peephole optimizer. This routine calls the instruction specific
 * optimization routines above for each instruction in the peep list.
 */
static void
opt3 P0(void)
{
    CODE   *ip;
    if (! opt_option)
        return;
    do {
	changes = 0;
	next_ip = peep_head;
	while (next_ip != NIL_CODE) {
	    ip = next_ip;
	    next_ip = ip->fwd;
	    switch (ip->opcode) {
	      case op_move:
		peep_move(ip);
		break;
	      case op_movem:
		peep_movem(ip);
		break;
	      case op_pea:
		peep_pea(ip);
		break;
	      case op_lea:
		peep_lea(ip);
		break;
	      case op_add:
	      case op_addq:
		peep_add(ip);
		break;
	      case op_and:
		peep_and(ip);
		break;
	      case op_clr:
		peep_clr(ip);
		break;
	      case op_sub:
		peep_sub(ip);
		break;
	      case op_cmp:
		peep_cmp(ip);
		break;
	      case op_tst:
		peep_tst(ip);
		break;
	      case op_ext:
		peep_ext(ip);
		break;
	      case op_asl:
		peep_shift(ip);
		break;
	      case op_beq:
	      case op_bne:
	      case op_bgt:
	      case op_bge:
	      case op_blt:
	      case op_bls:
	      case op_blo:
	      case op_bhi:
	      case op_bhs:
		peep_bxx(ip);
		break;
	      case op_rts:
		peep_uctran(ip);
		break;
	      case op_line:
		peep_line(ip);
		break;
	      case op_label:
		peep_label(ip);
		break;
	      case op_bra:
		peep_bra(ip);
		break;
	      case op_jmp:
		peep_jmp(ip);
		break;
	      default:
		break;
	    }
	}
#ifdef VERBOSE
	if (verbose_option && changes)
	    message (MSG_PEEPCHANGES, changes) ;
#endif /* VERBOSE */
    } while (changes);
}
#endif /* MC680X0 */
